"""Configuration file for the Sphinx documentation builder."""

from kernels import kernels_download
from sphinx_gallery.sorting import ExplicitOrder

from moon_coverage import __version__


# -- Project information -----------------------------------------------------

project = 'Moon Coverage Documentation'
copyright = '2020-2022 - All right reserved'  # pylint: disable=W0622
author = 'LPG Nantes'
version = __version__
release = __version__

# -- General configuration ---------------------------------------------------

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'sphinx.ext.intersphinx',
    'sphinx_copybutton',
    'sphinx_gallery.gen_gallery',
    'myst_nb',
]

master_doc = 'index'
language = 'en'
exclude_patterns = [
    '_build',
    'Thumbs.db',
    '.DS_Store',
    '.jupyter_cache',
    'jupyter_execute',
    '**.ipynb_checkpoints',
]

# -- Mapping -----------------------------------------------------------------

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'numpy': ('https://numpy.org/doc/stable/', None),
    'matplotlib': ('https://matplotlib.org/stable/', None),
    'spiceypy': ('https://spiceypy.readthedocs.io/en/stable/', None),
}

# -- Options for HTML output -------------------------------------------------

html_theme = 'sphinx_book_theme'
html_title = 'Moon Coverage Documentation'
html_logo = 'images/moon-coverage.svg'
html_favicon = 'images/favicon.ico'

html_theme_options = {
    'use_repository_button': False,
    'use_download_button': False,
    'use_fullscreen_button': False,
    'extra_navbar': '',
}

html_static_path = ['_static']

html_css_files = [
    'css/custom.css',
]

# -- Extension configuration -------------------------------------------------
autodoc_default_options = {
    'members': True,
    'show-inheritance': True,
}

autodoc_member_order = 'bysource'
pygments_style = None

myst_enable_extensions = [
    "colon_fence",
    "dollarmath",
]

jupyter_execute_notebooks = "cache"

copybutton_prompt_text = r'>>>\s?|\.\.\.\s|\$\s'
copybutton_prompt_is_regexp = True
copybutton_copy_empty_lines = True

# -- LaTeX config for PDF output ---------------------------------------------

latex_engine = 'xelatex'

# -- Gallery examples --------------------------------------------------------

sphinx_gallery_conf = {
    'examples_dirs': '../examples',
    'gallery_dirs': 'gallery',
    'filename_pattern': '.py$',
    'download_all_examples': False,
    'show_signature': False,
    'remove_config_comments': True,
    'pypandoc': True,
    'notebook_images': True,
    'first_notebook_cell': (
        '# This example is part of the moon-coverage examples gallery.\n'
        '# More examples can be found here: https://moon-coverage.univ-nantes.fr.'
    ),
    'subsection_order': ExplicitOrder([
        '../examples/trajectory',
        '../examples/cross-missions',
    ]),
}


def setup(app):
    """Fix multiple files warninging.

    See: https://github.com/executablebooks/MyST-NB/issues/363

    """
    app.registry.source_suffix.pop(".ipynb", None)


# -- Download the required kernels in advance ----------------------------------------

kernels_download()
