---
jupytext:
  formats: ipynb,md:myst
  text_representation:
    extension: .md
    format_name: myst
    format_version: 0.13
    jupytext_version: 1.13.0
kernelspec:
  display_name: Python 3
  language: python
  name: python3
---

# ESA events files

```{code-cell}
:tags: [remove-input]

import matplotlib.pyplot as plt

from moon_coverage import TourConfig
from moon_coverage.esa import EsaMissionPhases, EsaMissionTimeline, EsaMissionEvents

tour = TourConfig(mk=5.0, spacecraft='JUICE', target='Ganymede')
```

Recently the ESA-SOC published a collection of event files that contains the different phases of the mission, the dates and names of the flybys; as well as some technical information, e.g. the visibility of the DSN stations and the expected downlinks windows.

:::{Caution}
These files need to be considered as experimental and their format and location can evolve significantly in the future. Currently they are only provided for `the latest CReMA of JUICE mission` (`5.0`) on this [FTP repository](ftp://ftp.sciops.esa.int/pub/juicesoc/DATA/crema_5_0).
You can also download some of them in the [JUICE timeline tool](https://juicesoc.esac.esa.int/tm/).
:::

## Event file format

An event file, is a simple `.csv` file containing either a single time event, a start or stop time event or an event with both a start and a stop time.

For example, the [mission phases file](ftp://ftp.sciops.esa.int/pub/juicesoc/DATA/crema_5_0/Mission_Phases.csv) is defined like this:

:::{Hint}
:class: margin

All the events have a `t_start` and a `t_end` time and the primary key is stored in the `Name` column.
:::

```{code-cell}
:tags: [remove-input, output_scroll]

EsaMissionPhases('Mission_Phases.csv').csv
```

But [in some cases](ftp://ftp.sciops.esa.int/pub/juicesoc/DATA/crema_5_0/mission_timeline_event_file_5_0.csv), the events are only defined as single `event time`:

:::{Note}
:class: margin

The primary key (`Event name`) does not need to be unique and will be grouped later on.

Here, the `XXX_START` and `XXX_END` events will be paired.

If a pair is incomplete, a _warning_ message will be send to the user at runtime.
:::

```{code-cell}
:tags: [remove-input, output_scroll]

EsaMissionTimeline('mission_timeline_event_file_5_0.csv').csv
```

Lastly, [in some cases](ftp://ftp.sciops.esa.int/pub/juicesoc/DATA/crema_5_0/segmentation_proposal_Jupiter_Phase_all_crema_5_0_MALARGUE.csv), no explicit header is provided in the file.
By default we will assume that when it's the case, the default header should be:

:::{Tip}
:class: margin

As you can see below, any custom header can be provided.
:::

```{code-cell}
:tags: [remove-input, output_scroll]

EsaMissionEvents('segmentation_proposal_Jupiter_Phase_all_crema_5_0_MALARGUE.csv').csv
```

## Parsing an events file

All the event files can be parsed by the `EventsFile` object:

```{code-cell}
from moon_coverage.events import EventsFile

EventsFile(
    'segmentation_proposal_Jupiter_Phase_all_crema_5_0_MALARGUE.csv',
    primary_key='name',                                        # default
    header='# name, t_start, t_end, subgroup, working_group',  # optional
)
```

But in practice, it is easier to use its higher level implementations that will supply the default parameter for the files described above:

```{code-cell}
from moon_coverage.esa import EsaMissionPhases, EsaMissionTimeline, EsaMissionEvents

phases = EsaMissionPhases('Mission_Phases.csv')
timeline = EsaMissionTimeline('mission_timeline_event_file_5_0.csv')
malargue = EsaMissionEvents('segmentation_proposal_Jupiter_Phase_all_crema_5_0_MALARGUE.csv')
```

When parsed, you will get a list of `Event` and `EventWindow` objects:

```{code-cell}
:tags: [output_scroll]

phases
```

## Get an event by key

You can select a given event with its name:

:::{Tip}
:class: margin

On Jupyter and IPython environment, you can use the `<tab>`
key to auto-complete: `events['<tab>']` to show the available keys.
:::

```{code-cell}
gco_500 = phases['GCO500']

gco_500
```
When multiple event with the same name are detected, they will be grouped in an `EventsList`:

:::{Tip}
:class: margin

The search key is not sensitive to the case.
:::

```{code-cell}
downlinks = timeline['downlink_malargue']

downlinks
```

:::{Hint}
:class: margin

In the case of the __flybys__, they contain a special property (`.crema_names`) that host they canonic name defined by the SOC:

```python
>>> flybys.crema_names
['1G1', '2G2', ...]
```
:::

```{code-cell}
flybys = timeline['FLYBY_GANYMEDE']

flybys
```

Then, you can select a precise flyby if you want with either its `name` or its `id`:

:::{Tip}
:class: margin

The value in the ``contextual info`` column are also parsed as additional attributes (separated with a ``;``).
:::

```{code-cell}
flyby_1G1 = flybys['1G1']  # or flybys[0]

flyby_1G1
```

Multiple inputs is also supported:

:::{Tip}
:class: margin

The events will be __sorted__ by start time and
the __duplicates__ will be __discarded__.
:::

```{code-cell}
phases['GCO500', 'GCO5000']
```

You can also get all the events starting with a given keyword:

```{code-cell}
timeline.startswith('GANYMEDE')
```

or ending with a given keyword:

```{code-cell}
timeline.endswith('TRANSIT')
```

as well as [complex regular expression](https://docs.python.org/3/library/re.html):

:::{Note}
:class: margin

The `find()` function is the default function called by `__getitem__`
in last resort, if none of the key(s) were found.
:::

```{code-cell}
timeline.find('^VIS.*E$')  # Starting with 'VIS' and ending with a 'E'
```


## Events filtering

`EventsFile`, `EventsDict` and `EventsList` can be filtered with any date string with a `before()`, `after()` and `between()` functions:

:::{Hint}
:class: margin

You can also use the comparison operators ``<``, ``<=``, ``>``, ``>=``.
But you should avoid to chain them (see `Danger` note below).
:::

```{code-cell}
phases.before(2033)  # or: `phases <= 2033`
```

```{code-cell}
phases.after('2034-06-01', strict=True)  # or: `phases > '2034-06-01'`
```

```{code-cell}
timeline.between('2031-06', '2032-03')
```

:::{Danger}
Unfortunately, it is not possible to chain the operators without
explicit parenthesis. You should __NOT__ do:

```python
'2031-06' <= timeline <= '2032-03'
```

but you __should__ do:
```python
('2031-06' <= timeline) <= '2032-03'
```
or:
```python
'2031-06' <= (timeline <= '2032-03')
```
:::


## Events usage

All these events objects are compatible with `TourConfig` and `Trajectory` objects, for example you can use them to select a time window:

```{code-cell}
tour[gco_500:'1h']
```

You can also use an event to defined quickly the associated flyby:

```{code-cell}
tour.flyby(flyby_1G1)
```

You can also get multiple flyby trajectories at once (here the first 3 flybys above Ganymede defined before):

```{code-cell}
tour.flyby(flybys)
```

Temporal operation is also supported on the `Event` and `EventWindow` objects, for example you can take the trajectory `12 hours` around a flyby CA event (with a regular temporal step of `10 minutes`):

:::{Hint}
:class: margin

A _subtraction_ is performed on the `start time` and an _addition_ is performed on the `stop time`.
:::

```{code-cell}
tour[flyby_1G1 - '12 h':flyby_1G1 + '12 h':'10 mins']
```

You can also filter a trajectory with any `EventWindow` or `EventsList`.
Here we choose to take all the tour (from the beginning to the end of the mission with a temporal step of `1h`) and we only show the point during the `GCO500 phase`.

:::{Caution}
:class: margin

Filtering on a large temporal windows is __very inefficient__. Here the full trajectory contains more than 4,500 points when only 130 are really relevant. You should better use the solution presented above:

```python
tour[gco_500:'1 day']
```

to get the same result in the more efficient way.
:::

```{code-cell}
tour['start':'stop':'1 day'] & gco_500
```

## Trajectory filtering with events


Similarly to the `ROI`, you can also discard an `EventWindow` or an `EventsList`. For example if you want to mask all the downlink period, you can do:

```{code-cell}
traj_no_downlinks = tour['2031-01-20' : '2031-01-22'] ^ downlinks

traj_no_downlinks
```

We can represent the filtered trajectory parameter, with the 2 masked downlink windows:

```{code-cell}
:tags: [hide-input]

_, ax = plt.subplots(figsize=(14, 3))

ax.plot(traj_no_downlinks.utc, traj_no_downlinks.inc, 'o', ms=2)

y0, y1 = ax.get_ylim()
for downlink in downlinks[:2]:
    ax.fill_between([downlink.start, downlink.stop], y0, y1, color='0.95')
    ax.text(downlink.start + (downlink.stop - downlink.start)/2, (y0 + y1) / 2,
            'Downlink\nwindow', ha='center', va='center')

ax.set_ylim(y0, y1)
ax.set_xlabel('UTC datetime in 2031')
ax.set_ylabel('Incidence angle (degree)');
```

You can also exclude an `EventWindow` or an `EventsList` directly in `TourConfig` or in the `Trajectory` definition:

:::{Tip}
:class: margin

If you know in advance the events that you want to discard, it is recommended to use this method to avoid creating trajectory points that will not be masked later on. __Only the relevant points are kept__.
:::

```{code-cell}
TourConfig(
    mk='5.0',
    spacecraft='JUICE',
    target='Ganymede',
    exclude=downlinks
)['2031-01-20' : '2031-01-22']
```
