---
jupytext:
  formats: ipynb,md:myst
  text_representation:
    extension: .md
    format_name: myst
    format_version: 0.13
    jupytext_version: 1.13.0
kernelspec:
  display_name: Python 3
  language: python
  name: python3
---

# ESA export files

```{code-cell}
:tags: [remove-input]

import matplotlib.pyplot as plt

from moon_coverage import TourConfig, GanymedeROIs
from moon_coverage.esa import EsaMissionEvents

tour = TourConfig(mk=5.0, spacecraft='JUICE', target='Ganymede')
sc_traj = tour['2035-06-01':'2035-06-02']
```

The `moon-coverage` is designed to compute spacecraft trajectory, instrument coverage and region of interest visibility.
When you perform your analysis, you can export you calculations to external tools.

At the moment, only a few temporal events export are supported but [you want create your own if needed](../trajectory/spacecraft/masked_trajectory.html#segmented-trajectories).
Below you will find different examples to export to the JUICE timeline tool and to MAPPS.

## JUICE timeline export

The [JUICE Segment Harmonization Tool](https://juicesoc.esac.esa.int/tm/?trajectory=CREMA_5_0) (a.k.a the _JUICE timeline tool_) is an online events visualization tool that can be used to display a collection of temporal events in different timelines. It allows the ESA-SOC to plan events computed with other planning tools.

The `moon-coverage` is natively compatible with the JUICE timeline and can export CSV and JSON files. Then these files can be imported in the timeline tool.

```{code-cell}
from moon_coverage.esa import export_timeline
```

### Trajectory segments

If we re-use the masked trajectory computed in the [spacecraft trajectory section](../trajectory/spacecraft/masked_trajectory.html#multi-criteria-filter) below `500 km` and on the `dayside` of Ganymede during a few orbit in June 2035, we can export all the segments in a CSV file:

```{code-cell}
export_timeline(
    'Ganymede_2035-JUN-01-low_alt-dayside.csv',
    sc_traj.where((sc_traj.alt > 500) & sc_traj.day)
)
```
```{code-cell}
:tags: [hide-input]

# You can use `EsaMissionEvents` to read the csv file as any ESA segmentation file

EsaMissionEvents('Ganymede_2035-JUN-01-low_alt-dayside.csv', header=None)['TRAJECTORY_SEGMENT']
```

We can recognized the 8 segments [displayed previously](../trajectory/spacecraft/masked_trajectory.html#masked-trajectory-on-a-map).

![Trajectory segments imported in the JUICE Segment Harmonization tool.](Timeline-traj_segments.png)

### ROIs intersection

You can provide a second argument to the `export_timeline()` function to export only the intersection between a `Trajectory` and a `ROI` or a `ROIsCollection`.

Here, we will export the file as a JSON file just by changing the output filename extension (`.json`):

```{code-cell}
export_timeline(
    'Ganymede_2035-JUN-01-ROI-polar_deposits.json',
    sc_traj,
    GanymedeROIs[1],
)
```
```{code-cell}
:tags: [remove-input, output_scroll]

import json

with open('Ganymede_2035-JUN-01-ROI-polar_deposits.json', 'r') as f:
    print(json.dumps(json.load(f), indent=2))
```


We also restricted our intersection the intersection only on the [polar deposits ROI from Stephan et al. 2021](../trajectory/rois/intersection.html#representation-on-a-map).

![Polar deposit ROI intersected imported in the JUICE Segment Harmonization tool.](Timeline-rois_intersection.png)

More detail about the optional arguments on `export_timeline()` can be found [here](../api/esa.html#export-timeline).


## MAPPS export

Similarly to the JUICE timeline export described above, you can export your `Trajectory` segments and `ROI` intersections as EVF file that can be imported in MAPPS:

```{code-cell}
export_timeline(
    'Ganymede_2035-JUN-01-low_alt-dayside-ROI-polar_deposits.evf',
    sc_traj.where((sc_traj.alt > 500) & sc_traj.day),
    GanymedeROIs[1],
)
```
```{code-cell}
:tags: [remove-input]

!cat Ganymede_2035-JUN-01-low_alt-dayside-ROI-polar_deposits.evf
```
