"""Pre-build download for ReadTheDocs."""

import os
from datetime import datetime as dt
from multiprocessing import cpu_count
from pathlib import Path

from moon_coverage import TourConfig
from moon_coverage.debug import debug_download
from moon_coverage.esa import EsaCremasCollection
from moon_coverage.misc import logger


def kernels_download():
    """Download ESA kernels for the examples.

    By default we use the FTP ESA endpoint which is at least 20% faster
    than the HTTPS endpoint.

    """
    # Add RTD logger
    log, _ = logger('Docs')
    log.warning('CPU available: %i', cpu_count())

    # Select ESA download endpoint
    endpoint = os.environ.get(
        'ESA_KERNELS_ENDPOINT',
        'ftp://spiftp.esac.esa.int/data/SPICE'
    )

    # Enable download debugger
    if os.environ.get('DEBUG_DOWNLOAD'):
        debug_download(True)

    # Start download
    t_start = dt.now()
    log.warning('Start kernels download at: %s', t_start)

    # JUICE
    if not os.environ.get('KERNELS_JUICE'):
        raise KeyError('`KERNELS_JUICE` not found.')

    _ = TourConfig(
        mk=5.0,
        spacecraft='JUICE',
        remote_kernels=f'{endpoint}/JUICE/kernels',
        download_kernels=True,
    )

    t_juice = dt.now()
    log.warning('JUICE download duration: %s', t_juice - t_start)

    # Examples kernel directory output
    KERNELS_DIR = (Path(__file__).parent / '../examples/cross-missions/kernels').resolve()

    # Bepi
    mk = EsaCremasCollection('BepiColombo')['bc_plan.tm', 'v200_20181212_002']
    _ = TourConfig(
        mk=mk,
        spacecraft='MPO',
        kernels_dir=KERNELS_DIR,
        remote_kernels=f'{endpoint}/BEPICOLOMBO/kernels',
        download_kernels=True,
    )

    t_bepi = dt.now()
    log.warning('Bepi download duration: %s', t_bepi - t_juice)

    # EnVision
    mk = EsaCremasCollection('EnVision')['envision_study_north_voi_v020.tm']
    _ = TourConfig(
        mk=mk,
        spacecraft='EnVision',
        kernels_dir=KERNELS_DIR,
        remote_kernels=f'{endpoint}/EnVision/kernels',
        download_kernels=True,
    )

    t_envision = dt.now()
    log.warning('EnVision download duration: %s', t_envision - t_bepi)

    # Download complete
    t_end = dt.now()
    log.warning('Complete kernels download time: %s', t_end - t_start)
