---
jupytext:
  formats: ipynb,md:myst
  text_representation:
    extension: .md
    format_name: myst
    format_version: 0.13
    jupytext_version: 1.13.0
kernelspec:
  display_name: Python 3
  language: python
  name: python3
---

# SPICE Pool

The SPICE calculations are performed on a pre-loaded set of kernels, usually with the `spiceypy.furnsh()` function. These kernels are cached in the SPICE pool and accessible to all the SPICE functions.

To easily manipulate this pool, to add or remove kernels and to quickly check its content, we created a python `SpicePool` object that acts as a `singleton`.

```{code-cell}
from moon_coverage import SpicePool
```

The `SpicePool` contains the list of all the pre-loaded kernels:

```{code-cell}
SpicePool
```

## Add and remove kernels to the pool

You can add one or many kernel(s) with:

```{admonition} Download
:class: margin

[`naif0012.tls`](https://naif.jpl.nasa.gov/pub/naif/generic_kernels/lsk/naif0012.tls) kernel
```

```{code-cell}
SpicePool.add('naif0012.tls')

SpicePool
```


```{code-cell}
SpicePool.kernels
```

```{code-cell}
SpicePool.remove('naif0012.tls')

SpicePool
```

You can check if a kernel is already in the pool:

```{code-cell}
'naif0012.tls' in SpicePool
```

And you can also check the full content of the pool against a list of expected kernels:

```{code-cell}
SpicePool == ['naif0012.tls']
```

:::{Note}
:class: margin

If the `kernel` is already present in the pool, it will throw a `ValueError` exception but a kernel can only be loaded once.
:::

You can also use logical operations to add/remove a kernel(s) to/from the pool:

```{code-cell}
SpicePool + 'naif0012.tls'
SpicePool - 'naif0012.tls'
```

If needed, you can purge the whole pool at once:

```{code-cell}
SpicePool.purge()

SpicePool
```

## Get values from pool

```{admonition} Download
:class: margin

[`pck00010.tpc`](https://naif.jpl.nasa.gov/pub/naif/generic_kernels/pck/pck00010.tpc) kernel
```

```{code-cell}
SpicePool.add(['naif0012.tls', 'pck00010.tpc'])
```

When your set of kernels is loaded, you can request any values in the pool with the python getter syntax:

```{code-cell}
SpicePool['DELTET/DELTA_T_A']
```
```{code-cell}
SpicePool['BODY503_RADII']
```

If the value is not present in the pool a `ValueError` will be thrown:

```{code-cell}
:tags: [raises-exception]
SpicePool['FOO']
```

## Kernels temporal coverage

```{code-cell}
:tags: [remove-cell]

# Load all the kernels from JUICE 5.0 CReMA

from moon_coverage import TourConfig

TourConfig(mk=5.0)
```

Sometime it can be difficult to get the complete range of valid times that can be use with the kernels loaded in the pool. The `SpicePool` provide a helper that will search all `CK`, `PCK` and `SPK` kernels to find the coverage window associated with a given body:

```{code-cell}
SpicePool.windows('GANYMEDE')
```

You can select different types of output format, with the `fmt` argument:

```{code-cell}
SpicePool.windows('GANYMEDE', fmt='ET')
```

You can also provide a list of bodies to get all their coverage windows:

```{code-cell}
SpicePool.windows('JUICE', 'GANYMEDE')
```

If you need the overlapping coverage (i.e. max start time and min stop time), you can use the `.coverage()` function:

```{code-cell}
SpicePool.coverage('JUICE', 'GANYMEDE', fmt='TDB')
```

## Enable pool debugger

If you need to debug the SPICE pool, you can increase the level of the verbose with the `debug_spice_pool()` function.

```{code-cell}
from moon_coverage.debug import debug_spice_pool

debug_spice_pool(True)
```

:::{Tip}
:class: margin

If you want purge the pool before loading a new set of kernels, you can add `purge=True` argument.
:::


```{code-cell}
SpicePool.add('naif0012.tls', purge=True)
```

```{code-cell}
SpicePool.remove('naif0012.tls')
```

```{code-cell}
SpicePool.purge()
```

You can decrease the verbose level to `INFO` with an `'info'` key:

```{code-cell}
debug_spice_pool('info')
```

Now only, the purge of the pool will trigger a log output:

```{code-cell}
SpicePool.add('naif0012.tls')
```

```{code-cell}
SpicePool.remove('naif0012.tls')
```

```{code-cell}
SpicePool.purge()
```

To disable the SPICE pool verbose, pass a `False` key:

```{code-cell}
debug_spice_pool(False)
```
