---
jupytext:
  formats: ipynb,md:myst
  text_representation:
    extension: .md
    format_name: myst
    format_version: 0.13
    jupytext_version: 1.13.0
kernelspec:
  display_name: Python 3
  language: python
  name: python3
---

# SPICE References

```{code-cell}
:tags: [remove-cell]

from moon_coverage import TourConfig

TourConfig(mk=5.0)
```

Object references in the SPICE kernels are usually represented either by their name (which is not always unique) or by an internal `id` code. Most of the time, the user does not have to know the reference `id` and can pass directly its name to the function but it's not always the case. To provide a simpler access to the SPICE references, the `moon-coverage` provides a `SpiceRef` object that will extract the correct information from the [`SpicePool`](pool).


```{code-cell}
from moon_coverage import SpiceRef
```

For example, for `Jupiter barycenter` we have:

:::{Note}
:class: margin

If the reference is not known, a `spiceypy` error will be thrown with an additional `ValueError`:

```python
>>> SpiceRef('FOO')
...
NotFoundError: Spice returns not found for function: bods2c
...
ValueError: Unknown reference: `FOO`
```
:::

```{code-cell}
ref = SpiceRef('Jupiter Barycenter')
ref
```

## Generic `SpiceRef` properties

This reference can be used either as a string to get its `name` property:

```{code-cell}
str(ref)
```

or as an integer to get its `id` property:

```{code-cell}
int(ref)
```

You can compare a `SpiceRef` with its string and integer values:

```{code-cell}
ref == 'JUPITER BARYCENTER'
```

```{code-cell}
ref == 5
```

Similarly to the `SpicePool`, you can query the values from the pool directly on the `SpiceRef` object:

:::{Warning}
:class: margin

By default, the value is not specific to the `SpiceRef`, any value in the `SpicePool` is accessible.
:::
```{code-cell}
ref['BODY599_RADII']
```

## `SpiceBody` properties

If the `SpiceRef` is a planet or a satellite (like for `Ganymede`), the `SpiceRef` is promoted to a `SpiceBody`.

```{code-cell}
body = SpiceRef('Ganymede')
body
```

For a `SpiceBody`, you have access to additional attributes. For example, the body `radii` values:

```{code-cell}
body.radii  # [km]
```

its _geometric-mean_ radius:

```{code-cell}
body.radius  # [km]
```

or its centered body-fixed frame:

```{code-cell}
body.frame
```

For a `SpiceBody`, the item getter is prefix with `BODYXXX_` with `XXX` being the body code. Therefore you can do:

```{code-cell}
body['GM']  # Alias to `BODY503_GM`
```

## `SpiceSpacecraft` properties

If you select a spacecraft reference, it will be promoted to a `SpiceSpacecraft`:

```{code-cell}
sc = SpiceRef('JUICE')
sc
```

`SpiceSpacecraft` includes a `frame` and `boresight` attributes:

```{code-cell}
sc.frame
```
```{code-cell}
sc.boresight
```

You can also access the list of the spacecraft instruments available in the SPICE pool:

```{code-cell}
:tags: [output_scroll]

sc.instruments
```

or select a single one:

```{code-cell}
inst = sc.instr('JANUS')
inst
```

## `SpiceInstrument` properties

When an instrument is selected (as a `SpiceInstrument`), you can get its field of view parameters:

```{code-cell}
inst.boresight
```
```{code-cell}
inst.bounds
```
```{code-cell}
inst.shape
```
```{code-cell}
inst.frame
```

You can also retrieve the detector size:
```{code-cell}
inst.ns, inst.nl
```

or the full field-of-view along the track and cross track:
```{code-cell}
inst.fov_along_track, inst.fov_cross_track  # in radians
```

:::{Caution}
:class: margin
We assume that the sample direction is
aligned with the cross-track direction (ie. 1-line acquisition
in push-broom mode should be in the direction of flight).

This convention is not strictly enforced in the instrument kernel
and **the user must be careful** and adjust theses values according
to the detector orientation relative to the spacecraft motion.

:::

:::{Warning}
:class: margin
`JUICE_JANUS` instrument in `v06` does not follow this convention.
We manually manage this exception for the moment (see
[MR !27](https://juigitlab.esac.esa.int/datalab/moon-coverage/-/merge_requests/27)
for more details).
:::

as well as the pixel instantaneous field-of-view:
```{code-cell}
inst.ifov_along_track, inst.ifov_cross_track  # in radians
```

For a `SpiceInstrument`, the item getter is prefix with `INSXXX_` with `XXX` being the instrument code. Therefore you can do:

```{code-cell}
inst['PIXEL_SAMPLES']  # Alias to `INS-28200_PIXEL_SAMPLES`
```

You can display the `boresight`, the `bounds` vectors and `rays` of the field of view contour in the instrument frame:

```{code-cell}
inst.display();
```

Or you can represent it in the plane orthogonal to the field of view:

```{code-cell}
inst.view();
```
