"""Test ESA API module."""

from io import BytesIO
from json import JSONDecodeError
from urllib.error import HTTPError

from pytest import fixture, raises

import moon_coverage.esa.api as api_request
from moon_coverage.esa import get_mk, get_tag
from moon_coverage.esa.api import crema_ref, esa_api


LATEST_JUICE_SKD_TAG = 'v300_20220310_001'
LATEST_JUICE_MK = 'juice_crema_5_0b23_1.tm'


@fixture
def mock_url_resp(monkeypatch):
    """Mock URL response."""

    def url_response(data):
        """Dummy url response data/err."""
        resp = iter(data if isinstance(data, list) else [data])

        class DummyResponse:
            """Dummy data response."""

            def __init__(self, url):
                self.url = url

            def __enter__(self):
                """Dummy reader function."""
                return BytesIO(next(resp).encode())

            def __exit__(self, *args):  # pylint: disable=no-self-use
                pass

        def dummy_error(url):
            """Dummy error response."""
            raise HTTPError(url, 400, 'Bad request', {}, BytesIO())

        monkeypatch.setattr(api_request, 'urlopen',
                            DummyResponse if data else dummy_error)

    return url_response


def test_esa_api_cache(mock_url_resp):
    """Test ESA API cache."""
    # The first response to cache
    mock_url_resp('{"values":true}')
    assert esa_api('foo')

    # Change response on a new key
    mock_url_resp('{"children":{"values":false}}')
    assert not esa_api('bar')

    # Reload from cache (even if the response evolved)
    assert esa_api('foo')

    # Response splitted in 3 payloads
    mock_url_resp([
        '{"values":[0, 1],"nextPageStart":2,"start":0}',
        '{"values":[2, 3],"nextPageStart":4,"start":2}',
        '{"values":[4],"start":4}'
    ])
    assert esa_api('baz') == [0, 1, 2, 3, 4]

    # Response splitted in 2 payload but with a limit
    mock_url_resp([
        '{"values":[0, 1],"nextPageStart":2}',
        '{"values":[2]}'
    ])
    assert esa_api('qux', limit=2) == [0, 1]


def test_esa_api_err(mock_url_resp):
    """Test ESA API error."""
    # Bad request (400 error)
    mock_url_resp(None)
    with raises(IOError):
        _ = esa_api('400')

    # The response is not a JSON
    mock_url_resp('<html><body><p>Not a JSON</p></body></html>.')
    with raises(JSONDecodeError):
        _ = esa_api('html')

    # JSON without values keyword
    mock_url_resp('{"size":1,"limit":25,"isLastPage":true}')
    with raises(KeyError):
        _ = esa_api('missing-values')

    # Valid JSON without values
    mock_url_resp('{"size":1,"limit":25,"isLastPage":true,"values":[]}')
    with raises(ValueError):
        _ = get_tag('empty-values')


def test_esa_api_get_tag():
    """Test get tags from ESA API."""
    # Latest
    assert get_tag('juice') == LATEST_JUICE_SKD_TAG
    assert get_tag('juice', 'latest') == LATEST_JUICE_SKD_TAG

    # Get an older version from short version key
    assert get_tag('juice', 'v280') == 'v280_20210216_001'

    # Get an older version from long version key
    assert get_tag('juice', 'v280_20210216_001') == 'v280_20210216_001'

    # Try to get an invalid version
    with raises(ValueError):
        _ = get_tag('juice', 'v999')

    # Get all tags
    tags = get_tag('juice', 'all', limit=2)

    assert len(tags) == 2
    assert tags[0] == LATEST_JUICE_SKD_TAG

    # Invalid mission name
    with raises(AttributeError):
        _ = get_tag(None)


def test_esa_api_get_mk():
    """Test get mk form ESA API."""
    # Latest metakernel for the latest version
    mk = get_mk('JUICE')
    assert mk.name == LATEST_JUICE_MK
    assert mk.parent.name == LATEST_JUICE_SKD_TAG
    assert mk.parent.parent.name == 'juice'

    # Get an old metakernel for the latest version (with filename)
    mk = get_mk('juice', 'juice_crema_3_0.tm')
    assert mk.name == 'juice_crema_3_0.tm'

    # Get an old metakernel for the latest version (with shortcut)
    mk = get_mk('juice', '3.0')
    assert mk.name == 'juice_crema_3_0.tm'

    # Get an old metakernel from an old version (with full tag)
    mk = get_mk('juice', '3.0', 'v280_20210216_001')
    assert mk.name == 'juice_crema_3_0.tm'
    assert mk.parent.name == 'v280_20210216_001'

    # Get an old metakernel from an old version (with short version)
    mk = get_mk('juice', '3.0', 'v280')
    assert mk.name == 'juice_crema_3_0.tm'
    assert mk.parent.name == 'v280_20210216_001'

    # Get all the metakernels for a given tag
    mks = get_mk('juice', mk='all', version='v280_20210216_001')

    assert len(mks) == 10
    assert mks[0] == 'juice_crema_3_0.tm'
    assert mks[-1] == 'juice_crema_4_2b22_1.tm'

    # Get all the metakernels for all tag
    mks = get_mk('juice', mk='all', version='all')

    assert LATEST_JUICE_SKD_TAG in mks  # latest
    assert 'v160_20180216_001' in mks   # oldest

    assert len(mks['v160_20180216_001']) == 5
    assert mks['v160_20180216_001'][0] == 'juice_crema_2_0_ops.tm'
    assert mks['v160_20180216_001'][-1] == 'juice_crema_3_2_ops.tm'

    # Invalid version in MkDict
    assert 'foo' not in mks

    with raises(KeyError):
        _ = mks['foo']

    # Remove and re-downloaded the latest JUICE crema
    mk = get_mk('juice')
    mk.unlink()

    mk = get_mk('juice')
    assert mk.exists()

    # Invalid mission name
    with raises(AttributeError):
        _ = get_mk(None)

    # Invalid metakernel name
    with raises(FileNotFoundError):
        _ = get_mk('juice', '1.0')


def test_esa_api_crema_ref():
    """Test ESA CReMA reference formatter."""
    assert crema_ref(3.0) == '3_0'
    assert crema_ref('3.0') == '3_0'
    assert crema_ref('4.0b') == '4_0b'
    assert crema_ref('3_2 gan N56 Pp5 Q19') == '3_2_gan_N56_Pp5_Q19'
    assert crema_ref('3.2 gan N56 Pp5 Q19') == '3_2_gan_N56_Pp5_Q19'
    assert crema_ref('3.2-gan-N56-Pp5-Q19') == '3_2_gan_N56_Pp5_Q19'
    assert crema_ref('4.2 N56 Pp5 Q19') == '4_2_N56_Pp5_Q19'
    assert crema_ref('4.2 N56 Pn5 Q19') == '4_2_N56_Pn5_Q19'
    assert crema_ref('4.2 N54 Pp5 Q19') == '4_2_N54_Pp5_Q19'
    assert crema_ref('4.2b22.1') == '4_2b22_1'
    assert crema_ref('4_2b22_1') == '4_2b22_1'
    assert crema_ref('4.2b22.1 (cruise only)') == '4_2b22_1_cruise'
    assert crema_ref(5.0) == '5_0'
    assert crema_ref('5.0b23.1') == '5_0b23_1'
