"""Test ESA export functions."""

from pathlib import Path

from pytest import fixture, raises

from moon_coverage.esa import export_timeline
from moon_coverage.esa.export import datetime, extract_segments
from moon_coverage.rois import ROI, ROIsCollection


DATA = Path(__file__).parent / '..' / 'data'


@fixture
def traj():
    """Default trajectory with 2 segments."""

    class DummySegment:
        """Dummy segment."""
        def __init__(self, start, stop):
            self.start = start
            self.stop = stop

    class DummyTrajectory:
        """Dummy trajectory."""
        segments = [
            DummySegment('2031-01-01', '2031-01-01T12'),
            DummySegment('2032-01-01T01:23:45', '2032-01-01T12:34:56.7890'),
            DummySegment('2031-01-02', '2031-01-02T12'),
            DummySegment('2032-01-01T01:23:45', '2032-01-01T12:34:56'),
        ]

        def __and__(self, key):
            return self.segments[slice(2) if key == 'ROI_1' else slice(2, 4)]

        def __iter__(self):
            return iter(self.segments[:3])

    return DummyTrajectory()


@fixture
def rois(monkeypatch):
    """Default collection of ROIs with 2 ROIs."""

    monkeypatch.setattr(ROI, '__contains__', lambda self, _: True)

    return ROIsCollection({
        'ROI_1': {'lons_e': [1, 3, 3, 1], 'lats': [5, 5, 7, 7], 'name': 'foo'},
        'ROI_2': {'lons_e': [2, 4, 4, 2], 'lats': [6, 6, 8, 8], 'category': 'bar'},
    })


@fixture
def freeze_now(monkeypatch):
    """Freeze datetime now function."""
    class MockDatetime(datetime.datetime):
        """Mocked datetime."""
        @classmethod
        def now(cls, *args):
            """Patched now."""
            return datetime.datetime(2022, 2, 25, 18)

    monkeypatch.setattr(datetime, 'datetime', MockDatetime)

    return datetime.datetime.now()


def test_esa_export_extract_segments(traj, rois):
    """Test trajectory and roi segments extraction."""
    # Without a ROI
    segments = extract_segments(traj, subgroup='DUMMY_SUBGROUP', source='LOCAL')

    assert segments == [
        [
            'TRAJECTORY_SEGMENT',
            '2031-01-01T00:00:00Z',
            '2031-01-01T12:00:00Z',
            'DUMMY_SUBGROUP',
            'LOCAL',
        ], [
            'TRAJECTORY_SEGMENT',
            '2031-01-02T00:00:00Z',
            '2031-01-02T12:00:00Z',
            'DUMMY_SUBGROUP',
            'LOCAL',
        ], [
            'TRAJECTORY_SEGMENT',
            '2032-01-01T01:23:45Z',
            '2032-01-01T12:34:56.789Z',
            'DUMMY_SUBGROUP',
            'LOCAL',
        ],
    ]

    # With a single ROI
    segments = extract_segments(traj, rois['ROI_1'])

    assert segments == [
        [
            'ROI_INTERSECTION',
            '2031-01-01T00:00:00Z',
            '2031-01-01T12:00:00Z',
            'ROI_1',
            'GENERIC',
        ], [
            'ROI_INTERSECTION',
            '2032-01-01T01:23:45Z',
            '2032-01-01T12:34:56.789Z',
            'ROI_1',
            'GENERIC',
        ],
    ]

    # With a ROIsCollection
    segments = extract_segments(traj, rois)

    assert segments == [
        [
            'ROI_INTERSECTION',
            '2031-01-01T00:00:00Z',
            '2031-01-01T12:00:00Z',
            'ROI_1',
            'GENERIC',
        ], [
            'ROI_INTERSECTION',
            '2031-01-02T00:00:00Z',
            '2031-01-02T12:00:00Z',
            'ROI_2',
            'GENERIC',
        ], [
            'ROI_INTERSECTION',
            '2032-01-01T01:23:45Z',
            '2032-01-01T12:34:56Z',
            'ROI_2',
            'GENERIC',
        ], [
            'ROI_INTERSECTION',
            '2032-01-01T01:23:45Z',
            '2032-01-01T12:34:56.789Z',
            'ROI_1',
            'GENERIC',
        ],
    ]


def test_esa_export_export_timeline_csv(traj, rois):
    """Test JUICE timeline export as a csv file."""
    # Trajectory segments (no subgroup)
    csv = export_timeline(DATA / 'export_timeline_export.csv', traj)

    assert csv.read_text(encoding='utf-8') == \
        """# name,t_start,t_end,subgroup,source
TRAJECTORY_SEGMENT,2031-01-01T00:00:00Z,2031-01-01T12:00:00Z,,GENERIC
TRAJECTORY_SEGMENT,2031-01-02T00:00:00Z,2031-01-02T12:00:00Z,,GENERIC
TRAJECTORY_SEGMENT,2032-01-01T01:23:45Z,2032-01-01T12:34:56.789Z,,GENERIC"""

    # Trajectory and ROIs intersection
    csv = export_timeline(DATA / 'export_timeline_export.csv', traj, rois)

    assert csv.read_text(encoding='utf-8') == \
        """# name,t_start,t_end,subgroup,source
ROI_INTERSECTION,2031-01-01T00:00:00Z,2031-01-01T12:00:00Z,ROI_1,GENERIC
ROI_INTERSECTION,2031-01-02T00:00:00Z,2031-01-02T12:00:00Z,ROI_2,GENERIC
ROI_INTERSECTION,2032-01-01T01:23:45Z,2032-01-01T12:34:56Z,ROI_2,GENERIC
ROI_INTERSECTION,2032-01-01T01:23:45Z,2032-01-01T12:34:56.789Z,ROI_1,GENERIC"""

    csv.unlink()
    assert not csv.exists()


def test_esa_export_export_timeline_json(traj, rois, freeze_now):
    """Test JUICE timeline export as a json file."""
    assert str(freeze_now) == '2022-02-25 18:00:00'  # Check freeze fixture

    # Trajectory segments (no subgroup)
    json = export_timeline(DATA / 'export_timeline_export.json', traj)

    assert json.read_text(encoding='utf-8') == (
        '{"creationDate": "2022-02-25T18:00:00Z", '
        '"name": "export_timeline_export", '
        '"segments": [{"start": "2031-01-01T00:00:00Z", '
        '"end": "2031-01-01T12:00:00Z", '
        '"segment_definition": "TRAJECTORY_SEGMENT", '
        '"name": "TRAJECTORY_SEGMENT", '
        '"overwritten": false, "timeline": "LOCAL", '
        '"source": "GENERIC", '
        '"resources": []}, {"start": "2031-01-02T00:00:00Z", '
        '"end": "2031-01-02T12:00:00Z", '
        '"segment_definition": "TRAJECTORY_SEGMENT", '
        '"name": "TRAJECTORY_SEGMENT", '
        '"overwritten": false, "timeline": "LOCAL", '
        '"source": "GENERIC", '
        '"resources": []}, {"start": "2032-01-01T01:23:45Z", '
        '"end": "2032-01-01T12:34:56.789Z", '
        '"segment_definition": "TRAJECTORY_SEGMENT", '
        '"name": "TRAJECTORY_SEGMENT", '
        '"overwritten": false, "timeline": "LOCAL", '
        '"source": "GENERIC", '
        '"resources": []}], "segmentGroups": [], "trajectory": "CREMA_5_0", '
        '"localStoragePk": ""}'
    )

    # Trajectory segments (custom subgroup)
    json = export_timeline(DATA / 'export_timeline_export.json', traj, subgroup='FOO')

    assert json.read_text(encoding='utf-8') == (
        '{"creationDate": "2022-02-25T18:00:00Z", '
        '"name": "export_timeline_export", '
        '"segments": [{"start": "2031-01-01T00:00:00Z", '
        '"end": "2031-01-01T12:00:00Z", '
        '"segment_definition": "TRAJECTORY_SEGMENT", '
        '"name": "FOO", '
        '"overwritten": false, "timeline": "LOCAL", '
        '"source": "GENERIC", '
        '"resources": []}, {"start": "2031-01-02T00:00:00Z", '
        '"end": "2031-01-02T12:00:00Z", '
        '"segment_definition": "TRAJECTORY_SEGMENT", '
        '"name": "FOO", '
        '"overwritten": false, "timeline": "LOCAL", '
        '"source": "GENERIC", '
        '"resources": []}, {"start": "2032-01-01T01:23:45Z", '
        '"end": "2032-01-01T12:34:56.789Z", '
        '"segment_definition": "TRAJECTORY_SEGMENT", '
        '"name": "FOO", '
        '"overwritten": false, "timeline": "LOCAL", '
        '"source": "GENERIC", '
        '"resources": []}], "segmentGroups": [], "trajectory": "CREMA_5_0", '
        '"localStoragePk": ""}'
    )

    # Trajectory and ROIs intersection
    json = export_timeline(DATA / 'export_timeline_export.json', traj, rois)

    assert json.read_text(encoding='utf-8') == (
        '{"creationDate": "2022-02-25T18:00:00Z", '
        '"name": "export_timeline_export", '
        '"segments": [{"start": "2031-01-01T00:00:00Z", '
        '"end": "2031-01-01T12:00:00Z", '
        '"segment_definition": "ROI_INTERSECTION", '
        '"name": "ROI_1", '
        '"overwritten": false, "timeline": "LOCAL", '
        '"source": "GENERIC", '
        '"resources": []}, {"start": "2031-01-02T00:00:00Z", '
        '"end": "2031-01-02T12:00:00Z", '
        '"segment_definition": "ROI_INTERSECTION", '
        '"name": "ROI_2", '
        '"overwritten": false, "timeline": "LOCAL", '
        '"source": "GENERIC", '
        '"resources": []}, {"start": "2032-01-01T01:23:45Z", '
        '"end": "2032-01-01T12:34:56Z", '
        '"segment_definition": "ROI_INTERSECTION", '
        '"name": "ROI_2", '
        '"overwritten": false, "timeline": "LOCAL", '
        '"source": "GENERIC", '
        '"resources": []}, {"start": "2032-01-01T01:23:45Z", '
        '"end": "2032-01-01T12:34:56.789Z", '
        '"segment_definition": "ROI_INTERSECTION", '
        '"name": "ROI_1", '
        '"overwritten": false, "timeline": "LOCAL", '
        '"source": "GENERIC", '
        '"resources": []}], "segmentGroups": [], "trajectory": "CREMA_5_0", '
        '"localStoragePk": ""}'
    )

    json.unlink()
    assert not json.exists()


def test_esa_export_export_timeline_evf(traj, rois, freeze_now):
    """Test JUICE timeline export as a evf file."""
    assert str(freeze_now) == '2022-02-25 18:00:00'  # Check freeze fixture

    # Trajectory segments (no subgroup)
    evf = export_timeline(DATA / 'export_timeline_export.evf', traj)

    assert evf.read_text(encoding='utf-8') == \
        """# Events generated by the moon-coverage on 2022-FEB-25_18:00:00
2031-JAN-01_00:00:00      TRAJECTORY_SEGMENT_START  (COUNT = 1)
2031-JAN-01_12:00:00      TRAJECTORY_SEGMENT_END    (COUNT = 1)
2031-JAN-02_00:00:00      TRAJECTORY_SEGMENT_START  (COUNT = 2)
2031-JAN-02_12:00:00      TRAJECTORY_SEGMENT_END    (COUNT = 2)
2032-JAN-01_01:23:45      TRAJECTORY_SEGMENT_START  (COUNT = 3)
2032-JAN-01_12:34:56.789  TRAJECTORY_SEGMENT_END    (COUNT = 3)"""

    # Trajectory segments (custom subgroup)
    evf = export_timeline(DATA / 'export_timeline_export.evf', traj, subgroup='BAR')

    assert evf.read_text(encoding='utf-8') == \
        """# Events generated by the moon-coverage on 2022-FEB-25_18:00:00
2031-JAN-01_00:00:00      BAR_START  (COUNT = 1)
2031-JAN-01_12:00:00      BAR_END    (COUNT = 1)
2031-JAN-02_00:00:00      BAR_START  (COUNT = 2)
2031-JAN-02_12:00:00      BAR_END    (COUNT = 2)
2032-JAN-01_01:23:45      BAR_START  (COUNT = 3)
2032-JAN-01_12:34:56.789  BAR_END    (COUNT = 3)"""

    # Trajectory and ROIs intersection
    evf = export_timeline(DATA / 'export_timeline_export.evf', traj, rois)

    assert evf.read_text(encoding='utf-8') == \
        """# Events generated by the moon-coverage on 2022-FEB-25_18:00:00
2031-JAN-01_00:00:00      ROI_1_START  (COUNT = 1)
2031-JAN-01_12:00:00      ROI_1_END    (COUNT = 1)
2031-JAN-02_00:00:00      ROI_2_START  (COUNT = 1)
2031-JAN-02_12:00:00      ROI_2_END    (COUNT = 1)
2032-JAN-01_01:23:45      ROI_2_START  (COUNT = 2)
2032-JAN-01_01:23:45      ROI_1_START  (COUNT = 2)
2032-JAN-01_12:34:56      ROI_2_END    (COUNT = 2)
2032-JAN-01_12:34:56.789  ROI_1_END    (COUNT = 2)"""

    evf.unlink()
    assert not evf.exists()


def test_esa_export_errors(traj, rois):
    """Test esa export errors."""
    # Invalid rois input type
    with raises(TypeError):
        _ = extract_segments(traj, roi=False)

    # Invalid output format
    with raises(ValueError):
        _ = export_timeline(DATA / 'export_timeline_export.txt', traj, rois)
