"""Test events module."""

import numpy as np
from numpy.testing import assert_array_almost_equal as assert_array

from pytest import fixture, raises

from moon_coverage.events import (
    Event, EventsDict, EventsList, EventWindow, timedelta
)
from moon_coverage.events.event import AbstractEvent, AbstractEventsCollection
from moon_coverage.html import display_html


@fixture
def event_dict():
    """Single event attributes."""
    return {
        'event time [utc]': '2031-07-21T07:14:23Z',
        'contextual info': 'Crema name= 1G1;Altitude above ganymede [km]= 400.00697',
    }


@fixture
def event(event_dict):
    """Single event."""
    return Event('FLYBY_GANYMEDE', **event_dict)


@fixture
def event_window():
    """Windows event."""
    return EventWindow('GCO500',
                       t_start='2035-05-22T06:00:00',
                       t_end='2035-10-05T16:01:00')


@fixture
def events_dict(event, event_window):
    """Events dict."""
    return EventsDict([event_window, event])


@fixture
def events_list(event):
    """Events list."""
    return EventsList([event, event])


def test_event_single(event, event_dict):
    """Test single event attributes."""
    assert isinstance(event, Event)
    assert str(event) == 'FLYBY_GANYMEDE (2031-07-21)'

    assert event.start == event.stop == np.datetime64('2031-07-21T07:14:23')
    assert event.start_date == event.stop_date == '2031-07-21'

    assert isinstance(hash(event), int)

    # Contextual info are extracted with the data (but not parsed)
    assert tuple(event.keys()) == (
        'event time [utc]', 'Crema name', 'Altitude above ganymede [km]')

    assert tuple(event.values()) == ('2031-07-21T07:14:23Z', '1G1', '400.00697')

    assert 'Crema name' in event
    assert event['Crema name'] == '1G1'
    assert event.get('Crema name') == '1G1'

    assert 'FOO' not in event
    assert event.get('FOO') is None

    # Single event init with an EventWindow
    event = EventWindow('FLYBY_GANYMEDE', **event_dict)
    assert isinstance(event, Event)
    assert isinstance(hash(event), int)


def test_event_window(event_window):
    """Test event window attributes."""
    assert isinstance(event_window, EventWindow)
    assert str(event_window) == 'GCO500 (2035-05-22 -> 2035-10-05)'

    assert event_window.start == np.datetime64('2035-05-22T06:00:00')
    assert event_window.stop == np.datetime64('2035-10-05T16:01:00')
    assert event_window.start_date == '2035-05-22'
    assert event_window.stop_date == '2035-10-05'

    assert isinstance(hash(event_window), int)


def test_events_dict(event, event_window, events_dict):
    """Test events dictionary."""
    events = events_dict

    assert isinstance(events, EventsDict)
    assert len(events) == 2

    # Sorted by inscreasing time
    assert str(events) == '\n'.join([
        '(2031-07-21 -> 2035-10-05 | 2 keys):',
        ' - FLYBY_GANYMEDE (2031-07-21)',
        ' - GCO500 (2035-05-22 -> 2035-10-05)',
    ])

    # Getter
    assert events['FLYBY_GANYMEDE'] == event
    assert events['GCO500'] == event_window

    assert events[0] == event
    assert events[1] == event_window
    assert events[-1] == event_window
    assert events[-2] == event

    with raises(IndexError):  # out of range index
        _ = events[2]

    with raises(IndexError):  # out of range index
        _ = events[-3]

    items = events[:]
    assert isinstance(items, EventsDict)
    assert len(items) == 2

    items = events['GCO', 'FLYBY_', 'FLYBY_']  # Unsorted tuple with duplicates
    assert isinstance(items, EventsDict)
    assert len(items) == 2

    assert items[0] == event  # Sorted without duplicates
    assert items[1] == event_window

    item = events[r'GCO\d+']  # Single regex search (1 output only)
    assert isinstance(item, EventWindow)
    assert item == event_window

    with raises(KeyError, match='`FOO` not found'):
        _ = events['FOO']

    # Function filters
    items = events.find('G')
    assert isinstance(items, EventsDict)
    assert len(items) == 2

    item = events.startswith('g')  # Ignore case
    assert isinstance(item, EventWindow)

    item = events.endswith('E', 'DE')  # multi-letters and duplicates
    assert isinstance(item, Event)

    with raises(KeyError, match='`FOO`, `BAR` not found'):
        _ = events.find('FOO', 'BAR')

    # Iter
    for i, ev in enumerate(events):
        assert ev == event if i == 0 else event_window

    # Hash
    assert isinstance(hash(events), int)

    # Contains
    assert 'FLYBY_GANYMEDE' in events
    assert 'FOO' not in events

    assert '2035-05-23' in events
    assert '2035-10-05' in events
    assert '2035-05-22' not in events
    assert '2035-10-06' not in events

    # Keys
    assert tuple(events.keys()) == ('FLYBY_GANYMEDE', 'GCO500')

    # Event start and Stop times
    assert events.starts == [
        np.datetime64('2031-07-21T07:14:23'),
        np.datetime64('2035-05-22T06:00:00'),
    ]
    assert events.stops == [
        np.datetime64('2031-07-21T07:14:23'),
        np.datetime64('2035-10-05T16:01:00'),
    ]
    assert events.windows == [
        (np.datetime64('2031-07-21T07:14:23'), np.datetime64('2031-07-21T07:14:23')),
        (np.datetime64('2035-05-22T06:00:00'), np.datetime64('2035-10-05T16:01:00')),
    ]

    assert events.start == np.datetime64('2031-07-21T07:14:23')
    assert events.stop == np.datetime64('2035-10-05T16:01:00')
    assert events.start_date == '2031-07-21'
    assert events.stop_date == '2035-10-05'


def test_events_list(event, events_list):
    """Test events dictionary."""
    events = events_list

    assert isinstance(events, EventsList)
    assert str(events) == 'FLYBY_GANYMEDE (2031-07-21 -> 2031-07-21 | 2 events)'
    assert len(events) == 2

    # Getter
    assert events[0] == event
    assert events[1] == event

    items = events[:1]  # with a slice
    assert isinstance(items, EventsList)
    assert len(items) == 1
    assert items[0] == event

    # Iter
    for ev in events:
        assert ev == event

    # Get crema names if present
    assert events.crema_names == ['1G1', '1G1']

    # Get item by flyby crema name
    assert events['1G1'] == event

    # Get item with multiple flyby names (tuple)
    items = events['1G1', '1G1']
    assert isinstance(items, EventsList)
    assert len(items) == 2
    assert items[0]['Crema name'] == '1G1'

    # Event start and Stop times
    assert events.starts == [
        np.datetime64('2031-07-21T07:14:23'),
        np.datetime64('2031-07-21T07:14:23'),
    ]
    assert events.stops == [
        np.datetime64('2031-07-21T07:14:23'),
        np.datetime64('2031-07-21T07:14:23'),
    ]
    assert events.windows == [
        (np.datetime64('2031-07-21T07:14:23'), np.datetime64('2031-07-21T07:14:23')),
        (np.datetime64('2031-07-21T07:14:23'), np.datetime64('2031-07-21T07:14:23')),
    ]

    assert events.start == events.stop == np.datetime64('2031-07-21T07:14:23')
    assert events.start_date == events.stop_date == '2031-07-21'

    # Hash
    assert isinstance(hash(events), int)

    # Contains
    assert 'FLYBY_GANYMEDE' not in events  # no keys

    assert '2031-07-21T07:14:23' in events
    assert '2031-07-21' not in events
    assert '2031-07-22' not in events


def test_events_timedelta():
    """Test events timedelta inputs."""
    assert timedelta('1 s') == np.timedelta64(1, 's')
    assert timedelta('1 sec') == np.timedelta64(1, 's')
    assert timedelta('2 seconds') == np.timedelta64(2, 's')
    assert timedelta('1s') == np.timedelta64(1, 's')

    assert timedelta('1 m') == np.timedelta64(1, 'm')
    assert timedelta('1 min') == np.timedelta64(1, 'm')
    assert timedelta('1 minute') == np.timedelta64(1, 'm')

    assert timedelta('12 h') == np.timedelta64(12, 'h')
    assert timedelta('12 hour') == np.timedelta64(12, 'h')
    assert timedelta('12 H') == np.timedelta64(12, 'h')

    assert timedelta('1 D') == np.timedelta64(1, 'D')
    assert timedelta('1 d') == np.timedelta64(1, 'D')
    assert timedelta('2 days') == np.timedelta64(2, 'D')

    assert timedelta('1 M') == np.timedelta64(1, 'M')
    assert timedelta('1 month') == np.timedelta64(1, 'M')

    assert timedelta('1 Y') == np.timedelta64(1, 'Y')
    assert timedelta('1 y') == np.timedelta64(1, 'Y')
    assert timedelta('2 years') == np.timedelta64(2, 'Y')

    # With timedelta input (no change)
    assert timedelta(np.timedelta64(1, 'Y')) == np.timedelta64(1, 'Y')

    # Invalid input (only int are accepted)
    with raises(ValueError):
        _ = timedelta('1.5 s')


def test_events_operations(event, event_window):
    """Test events operations."""
    # Add/substract temporal step as a string
    assert event - '12 h' == np.datetime64('2031-07-20T19:14:23')
    assert event + '12 h' == np.datetime64('2031-07-21T19:14:23')

    # Add/substract temporal step as deltatime64
    assert event - np.timedelta64(12, 'h') == np.datetime64('2031-07-20T19:14:23')
    assert event + np.timedelta64(12, 'h') == np.datetime64('2031-07-21T19:14:23')

    # With EventWindow
    assert event_window - '12 h' == np.datetime64('2035-05-21T18:00:00')
    assert event_window + '12 h' == np.datetime64('2035-10-06T04:01:00')


def test_events_intersection(event_window):
    """Test events intersection on an event window and events list."""
    # With a single time string
    assert '2035-05-23' in event_window
    assert '2035-10-05' in event_window
    assert '2035-05-22' not in event_window
    assert '2035-10-06' not in event_window

    # With a numpy.datetime64
    assert np.datetime64('2035-05-23') in event_window

    # With a list / tuple
    assert ['2035-05-22', '2035-05-23', '2035-10-05', '2035-10-06'] in event_window
    assert_array(
        event_window.contains(['2035-05-22', '2035-05-23', '2035-10-05', '2035-10-06']),
        [False, True, True, False]
    )
    assert [np.datetime64('2035-05-23')] in event_window

    # With a custom instance
    class A:
        """Dummy class."""
        @property
        def utc(self):
            """UTC times"""
            return [
                '2035-05-22',
                '2035-05-23',
                '2035-10-05',
                '2035-10-06',
            ]

    a = A()

    assert a in event_window
    assert_array(event_window.contains(a), [False, True, True, False])

    # Same with an EventsList
    events = EventsList([event_window])

    # With a single time string
    assert '2035-05-23' in events
    assert '2035-10-05' in events
    assert '2035-05-22' not in events
    assert '2035-10-06' not in events

    # With a numpy.datetime64
    assert np.datetime64('2035-05-23') in events

    # With a list / tuple
    assert ['2035-05-22', '2035-05-23', '2035-10-05', '2035-10-06'] in events
    assert_array(
        events.contains(['2035-05-22', '2035-05-23', '2035-10-05', '2035-10-06']),
        [False, True, True, False]
    )
    assert [np.datetime64('2035-05-23')] in events

    # With a custom instance
    assert a in events
    assert_array(events.contains(a), [False, True, True, False])


def test_events_comparisons(event_window):
    """Test event comparison operations."""
    assert not event_window > '2035-05-22T06:00:00'
    assert event_window >= '2035-05-22T06:00:00'
    assert not event_window > '2035-10-05T16:01:00'
    assert not event_window >= '2035-10-05T16:01:00'

    assert event_window > '2035-05-22'
    assert not event_window > '2035-05-23'
    assert not event_window < '2035-10-05'
    assert event_window < '2035-10-06'

    assert not event_window < '2035-05-22T06:00:00'
    assert not event_window <= '2035-05-22T06:00:00'
    assert not event_window < '2035-10-05T16:01:00'
    assert event_window <= '2035-10-05T16:01:00'

    # Check input types
    assert event_window > '2035'
    assert event_window > 2035
    assert event_window > np.datetime64('2035')


def test_events_errors():
    """Test events errors."""
    with raises(ValueError):
        _ = AbstractEvent('missing-time', a=1)

    class DummyEvent(AbstractEvent):  # pylint: disable=abstract-method
        """Dummy event object."""
        def __init__(self, *args, **kwargs):
            pass

    event = DummyEvent()
    with raises(NotImplementedError):
        _ = event.start

    with raises(NotImplementedError):
        _ = event.stop

    class DummyEvents(AbstractEventsCollection):  # pylint: disable=abstract-method
        """Dummy event object."""
        def __init__(self, *args, **kwargs):
            pass

    events = DummyEvents()
    with raises(NotImplementedError):
        _ = hash(events)


def test_events_ipython_autocompletion(event, event_window, events_dict, events_list):
    """Test Events IPython autocompletion."""
    def tab(obj):
        """Test IPython <tab> autocompletion."""
        return tuple(obj._ipython_key_completions_())  # pylint: disable=protected-access

    assert tab(event) == (
        'event time [utc]', 'Crema name', 'Altitude above ganymede [km]')
    assert tab(event_window) == ('t_start', 't_end')
    assert tab(events_dict) == ('FLYBY_GANYMEDE', 'GCO500')
    assert tab(events_list) == ('1G1', '1G1')


def test_events_html_repr(event, event_window, events_dict, events_list):
    """Test Events HTML representation."""
    assert display_html(event) == (
        '<table>'
        '<tr>'
        '<td>event time [utc]</td>'
        '<td>2031-07-21T07:14:23Z</td>'
        '</tr>'
        '<tr>'
        '<td>Crema name</td>'
        '<td>1G1</td>'
        '</tr>'
        '<tr>'
        '<td>Altitude above ganymede [km]</td>'
        '<td>400.00697</td>'
        '</tr>'
        '</table>'
    )

    assert display_html(event_window) == (
        '<table>'
        '<tr>'
        '<td>t_start</td>'
        '<td>2035-05-22T06:00:00</td>'
        '</tr>'
        '<tr>'
        '<td>t_end</td>'
        '<td>2035-10-05T16:01:00</td>'
        '</tr>'
        '</table>'
    )

    assert display_html(events_dict) == (
        '<table>'
        '<thead>'
        '<tr>'
        '<th>'
        '</th>'
        '<th>event</th>'
        '<th>#</th>'
        '<th>t_start</th>'
        '<th>t_stop</th>'
        '</tr>'
        '</thead>'
        '<tbody>'
        '<tr>'
        '<td>'
        '<em>0</em>'
        '</td>'
        '<td>FLYBY_GANYMEDE</td>'
        '<td>-</td>'
        '<td>2031-07-21</td>'
        '<td>2031-07-21</td>'
        '</tr>'
        '<tr>'
        '<td>'
        '<em>1</em>'
        '</td>'
        '<td>GCO500</td>'
        '<td>-</td>'
        '<td>2035-05-22</td>'
        '<td>2035-10-05</td>'
        '</tr>'
        '</tbody>'
        '</table>'
    )

    assert display_html(events_list) == (
        '<table>'
        '<thead>'
        '<tr>'
        '<th>'
        '</th>'
        '<th>event time [utc]</th>'
        '<th>Crema name</th>'
        '<th>Altitude above ganymede [km]</th>'
        '</tr>'
        '</thead>'
        '<tbody>'
        '<tr>'
        '<td>'
        '<em>0</em>'
        '</td>'
        '<td>2031-07-21T07:14:23Z</td>'
        '<td>1G1</td>'
        '<td>400.00697</td>'
        '</tr>'
        '<tr>'
        '<td>'
        '<em>1</em>'
        '</td>'
        '<td>2031-07-21T07:14:23Z</td>'
        '<td>1G1</td>'
        '<td>400.00697</td>'
        '</tr>'
        '</tbody>'
        '</table>'
    )
