"""Test Map module."""

from pathlib import Path

import numpy as np

import matplotlib.pyplot as plt

from pytest import fixture, raises

from moon_coverage import (
    CALLISTO, EARTH, EUROPA, GANYMEDE, IO, JUPITER, MAPS, MERCURY, MOON, VENUS
)
from moon_coverage.maps import Map
from moon_coverage.spice import SpiceRef


DATA = Path(__file__).parent / '..' / 'data'


@fixture
def m():
    """Default Map object."""
    return Map(GANYMEDE.fname, size=(512, 256))


def test_map_attr(m):
    """Test Map object attributes."""
    assert str(m) == m.fname.name == 'Ganymede.jpg'
    assert m.body is None


def test_map_img_attr(m):
    """Test Map image attributes."""
    img = m.img
    assert img.size == (512, 256)

    data = np.asarray(img)
    assert data[-1, 0] == 0
    assert data[-1, 256] != 0


def test_map_centered_0_2d():
    """Test Map image centered at 0 in 2D."""
    m = Map(GANYMEDE.fname, centered_0=True, size=(512, 256))

    img = m.img
    assert img.size == (512, 256)

    # Ganymede flipped image
    data = np.asarray(img)
    assert data[-1, 0] != 0
    assert data[-1, 256] == 0


def test_map_centered_0_3d():
    """Test Map image centered at 0 in 3D."""
    m = Map(IO.fname, centered_0=True, size=(512, 256))

    img = m.img
    assert img.size == (512, 256)

    # Io flipped image
    data = np.asarray(img)
    assert data[0, 0, 0] == 153
    assert data[0, 256, 0] == 152


def test_map_plot(m):
    """Test map plot with grid."""
    fout = DATA / 'foo.jpg'

    if fout.exists():
        fout.unlink()

    m.map(fout=fout)

    assert fout.exists()
    fout.unlink()


def test_maps_defaults():
    """Test default maps."""
    assert str(CALLISTO) == 'Callisto'
    assert str(EARTH) == 'Earth'
    assert str(EUROPA) == 'Europa'
    assert str(GANYMEDE) == 'Ganymede'
    assert str(MOON) == 'Moon'
    assert str(IO) == 'Io'
    assert str(JUPITER) == 'Jupiter'
    assert str(VENUS) == 'Venus'
    assert str(MERCURY) == 'Mercury'

    assert CALLISTO.radius == 2_410.3
    assert EARTH.radius == 6_371.0
    assert EUROPA.radius == 1_560.8
    assert GANYMEDE.radius == 2_631.2
    assert MOON.radius == 1_737.4
    assert IO.radius == 1_821.5
    assert JUPITER.radius == 69_911.3
    assert VENUS.radius == 6_051.8
    assert MERCURY.radius == 2_439.7


def test_maps_registry():
    """Test maps registry."""
    assert len(MAPS) == 9
    assert 'Ganymede' in MAPS

    ganymede = MAPS['ganymede']  # Case insensitive

    assert isinstance(ganymede, Map)
    assert str(ganymede) == 'Ganymede'

    MAPS['foo'] = SpiceRef('Saturn')

    assert MAPS['FOO'] == 'SATURN'


def test_map_projection():
    """Test map projection."""
    fig = plt.figure(figsize=(12, 12))
    ax = fig.add_subplot(projection=GANYMEDE)

    assert ax.proj.PROJ4 == 'eqc'
    assert ax.bg == GANYMEDE.fname
    assert ax.bg_extent == [0, 360, -90, 90]
    assert ax.target == 'Ganymede'

    with raises(NotImplementedError):
        _ = fig.add_subplot(projection=Map(GANYMEDE.fname, centered_0=True))
