"""Test vector module."""

import numpy as np
from numpy.testing import assert_array_almost_equal as assert_array

from pytest import approx, raises

from moon_coverage.math import angle, cs, lonlat, xyz
from moon_coverage.math.vectors import (
    boresight_rot_m, ell_norm, hat, norm, rot_axis,
    scalar_proj, scalar_rejection, vdot, vector_rejection
)


def test_cos_sin():
    """Test cosines and sines on angles in degrees."""
    cos, sin = cs(0)
    assert cos == approx(1, abs=1e-9)
    assert sin == approx(0, abs=1e-9)

    cos, sin = cs(45)
    assert cos == approx(0.707, abs=1e-3)
    assert sin == approx(0.707, abs=1e-3)

    cos, sin = cs(90)
    assert cos == approx(0, abs=1e-9)
    assert sin == approx(1, abs=1e-9)

    cos, sin = cs(180)
    assert cos == approx(-1, abs=1e-9)
    assert sin == approx(0, abs=1e-9)

    cos, sin = cs(-90)
    assert cos == approx(0, abs=1e-9)
    assert sin == approx(-1, abs=1e-9)

    # 1D array
    cos, sin = cs([0, 45, 90, 180, -90])
    assert_array(cos, [1, 0.707, 0, -1, 0], decimal=3)
    assert_array(sin, [0, 0.707, 1, 0, -1], decimal=3)

    # 2D array
    cos, sin = cs([[0, 45], [180, -90]])
    assert_array(cos, [[1, 0.707], [-1, 0]], decimal=3)
    assert_array(sin, [[0, 0.707], [0, -1]], decimal=3)


def test_norm():
    """Test vector norm."""
    assert norm([1, 0, 0]) == 1
    assert norm([0, 1, 0]) == 1
    assert norm([0, 0, 1]) == 1
    assert norm([0, 0, 0]) == 0
    assert norm([1, 1, 0]) == np.sqrt(2)
    assert norm([1, 1, 1]) == np.sqrt(3)
    assert norm([-1, 0, 0]) == 1

    # 1D array of 3D vectors
    v1 = [[1, 0, 0], [0, 1, 0], [1, 1, 0], [1, 1, 1]]

    assert_array(norm(v1), [1, 1, np.sqrt(2), np.sqrt(3)])

    # 2D array of 3D vectors
    v2 = [[[1, 0, 0], [0, 1, 0]], [[1, 1, 0], [1, 1, 1]]]
    assert_array(norm(v2), [[1, 1], [np.sqrt(2), np.sqrt(3)]])


def test_hat():
    """Test normalized vector."""
    assert_array(hat([1, 0, 0]), [1, 0, 0])
    assert_array(hat([0, 1, 0]), [0, 1, 0])
    assert_array(hat([1, 1, 0]), [0.707, 0.707, 0], decimal=3)
    assert_array(hat([1, 1, 1]), [0.577, 0.577, 0.577], decimal=3)

    # Norm == 0
    assert_array(hat([0, 0, 0]), [0, 0, 0])

    # 1D array of 3D vectors
    v1 = [[1, 0, 0], [0, 0, 0], [1, 1, 0], [1, 1, 1]]

    assert_array(hat(v1), [
        [1, 0, 0], [0, 0, 0],
        [0.707, 0.707, 0], [0.577, 0.577, 0.577]
    ], decimal=3)

    # 2D array of 3D vectors
    v2 = [[[1, 0, 0], [0, 0, 0]], [[1, 1, 0], [1, 1, 1]]]

    assert_array(hat(v2), [
        [[1, 0, 0], [0, 0, 0]],
        [[0.707, 0.707, 0], [0.577, 0.577, 0.577]]
    ], decimal=3)


def test_lonlat():
    """Test cartesian to spheric conversion (east longitude)."""
    assert_array(lonlat([1, 0, 0]), [0, 0])
    assert_array(lonlat([0, 1, 0]), [90, 0])
    assert_array(lonlat([1, 1, 0]), [45, 0])
    assert_array(lonlat([1, 0, 1]), [0, 45])
    assert_array(lonlat([0, -1, 0]), [270, 0])

    # Norm == 0
    assert_array(lonlat([0, 0, 0]), [0, 0])

    # 1D array
    v1 = [[0, 0, 0], [0, 1, 0], [1, 1, 0], [1, 0, 1], [0, -1, 0]]

    lon, lat = lonlat(v1)
    assert_array(lon, [0, 90, 45, 0, 270])
    assert_array(lat, [0, 0, 0, 45, 0])

    # 2D array
    v2 = [[[0, 0, 0], [0, 1, 0]], [[1, 1, 0], [1, 0, 1]]]

    lon, lat = lonlat(v2)
    assert_array(lon, [[0, 90], [45, 0]])
    assert_array(lat, [[0, 0], [0, 45]])


def test_xyz():
    """Test spherical to cartesian conversion."""
    assert_array(xyz(0, 0), [1, 0, 0])
    assert_array(xyz(90, 0), [0, 1, 0])
    assert_array(xyz(45, 0), [0.707, 0.707, 0], decimal=3)
    assert_array(xyz(0, 45), [0.707, 0, 0.707], decimal=3)
    assert_array(xyz(270, 0), [0, -1, 0])

    # With a defined radius
    assert_array(xyz(45, 0, r=np.sqrt(2)), [1, 1, 0])

    # 1D array (single)
    assert_array(xyz([0, 90, 45, 270], 0), [
        [1, 0, 0], [0, 1, 0], [0.707, 0.707, 0], [0, -1, 0]
    ], decimal=3)

    assert_array(xyz(0, [0, 45]), [[1, 0, 0], [0.707, 0, 0.707]], decimal=3)

    assert_array(xyz(45, 0, r=[1, np.sqrt(2)]), [[.707, .707, 0], [1, 1, 0]], decimal=3)

    # 1D array (multi)
    assert_array(xyz([0, 90, 0, 270], [0, 0, 45, 0]), [
        [1, 0, 0], [0, 1, 0], [0.707, 0, 0.707], [0, -1, 0]
    ], decimal=3)

    assert_array(xyz([0, 90, 0, 270], [0, 0, 45, 0], r=[1, 1, np.sqrt(2), 1]), [
        [1, 0, 0], [0, 1, 0], [1, 0, 1], [0, -1, 0]
    ], decimal=3)

    # 2D array
    assert_array(xyz([[0, 90], [0, 270]], [[0, 0], [45, 0]]), [
        [[1, 0, 0], [0, 1, 0]], [[0.707, 0, 0.707], [0, -1, 0]]
    ], decimal=3)

    assert_array(xyz(
        [[0, 90], [0, 270]], [[0, 0], [45, 0]], r=[[1, 1], [np.sqrt(2), 1]]), [
        [[1, 0, 0], [0, 1, 0]], [[1, 0, 1], [0, -1, 0]]
    ], decimal=3)


def test_vdot():
    """Test dot product between two vectors."""
    assert vdot([1, 0, 0], [1, 0, 0]) == 1
    assert vdot([1, 0, 0], [0, 1, 0]) == 0
    assert vdot([1, 0, 0], [0, 0, 1]) == 0
    assert vdot([0, 0, 0], [0, 0, 0]) == 0
    assert vdot([1, 0, 0], [-1, 0, 0]) == -1

    # 1D array
    v1 = [[1, 0, 0], [1, 0, 0], [1, 0, 0], [1, 0, 0]]
    v2 = [[1, 0, 0], [0, 1, 0], [0, 0, 1], [-1, 0, 0]]

    assert_array(vdot([1, 0, 0], v2), [1, 0, 0, -1])
    assert_array(vdot(v2, [1, 0, 0]), [1, 0, 0, -1])

    assert_array(vdot(v1, v2), [1, 0, 0, -1])

    # 2D array
    v1 = [[[1, 0, 0], [1, 0, 0]], [[1, 0, 0], [1, 0, 0]]]
    v2 = [[[1, 0, 0], [0, 1, 0]], [[-1, 0, 0], [0, 0, 1]]]

    assert_array(vdot(v1, v2), [[1, 0], [-1, 0]])

    with raises(ValueError):
        _ = vdot([[1, 0, 0]], v1)


def test_angle():
    """Test dot product."""
    assert angle([1, 0, 0], [1, 0, 0]) == 0
    assert angle([1, 0, 0], [0, 1, 0]) == 90
    assert angle([1, 0, 0], [-1, 0, 0]) == 180
    assert vdot([1, 0, 0], [0, 0, 0]) == 0

    assert angle([1, 0, 0], [2, 0, 0]) == 0
    assert angle([1, 0, 0], [0, 2, 0]) == 90
    assert angle([1, 0, 0], [-2, 0, 0]) == 180

    # Not unitary vector
    assert angle([1, 0, 0], [2, 0, 0]) == 0

    # Norm == 0
    assert angle([0, 0, 0], [1, 0, 0]) == 90
    assert angle([1, 0, 0], [0, 0, 0]) == 90

    # 1D array
    v1 = [[1, 0, 0], [1, 0, 0], [1, 0, 0], [1, 0, 0]]
    v2 = [[1, 0, 0], [0, 1, 0], [0, 0, 1], [-1, 0, 0]]

    assert_array(angle([1, 0, 0], v2), [0, 90, 90, 180])
    assert_array(angle(v2, [1, 0, 0]), [0, 90, 90, 180])

    assert_array(angle(v1, v2), [0, 90, 90, 180])

    # 2D array
    v1 = [[[1, 0, 0], [1, 0, 0]], [[1, 0, 0], [1, 0, 0]]]
    v2 = [[[1, 0, 0], [0, 1, 0]], [[-1, 0, 0], [0, 0, 1]]]

    assert_array(angle(v1, v2), [[0, 90], [180, 90]])

    # NaN support
    v1 = [[[1, 0, 0], [np.nan, 0, 0]], [[1, 0, 0], [1, 0, np.nan]]]
    v2 = [[[1, 0, 0], [0, 1, 0]], [[-1, np.nan, 0], [np.nan, 0, 1]]]

    assert_array(angle(v1, v2), [[0, np.nan], [np.nan, np.nan]])


def test_scalar_proj():
    """Test scalar projection between two vectors."""
    assert scalar_proj([1, 0, 0], [1, 0, 0]) == 1
    assert scalar_proj([2, 0, 0], [1, 0, 0]) == 2
    assert scalar_proj([0, 0, 0], [1, 0, 0]) == 0
    assert scalar_proj([1, 1, 0], [1, 0, 0]) == 1
    assert scalar_proj([1, 1, 1], [1, 0, 0]) == 1
    assert scalar_proj([-2, 1, 3], [1, 0, 0]) == -2

    assert scalar_proj([1, 0, 0], [0, 1, 0]) == 0
    assert scalar_proj([1, 0, 0], [0, 0, 1]) == 0
    assert scalar_proj([1, 0, 0], [0, 0, 0]) == 0
    assert scalar_proj([1, 0, 0], [2, 0, 0]) == 1
    assert scalar_proj([1, 0, 0], [-1, 0, 0]) == -1
    assert scalar_proj([1, 0, 0], [-2, 0, 0]) == -1

    # 1D array
    v = [[1, 0, 0], [2, 0, 0], [0, 0, 0], [1, 1, 0], [1, 1, 1], [-2, 1, 3]]
    n = [[0, 1, 0], [0, 0, 1], [0, 0, 0], [2, 0, 0], [-1, 0, 0], [-2, 0, 0]]

    assert_array(scalar_proj(v, [1, 0, 0]), [1, 2, 0, 1, 1, -2])
    assert_array(scalar_proj([1, 0, 0], n), [0, 0, 0, 1, -1, -1])

    assert_array(scalar_proj(v, n), [0, 0, 0, 1, -1, 2])

    # 2D array
    v = [[[1, 0, 0], [2, 0, 0], [0, 0, 0]], [[1, 1, 0], [1, 1, 1], [-2, 1, 3]]]
    n = [[[0, 1, 0], [0, 0, 1], [0, 0, 0]], [[2, 0, 0], [-1, 0, 0], [-2, 0, 0]]]

    assert_array(scalar_proj(v, n), [[0, 0, 0], [1, -1, 2]])


def test_scalar_rejection():
    """Test scalar rejection between two vectors."""
    assert scalar_rejection([1, 0, 0], [1, 0, 0]) == 0
    assert scalar_rejection([0, 1, 0], [1, 0, 0]) == 1
    assert scalar_rejection([0, 0, 1], [1, 0, 0]) == 1
    assert scalar_rejection([0, 1, 1], [1, 0, 0]) == approx(1.414, abs=1e-3)
    assert scalar_rejection([0, 0, 0], [1, 0, 0]) == 0
    assert scalar_rejection([0, -2, 0], [1, 0, 0]) == 2
    assert scalar_rejection([0, 0, 0], [0, 0, 0]) == 0

    assert scalar_rejection([1, 0, 0], [0, 1, 0]) == 1
    assert scalar_rejection([1, 0, 0], [0, 0, 1]) == 1
    assert scalar_rejection([1, 0, 0], [0, 0, 0]) == 1
    assert scalar_rejection([1, 0, 0], [2, 0, 0]) == 0
    assert scalar_rejection([1, 0, 0], [0, 2, 0]) == 1
    assert scalar_rejection([1, 0, 0], [0, -1, 0]) == 1

    # 1D array
    v = [[1, 0, 0], [0, 1, 0], [0, 0, 1], [0, 1, 1], [0, 0, 0], [0, -2, 0]]
    n = [[0, 1, 0], [0, 0, 1], [0, 0, 0], [2, 0, 0], [0, 2, 0], [0, -1, 0]]

    assert_array(scalar_rejection(v, [1, 0, 0]), [0, 1, 1, 1.414, 0, 2], decimal=3)
    assert_array(scalar_rejection([1, 0, 0], n), [1, 1, 1, 0, 1, 1])

    assert_array(scalar_rejection(v, n), [1, 1, 1, 1.414, 0, 0], decimal=3)

    # 2D array
    v = [[[1, 0, 0], [0, 1, 0], [0, 0, 1]], [[0, 1, 1], [0, 0, 0], [0, -2, 0]]]
    n = [[[0, 1, 0], [0, 0, 1], [0, 0, 0]], [[2, 0, 0], [0, 2, 0], [0, -1, 0]]]

    assert_array(scalar_rejection(v, n), [[1, 1, 1], [1.414, 0, 0]], decimal=3)


def test_vector_rejection():
    """Test vector rejection between two vectors."""
    assert_array(vector_rejection([1, 0, 0], [1, 0, 0]), [0, 0, 0])
    assert_array(vector_rejection([0, 1, 0], [1, 0, 0]), [0, 1, 0])
    assert_array(vector_rejection([0, 0, 1], [1, 0, 0]), [0, 0, 1])
    assert_array(vector_rejection([0, 1, 1], [1, 0, 0]), [0, 1, 1])
    assert_array(vector_rejection([1, 1, 1], [1, 0, 0]), [0, 1, 1])
    assert_array(vector_rejection([0, 0, 0], [1, 0, 0]), [0, 0, 0])
    assert_array(vector_rejection([0, -2, 0], [1, 0, 0]), [0, -2, 0])
    assert_array(vector_rejection([0, 0, 0], [0, 0, 0]), [0, 0, 0])

    assert_array(vector_rejection([1, 0, 0], [0, 1, 0]), [1, 0, 0])
    assert_array(vector_rejection([1, 0, 0], [0, 0, 1]), [1, 0, 0])
    assert_array(vector_rejection([1, 0, 0], [0, 0, 0]), [1, 0, 0])
    assert_array(vector_rejection([1, 0, 0], [2, 0, 0]), [0, 0, 0])
    assert_array(vector_rejection([1, 0, 0], [0, 2, 0]), [1, 0, 0])
    assert_array(vector_rejection([1, 0, 0], [0, -1, 0]), [1, 0, 0])

    # 1D array
    v = [[1, 0, 0], [0, 1, 0], [0, 0, 1], [0, 1, 1], [0, 0, 0], [0, -2, 0]]
    n = [[0, 1, 0], [0, 0, 1], [0, 0, 0], [2, 0, 0], [0, 2, 0], [0, -1, 0]]

    assert_array(vector_rejection(v, [1, 0, 0]), [
        [0, 0, 0],
        [0, 1, 0],
        [0, 0, 1],
        [0, 1, 1],
        [0, 0, 0],
        [0, -2, 0],
    ])

    assert_array(vector_rejection([1, 0, 0], n), [
        [1, 0, 0],
        [1, 0, 0],
        [1, 0, 0],
        [0, 0, 0],
        [1, 0, 0],
        [1, 0, 0],
    ])

    assert_array(vector_rejection(v, n), [
        [1, 0, 0],
        [0, 1, 0],
        [0, 0, 1],
        [0, 1, 1],
        [0, 0, 0],
        [0, 0, 0],
    ])

    # 2D array
    v = [[[1, 0, 0], [0, 1, 0], [0, 0, 1]], [[0, 1, 1], [0, 0, 0], [0, -2, 0]]]
    n = [[[0, 1, 0], [0, 0, 1], [0, 0, 0]], [[2, 0, 0], [0, 2, 0], [0, -1, 0]]]

    assert_array(vector_rejection(v, n), [
        [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        [[0, 1, 1], [0, 0, 0], [0, 0, 0]],
    ])


def test_ell_norm():
    """Test normal vector on a ellipsoid."""
    # Check on a sphere
    radii = [1, 1, 1]

    assert_array(ell_norm([1, 0, 0], radii), [1, 0, 0])
    assert_array(ell_norm([0, 1, 0], radii), [0, 1, 0])
    assert_array(ell_norm([0, 0, 1], radii), [0, 0, 1])
    assert_array(ell_norm([1, 1, 1], radii), [0.577, 0.577, 0.577], decimal=3)

    # Check on an ellipsoid
    radii = [1, 2, 3]

    assert_array(ell_norm([1, 0, 0], radii), [1, 0, 0])  # Invariant along the x-axis
    assert_array(ell_norm([0, 1, 0], radii), [0, 1, 0])  # Invariant along the y-axis
    assert_array(ell_norm([0, 0, 1], radii), [0, 0, 1])  # Invariant along the z-axis
    assert_array(ell_norm([1, 1, 0], radii), [0.97, 0.24, 0], decimal=2)
    assert_array(ell_norm([1, 0, 1], radii), [0.99, 0, 0.11], decimal=2)
    assert_array(ell_norm([0, 1, 1], radii), [0, 0.91, 0.41], decimal=2)

    # 1D array
    v = [[1, 0, 0], [1, 1, 0], [1, 0, 1], [0, 1, 1]]

    assert_array(ell_norm(v, radii), [
        [1, 0, 0],
        [0.97, 0.24, 0],
        [0.99, 0, 0.11],
        [0, 0.91, 0.41],
    ], decimal=2)

    # 2D array
    v = [[[1, 0, 0], [1, 1, 0]], [[1, 0, 1], [0, 1, 1]]]

    assert_array(ell_norm(v, radii), [
        [[1, 0, 0], [0.97, 0.24, 0]],
        [[0.99, 0, 0.11], [0, 0.91, 0.41]],
    ], decimal=2)

    # Errors
    with raises(ValueError):
        _ = ell_norm([1, 0, 0], [-1, 0, 0])

    # NaN support
    v = [[[1, 0, 0], [np.nan, 1, 0]], [[1, np.nan, 1], [0, 1, np.nan]]]

    assert_array(ell_norm(v, radii), [
        [[1, 0, 0], [np.nan, np.nan, np.nan]],
        [[np.nan, np.nan, np.nan], [np.nan, np.nan, np.nan]],
    ], decimal=2)


def test_rot_axis():
    """Test rotation matrix along an axis."""
    assert_array(rot_axis(0, (0, 0, 1)), [
        [1, 0, 0],
        [0, 1, 0],
        [0, 0, 1],
    ])

    assert_array(rot_axis(90, (0, 0, 1)), [
        [0, -1, 0],
        [1, 0, 0],
        [0, 0, 1],
    ])

    assert_array(rot_axis(-90, (0, 0, 1)), [
        [0, 1, 0],
        [-1, 0, 0],
        [0, 0, 1],
    ])

    assert_array(rot_axis(180, (0, 0, 1)), [
        [-1, 0, 0],
        [0, -1, 0],
        [0, 0, 1],
    ])

    assert_array(rot_axis(90, (1, 0, 0)), [
        [1, 0, 0],
        [0, 0, -1],
        [0, 1, 0],
    ])

    assert_array(rot_axis(90, (0, 1, 0)), [
        [0, 0, 1],
        [0, 1, 0],
        [-1, 0, 0],
    ])

    assert_array(rot_axis(120, (1, 1, 1)), [
        [0, 0, 1],
        [1, 0, 0],
        [0, 1, 0],
    ])

    assert_array(rot_axis(-120, (1, 1, 1)), [
        [0, 1, 0],
        [0, 0, 1],
        [1, 0, 0],
    ])

    # On null-vector
    assert_array(rot_axis(90, (0, 0, 0)), [
        [0, 0, 0],
        [0, 0, 0],
        [0, 0, 0],
    ])

    # 1D angle list
    assert_array(rot_axis([0, 90, -90, 180], (0, 0, 1)), [
        [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        [[0, -1, 0], [1, 0, 0], [0, 0, 1]],
        [[0, 1, 0], [-1, 0, 0], [0, 0, 1]],
        [[-1, 0, 0], [0, -1, 0], [0, 0, 1]],
    ])


def test_boresight_rot_matrix():
    """Test boresight rotation matrix."""
    assert_array(boresight_rot_m((0, 0, 1)), np.identity(3))

    # 90° along y
    m = boresight_rot_m((1, 0, 0))

    assert_array(m, [
        [0, 0, -1], [0, 1, 0], [1, 0, 0]
    ])
    assert_array(m @ [1, 0, 0], [0, 0, 1])

    # -90° along y
    m = boresight_rot_m((-1, 0, 0))

    assert_array(m, [
        [0, 0, -1], [0, -1, 0], [-1, 0, 0]
    ])
    assert_array(m @ [-1, 0, 0], [0, 0, 1])

    # 180° along y
    m = boresight_rot_m((0, 0, -1))

    assert_array(m, [
        [-1, 0, 0], [0, 1, 0], [0, 0, -1]
    ])
    assert_array(m @ [0, 0, -1], [0, 0, 1])

    # 45° along y
    m = boresight_rot_m((1, 0, 1))

    r = np.sqrt(2)
    assert_array(m, [
        [r / 2, 0, -r / 2], [0, 1, 0], [r / 2, 0, r / 2]
    ], decimal=2)
    assert_array(m @ [1, 0, 1], [0, 0, r], decimal=3)

    # 90° along z and 90° along y
    m = boresight_rot_m((0, -1, 0))

    assert_array(m, [
        [0, 0, -1], [1, 0, 0], [0, -1, 0]
    ])
    assert_array(m @ [0, -1, 0], [0, 0, 1])

    # -90° along z and 90° along y
    m = boresight_rot_m((0, 1, 0))

    assert_array(m, [
        [0, 0, -1], [-1, 0, 0], [0, 1, 0]
    ])
    assert_array(m @ [0, 1, 0], [0, 0, 1])

    # -45° along z and 45° along y
    m = boresight_rot_m((1, 1, 1))

    assert_array(m @ m.T, np.identity(3))
    assert_array(m.T @ m, np.identity(3))

    assert_array(m @ [1, 1, 1], [0, 0, np.sqrt(3)], decimal=3)

    # Error
    with raises(ValueError):
        _ = boresight_rot_m((0, 0, 0))
