"""Test miscellaneous cache module."""
# pylint: disable=comparison-with-callable

from moon_coverage.debug import debug_cache
from moon_coverage.misc import cached_property


def test_cache_property():
    """Test cache property decorator."""
    class A:
        """Dummy A object."""
        i = 1

        @cached_property
        def double(self):
            """Double the internal value."""
            self.i *= 2
            return self.i

    a = A()
    assert a.i == 1

    # Check that the cache is empty
    assert '_double_cached' not in a.__dict__
    assert '__cache_registry' not in a.__dict__
    assert 'clear_cache' not in a.__dict__

    # Compute from no cache
    assert a.double == 2
    assert a.i == 2

    # Check cache content
    assert '_double_cached' in a.__dict__
    assert '__cache_registry' in a.__dict__
    assert 'clear_cache' in a.__dict__

    # Reload value from the cache
    assert a.double == 2
    assert a.i == 2

    # Remove value from the cache
    del a.double
    assert '_double_cached' not in a.__dict__

    # Verify the value can be re-computed and cached again
    assert a.i == 2
    assert a.double == 4
    assert a.i == 4
    assert '_double_cached' in a.__dict__

    # Delete with delattr
    delattr(a, 'double')
    assert '_double_cached' not in a.__dict__

    # Refill the cache
    assert a.double == 8

    # Check clear cache
    assert callable(a.clear_cache)  # pylint: disable=no-member

    a.clear_cache()  # pylint: disable=no-member

    assert '_double_cached' not in a.__dict__
    assert '__cache_registry' not in a.__dict__
    assert 'clear_cache' not in a.__dict__


def test_cache_property_debugger(caplog):
    """Test cache property decorator debugger."""
    class B:
        """Dummy B object."""
        i = 1

        @cached_property
        def plus_one(self):
            """Add one to the internal value."""
            self.i += 1
            return self.i

    b = B()

    # Enable cache debugger
    debug_cache(True)

    assert caplog.messages == ['Change logger level to DEBUG']
    caplog.clear()

    # Fill the cache
    assert b.i == 1
    assert b.plus_one == 2
    assert b.i == 2

    assert caplog.messages == [
        '`plus_one` is not in the cache',
        '`plus_one` value is 2',
        '`plus_one` is now saved in the cache',
        '`__cache_registry` and `clear_cache()` were added',
        '`plus_one` parent(s) are registered in the cache',
    ]
    caplog.clear()

    # Load from the cache
    assert b.plus_one == 2
    assert b.i == 2

    assert caplog.messages == ['`plus_one` is loaded from the cache']
    caplog.clear()

    # Remove from the cache
    del b.plus_one

    assert caplog.messages == ['`plus_one` was removed from the cache']
    caplog.clear()

    # Clear all cache
    b.clear_cache()  # pylint: disable=no-member

    assert caplog.messages == [
        '`__cache_registry` and `clear_cache()` were removed',
        'The cache is now clean',
    ]

    # Disable cache debugger
    debug_cache(False)


def test_cache_property_parent():
    """Test cache property parent.

    Warning
    -------
    Here we voluntary change the internal value
    of i to see the evolution of the cache state.
    In a real objet, if a property is cached
    you need to invalid it when the values
    in its method change.

    Garbage collection is implement for the
    list of the parents.

    """
    class C:
        """Dummy C object."""
        i = 1

        @cached_property
        def double(self):
            """Double the internal value."""
            self.i *= 2
            return self.i

        @cached_property(parent='double')
        def double_plus_one(self):
            """Add one to the doubled value."""
            self.i = self.double + 1
            return self.i

        @cached_property()  # Empty parents
        def triple(self):
            """Add three to the doubled value."""
            self.i *= 3
            return self.i

        @cached_property(parent=['double', 'triple'])
        def double_plus_triple(self):
            """Add doubled and the triple."""
            self.i = self.double + self.triple
            return self.i

        @cached_property(parent=('double_plus_triple',))
        def double_plus_triple_plus_one(self):
            """Add doubled and the triple plus one."""
            self.i = self.double_plus_triple + 1
            return self.i

    c = C()

    # Load all the caches
    assert c.i == 1                    # Initial value
    assert c.double == 2               # Computed
    assert c.i == 2                    # Updated value
    assert c.double_plus_one == 3      # Computed
    assert c.double == 2               # Loaded from cache (old value)
    assert c.i == 3                    # Updated value
    assert c.triple == 9               # Computed
    assert c.double_plus_one == 3      # Loaded from cache (old value)
    assert c.i == 9                    # Updated value
    assert c.double_plus_triple == 11  # Loaded from cache (old value)
    assert c.i == 11                   # Updated value

    # Clear with parent but no children
    del c.double_plus_one

    assert c.double_plus_one == 3      # Recompute with old values (not cleared)
    assert c.i == 3                    # Updated value
    assert c.double == 2               # Loaded from cache (old value)
    assert c.triple == 9               # Loaded from cache (old value)
    assert c.double_plus_triple == 11  # Loaded from cache (old value)

    # Clear with parent with one children
    del c.triple

    c.i = 5                            # Updated value
    assert c.double == 2               # Loaded from cache (old value)
    assert c.double_plus_triple == 17  # Recomputed with old double value but new triple
    assert c.i == 17                   # Updated value
    assert c.triple == 15              # Triple was updated on double_plus_triple
    assert c.double_plus_triple_plus_one == 18  # Computed
    assert c.i == 18                   # Updated value

    # Clear with parent with two nested children
    del c.double

    assert '_triple_cached' in c.__dict__
    assert '_double_cached' not in c.__dict__
    assert '_double_plus_triple_cached' not in c.__dict__
    assert '_double_plus_triple_plus_one_cached' not in c.__dict__
