"""Test miscellaneous depreciation module."""

from moon_coverage.misc import DeprecationHelper, depreciated_renamed


def test_depreciation_renamed(caplog):
    """Test depreciation renamed class decorator."""
    class New:
        """Dummy new object."""
        def __init__(self, a, b=0):
            self.a = a
            self.b = b

    @depreciated_renamed
    class Old(New):
        """Dummy old object."""

    old = Old(1, b=10)

    assert isinstance(old, Old)

    assert old.a == 1
    assert old.b == 10

    assert old.__doc__ == '[Depreciated] Dummy old object.'

    assert caplog.messages == [
        '`Old` has been renamed `New`. '
        'Please update your code to use this new denomination.'
    ]


def test_depreciation_helper_dict(caplog):
    """Test depreciation replaced helper on a dict."""
    NEW = {'key': 'value'}

    OLD = DeprecationHelper('OLD', 'NEW', NEW)

    assert repr(OLD) == repr(NEW)

    assert 'key' in OLD
    assert 'missing' not in OLD

    assert OLD['key'] == 'value'

    assert len(caplog.messages) == 4

    assert set(caplog.messages) == {
        '`OLD` has been replaced by `NEW`. '
        'Please update your code to use this new denomination.'
    }


def test_depreciation_helper_func(caplog):
    """Test depreciation replaced helper on a function."""
    def new(a, b=0):
        """Dummy new function."""
        return a + b

    old = DeprecationHelper('old', 'new', new)

    assert old(1) == 1
    assert old(1, 2) == 3
    assert old(1, b=3) == 4

    assert len(caplog.messages) == 3

    assert set(caplog.messages) == {
        '`old` has been replaced by `new`. '
        'Please update your code to use this new denomination.'
    }


def test_depreciation_helper_class(caplog):
    """Test depreciation replaced helper on a class."""
    class New:
        """Dummy new class."""
        A = 2

        def __init__(self, a):
            self.a = a

        def __repr__(self):
            return 'new-function'

        def add(self, b):
            """Dummy add function."""
            return self.a + b

        @classmethod
        def attr(cls):
            """Dummy class method."""
            return cls.A

    Old = DeprecationHelper('Old', 'New', New)

    old = Old(1)  # log __call__

    assert repr(old) == 'new-function'

    assert old.a == 1
    assert old.add(2) == 3

    assert old.A == 2

    assert Old.attr() == 2  # log __getattr__

    assert len(caplog.messages) == 2

    assert set(caplog.messages) == {
        '`Old` has been replaced by `New`. '
        'Please update your code to use this new denomination.'
    }
