"""Test ground projection axes module."""

from numpy.testing import assert_array_almost_equal as assert_array

import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection, PatchCollection
from matplotlib.colorbar import Colorbar

from pytest import fixture, raises

from moon_coverage.maps import GANYMEDE
from moon_coverage.projections.axes import ProjectionMapTargetError
from moon_coverage.rois import CallistoROIs, GanymedeROIs
from moon_coverage.ticks import km_ticks, m_pix_ticks


@fixture
def traj():
    """Default trajectory."""
    class Trajectory:
        """Dummy trajectory."""
        lonlat = [(0, 10, 20), (0, 5, 10)]
        alt = [0, 100, 200]
        inc = [10, 30, 45]
        pixel_scale = [0.1, 0.3, 0.45]

    return Trajectory()


def test_axes_attr():
    """Test projection axes attributes."""
    fig = plt.figure()
    ax = fig.add_subplot(projection=GANYMEDE)

    assert ax.proj.PROJ4 == 'eqc'
    assert ax.bg == GANYMEDE.fname
    assert ax.bg_extent == [0, 360, -90, 90]
    assert ax.target == 'Ganymede'

    plt.close()


def test_axes_plot(traj):
    """Test projection axes plot."""
    fig = plt.figure()
    ax = fig.add_subplot(projection=GANYMEDE)

    # Trajectory-like (with `lonlat` attribute)
    ax.plot(traj)

    # Lon and Lat array
    lonlat = [(0, 10, 20), (0, 5, 10)]
    ax.plot(lonlat)

    # Lon and Lat individual variables
    lon, lat = (0, 10, 20), (0, 5, 10)
    ax.plot(lon, lat)

    # With dots
    ax.plot(lon, lat, 'ro')
    ax.plot(lon, lat, '.', color='red')

    # With lines
    ax.plot(lon, lat, 'r--')
    ax.plot(lon, lat, '--', color='red')

    # With 1 single point
    ax.plot(lon[0], lat[0])
    ax.plot(lon[0], lat[0], '.', color='red')

    plt.close()


def test_axes_colorline(traj):
    """Test projection axes color lines."""
    fig = plt.figure()
    ax = fig.add_subplot(projection=GANYMEDE)

    # Color lines with colorbar
    cbar = ax.plot(traj, 'alt')

    assert isinstance(cbar, Colorbar)
    assert cbar.ax.get_xlabel() == 'Altitude'
    assert cbar.cmap.name == 'turbo_r'
    assert cbar.vmin == 0
    assert cbar.vmax == 200
    assert_array(cbar.get_ticks(), [0, 50, 100, 150, 200])

    assert cbar.formatter(0) == '0 km'
    assert cbar.formatter(200) == '200 km'

    # Override vmin/vmax and label
    cbar = ax.plot(traj, 'inc', cmap='Reds', vmin=0, vmax=90, label='foo')

    assert isinstance(cbar, Colorbar)
    assert cbar.ax.get_xlabel() == 'foo'
    assert cbar.cmap.name == 'Reds'
    assert cbar.vmin == 0
    assert cbar.vmax == 90
    assert_array(cbar.get_ticks(), [0, 20, 40, 60, 80])

    assert cbar.formatter(0) == '0°'
    assert cbar.formatter(90) == '90°'

    # Override format
    cbar = ax.plot(traj, 'pixel_scale', vmin=0, vmax=1, fmt=m_pix_ticks)

    assert isinstance(cbar, Colorbar)
    assert cbar.ax.get_xlabel() == 'Pixel scale'
    assert cbar.vmin == 0
    assert cbar.vmax == 1
    assert_array(cbar.get_ticks(), [0, 0.2, 0.4, 0.6, 0.8, 1])

    assert cbar.formatter(0) == '0 m/pix'
    assert cbar.formatter(1) == '1,000 m/pix'

    # Color lines without colorbar
    lines = ax.plot(traj, 'inc', cbar=False)
    assert isinstance(lines, LineCollection)

    plt.close()


def test_axes_colorbar():
    """Test projection standalone colorbar."""
    fig = plt.figure()
    ax = fig.add_subplot(projection=GANYMEDE)

    # No initial colorbar
    with raises(ValueError, match='No parent colorbar found.'):
        ax.twin_colorbar()

    # Add a standalone horizontal colorbar
    cbar = ax.colorbar(vmin=0, vmax=90, cmap='Reds', label='inc')

    assert isinstance(cbar, Colorbar)
    assert cbar.ax.get_xlabel() == 'Incidence angle'
    assert cbar.cmap.name == 'Reds'
    assert cbar.vmin == 0
    assert cbar.vmax == 90
    assert_array(cbar.get_ticks(), [0, 20, 40, 60, 80])

    assert cbar.formatter(0) == '0°'
    assert cbar.formatter(90) == '90°'

    # Check twin colorbar
    cax = ax.twin_colorbar(label='Twin label', format=2 * km_ticks,
                           ticks=[0, 20, 40, 80, 100])

    assert cax.get_xlabel() == 'Twin label'
    assert cax.get_xlim() == (0, 90)
    assert_array(cax.get_xticks(), [0, 10, 20, 40, 50])

    fmt = cax.xaxis.get_major_formatter()
    assert fmt(0) == '0 km'
    assert fmt(100) == '200 km'

    ax.colorbar(vmin=0, vmax=1, orientation='vertical')
    cax = ax.twin_colorbar(label='Vertical twin', format=3 * km_ticks,
                           ticks=[0, 0.3, 0.9])

    assert cax.get_ylabel() == 'Vertical twin'
    assert cax.get_ylim() == (0, 1)

    assert_array(cax.get_yticks(), [0, 0.1, 0.3])

    fmt = cax.yaxis.get_major_formatter()
    assert fmt(0) == '0 km'
    assert fmt(100) == '300 km'

    plt.close()


def test_axes_patch_coll():
    """Test projection axes patch collection plot."""
    fig = plt.figure()
    ax = fig.add_subplot(projection=GANYMEDE)

    ax.add_path(GanymedeROIs[2, 0, 4].path)

    ax.add_patch(GanymedeROIs[2, 0, 4])
    ax.add_patch(GanymedeROIs[2, 0, 4](facecolor='blue'))

    ax.add_collection(GanymedeROIs)
    ax.add_collection(GanymedeROIs(facecolors='blue'))

    plt.close()


def test_axes_patch_coll_legend():
    """Test projection axes patch collection legend."""
    fig = plt.figure()
    ax = fig.add_subplot(projection=GANYMEDE)

    ax.plot([], [], label='Foo')

    handles, labels = ax.get_legend_handles_labels()
    assert len(handles) == len(labels) == 1
    assert labels[0] == 'Foo'

    leg = ax.legend()
    handler_map = leg.get_legend_handler_map()

    assert PatchCollection in handler_map

    plt.close()


def test_axes_proj_target_error(traj):
    """Test projection target errors."""
    fig = plt.figure()
    ax = fig.add_subplot(projection=GANYMEDE)

    traj.target = 'Callisto'
    with raises(ProjectionMapTargetError):
        _ = ax.plot(traj)

    with raises(ProjectionMapTargetError):
        _ = ax.add_patch(CallistoROIs[1, 0, 4])

    with raises(ProjectionMapTargetError):
        _ = ax.add_patch(CallistoROIs[1, 0, 4](facecolor='blue'))

    with raises(ProjectionMapTargetError):
        _ = ax.add_collection(CallistoROIs)

    with raises(ProjectionMapTargetError):
        _ = ax.add_collection(CallistoROIs(facecolors='blue'))

    plt.close()
