"""Test equirectangular projection module."""

import numpy as np
from numpy.testing import assert_array_almost_equal as assert_array

from pytest import approx, fixture

from moon_coverage.projections import Equirectangular


@fixture
def proj():
    """Default equirectangular projection."""
    return Equirectangular()  # Default: lon_e_0 = 180


@fixture
def proj_ganymede():
    """Equirectangular projection on Ganymede surface."""
    return Equirectangular(target='Ganymede', radius=2631.2)


def test_equi_attr(proj_ganymede):
    """Test equirectangular projection attributes."""
    assert str(proj_ganymede) == 'Equirectangular'
    assert proj_ganymede.PROJ4 == 'eqc'

    assert proj_ganymede.lat_0 == 0
    assert proj_ganymede.lon_e_0 == 180
    assert proj_ganymede.lat_ts == 0
    assert proj_ganymede.rc == 1

    assert proj_ganymede.x_0 == approx(2631.2e3 * np.pi, abs=1)

    assert proj_ganymede.xc == approx(2631.2e3 * np.pi, abs=1)
    assert proj_ganymede.yc == approx(2631.2e3 * np.pi / 2, abs=1)

    assert_array(proj_ganymede.extent,
                 [0, 2631.2e3 * 2 * np.pi,
                  -2631.2e3 * np.pi / 2, 2631.2e3 * np.pi / 2],
                 decimal=1)

    assert proj_ganymede.proj4 == (
        '+proj=eqc +lat_0=0 +lon_0=180 +lat_ts=0 +x_0=8266158.590 +y_0=0.000 '
        '+a=2631200.0 +b=2631200.0 +units=m +no_defs')

    assert proj_ganymede.wkt == (
        'PROJCS["PROJCS_Ganymede_Equirectangular",'
        'GEOGCS["GCS_Ganymede",'
        'DATUM["D_Ganymede",'
        'SPHEROID["Ganymede_Mean_Sphere", 2631200, 0]],'
        'PRIMEM["Greenwich",0],'
        'UNIT["Degree",0.017453292519943295]],'
        'PROJECTION["Equirectangular"],'
        'PARAMETER["false_easting", 8266158.590],'
        'PARAMETER["false_northing", 0.000],'
        'PARAMETER["standard_parallel_1", 0],'
        'PARAMETER["central_meridian", 180],'
        'PARAMETER["latitude_of_origin", 0],'
        'UNIT["Meter", 1]]')


def test_equi_xy(proj):
    """Test equirectangular projection values."""
    assert_array(proj(0, 0), (0, 0), decimal=6)
    assert_array(proj(90, 0), (90, 0), decimal=6)
    assert_array(proj(180, 0), (180, 0), decimal=6)
    assert_array(proj(270, 0), (270, 0), decimal=6)

    # Wrap longitudes
    assert_array(proj(359, 0), (359, 0), decimal=6)
    assert_array(proj(360, 0), (360, 0), decimal=6)  # Right side longitude
    assert_array(proj(361, 0), (1, 0), decimal=6)

    assert_array(proj(540, 0), (180, 0), decimal=6)
    assert_array(proj(720, 0), (0, 0), decimal=6)

    assert_array(proj(-90, 0), (270, 0), decimal=6)
    assert_array(proj(-180, 0), (180, 0), decimal=6)
    assert_array(proj(-360, 0), (360, 0), decimal=6)

    # Check latitudes
    assert_array(proj(0, 90), (0, 90), decimal=6)
    assert_array(proj(360, -90), (360, -90), decimal=6)

    # 1D array
    assert_array(
        proj([0, 90, 180, 270, 360, 361, -90], 0),
        ([0, 90, 180, 270, 360, 1, 270], [0, 0, 0, 0, 0, 0, 0]),
        decimal=6)

    assert_array(
        proj(180, [0, 90, -90]),
        ([180, 180, 180], [0, 90, -90]),
        decimal=6)

    assert_array(
        proj([0, 180, 360, 361, -90], [-90, 90, 0, 0, 0]),
        ([0, 180, 360, 1, 270], [-90, 90, 0, 0, 0]),
        decimal=6)

    # 2D array
    assert_array(
        proj([[0, 180], [360, -90]], [[-90, 90], [0, 0]]),
        ([[0, 180], [360, 270]], [[-90, 90], [0, 0]]),
        decimal=6)

    # Projection with an embedded value
    assert_array(proj(0, 90, value=1.234), (0, 90, 1.234), decimal=6)
    assert_array(proj([0], [90], value=[1.234]), ([0], [90], [1.234]), decimal=6)


def test_equi_lonlat(proj):
    """Test equirectangular projection reverse values."""
    assert_array(proj(0, 0, invert=True), (0, 0), decimal=6)
    assert_array(proj(90, 0, invert=True), (90, 0), decimal=6)
    assert_array(proj(180, 0, invert=True), (180, 0), decimal=6)
    assert_array(proj(270, 0, invert=True), (270, 0), decimal=6)
    assert_array(proj(360, 0, invert=True), (360, 0), decimal=6)

    # Wrap longitudes
    assert_array(proj(359, 0, invert=True), (359, 0), decimal=6)
    assert_array(proj(360, 0, invert=True), (360, 0), decimal=6)  # Right side longitude
    assert_array(proj(361, 0, invert=True), (1, 0), decimal=6)

    assert_array(proj(540, 0, invert=True), (180, 0), decimal=6)
    assert_array(proj(720, 0, invert=True), (0, 0), decimal=6)

    assert_array(proj(-90, 0, invert=True), (270, 0), decimal=6)
    assert_array(proj(-180, 0, invert=True), (180, 0), decimal=6)
    assert_array(proj(-360, 0, invert=True), (360, 0), decimal=6)

    # Check latitudes
    assert_array(proj(0, 90, invert=True), (0, 90), decimal=6)
    assert_array(proj(360, -90, invert=True), (360, -90), decimal=6)

    # 1D array
    assert_array(
        proj([0, 90, 180, 270, 360, 361, -90], 0, invert=True),
        ([0, 90, 180, 270, 360, 1, 270], [0, 0, 0, 0, 0, 0, 0]),
        decimal=6)

    assert_array(
        proj(180, [0, 90, -90], invert=True),
        ([180, 180, 180], [0, 90, -90]),
        decimal=6)

    assert_array(
        proj([0, 180, 360, 361, -90], [-90, 90, 0, 0, 0], invert=True),
        ([0, 180, 360, 1, 270], [-90, 90, 0, 0, 0]),
        decimal=6)

    # 2D array
    assert_array(
        proj([[0, 180], [360, -90]], [[-90, 90], [0, 0]], invert=True),
        ([[0, 180], [360, 270]], [[-90, 90], [0, 0]]),
        decimal=6)


def test_equi_x_0():
    """Test equirectangular projection values centered on 180° with null false easting."""
    proj = Equirectangular(lon_e_0=180, x_0=0)

    assert proj.x_0 == 0

    # Direct
    assert_array(proj(0, 0), (-180, 0), decimal=6)
    assert_array(proj(180, 0), (0, 0), decimal=6)
    assert_array(proj(360, 0), (180, 0), decimal=6)

    # Inverse
    assert_array(proj(-180, 0, invert=True), (0, 0), decimal=6)
    assert_array(proj(0, 0, invert=True), (180, 0), decimal=6)
    assert_array(proj(180, 0, invert=True), (360, 0), decimal=6)


def test_equi_lon_e_0():
    """Test equirectangular projection values centered on 0."""
    proj = Equirectangular(lon_e_0=0)

    assert proj.x_0 == 0

    # Direct
    assert_array(proj(0, 0), (0, 0), decimal=6)
    assert_array(proj(0, 90), (0, 90), decimal=6)
    assert_array(proj(0, -90), (0, -90), decimal=6)
    assert_array(proj(90, 0), (90, 0), decimal=6)
    assert_array(proj(-90, 0), (-90, 0), decimal=6)

    assert_array(proj(180, 0), (180, 0), decimal=6)    # Right side
    assert_array(proj(-180, 0), (-180, 0), decimal=6)  # Left side
    assert_array(proj(270, 0), (-90, 0), decimal=6)    # Wrap longitudes
    assert_array(proj(360, 0), (0, 0), decimal=6)      # Wrap longitudes

    # Inverse
    assert_array(proj(0, 0, invert=True), (0, 0), decimal=6)
    assert_array(proj(0, 90, invert=True), (0, 90), decimal=6)
    assert_array(proj(0, -90, invert=True), (0, -90), decimal=6)
    assert_array(proj(90, 0, invert=True), (90, 0), decimal=6)
    assert_array(proj(-90, 0, invert=True), (-90, 0), decimal=6)

    assert_array(proj(180, 0, invert=True), (180, 0), decimal=6)    # Right side
    assert_array(proj(-180, 0, invert=True), (-180, 0), decimal=6)  # Left side
    assert_array(proj(270, 0, invert=True), (-90, 0), decimal=6)    # Wrap longitudes
    assert_array(proj(360, 0, invert=True), (0, 0), decimal=6)      # Wrap longitudes


def test_equi_xy_plot(proj):
    """Test equirectangular projection of plotted values."""
    x, y = proj.xy_plot(
        [10, 90, 180, 270, 340, 20],
        [10, 20, 30, 40, 50, 60],
    )

    assert len(x) == len(y) == 9
    assert_array(x, [10, 90, 180, 270, 340, 360, np.nan, 0, 20])
    assert_array(y, [10, 20, 30, 40, 50, 55, np.nan, 55, 60])

    # Reversed order
    x, y = proj.xy_plot(
        [20, 340, 270, 180, 90, 10],
        [60, 50, 40, 30, 20, 10],
    )

    assert len(x) == len(y) == 9
    assert_array(x, [20, 0, np.nan, 360, 340, 270, 180, 90, 10])
    assert_array(y, [60, 55, np.nan, 55, 50, 40, 30, 20, 10])

    # With attached data
    x, y, v = proj.xy_plot(
        [10, 90, 180, 270, 340, 20],
        [10, 20, 30, 40, 50, 60],
        values=[1, 2, 3, 4, 5, 6]
    )

    assert len(x) == len(y) == len(v) == 9
    assert_array(x, [10, 90, 180, 270, 340, 360, np.nan, 0, 20])
    assert_array(y, [10, 20, 30, 40, 50, 55, np.nan, 55, 60])
    assert_array(v, [1, 2, 3, 4, 5, 5.5, np.nan, 5.5, 6])
