"""Test equirectangular projection module."""

from numpy.testing import assert_array_almost_equal as assert_array

from matplotlib.path import Path

from pytest import fixture, raises

from moon_coverage.projections import Equirectangular


@fixture
def proj():
    """Equirectangular projection on degree unit surface centered at 180° (default)."""
    return Equirectangular()


@fixture
def proj_0():
    """Equirectangular projection on degree unit surface centered at 0°."""
    return Equirectangular(lon_e_0=0)


@fixture
def proj_x0():
    """Equirectangular projection centered at 180° with 0 false easting."""
    return Equirectangular(x_0=0)


def test_equi_path_null(proj):
    """Test equirectangular projection on null path."""
    assert proj.xy_path(None) is None


def test_equi_path_closed(proj):
    """Test equirectangular projection on path is closed."""
    path = proj(Path([
        (170, 0),
        (180, 10),
        (190, 0),
        (180, -10),
    ], [Path.MOVETO] + 3 * [Path.LINETO]))

    assert len(path.vertices) == len(path.codes) == 5
    assert_array(path.vertices[0], path.vertices[-1])

    assert_array(path.vertices, [
        (170, 0),
        (180, 10),
        (190, 0),
        (180, -10),
        (170, 0),  # Added
    ], decimal=6)

    assert_array(
        path.codes,
        [Path.MOVETO] + 3 * [Path.LINETO] + [Path.CLOSEPOLY]
    )


def test_equi_path_npole_right(proj):
    """Test equirectangular projection North Pole crossing (right crossing)."""
    path = proj(Path([
        (20, 80),
        (100, 60),
        (200, 80),
        (300, 60),
    ]))

    assert_array(path.vertices, [
        (20, 80),
        (100, 60),
        (200, 80),
        (300, 60),
        (360, 75),
        (360, 90),
        (0, 90),
        (0, 75),
        (20, 80),
    ], decimal=6)

    assert_array(
        path.codes,
        [Path.MOVETO] + 7 * [Path.LINETO] + [Path.CLOSEPOLY]
    )


def test_equi_path_npole_left(proj):
    """Test equirectangular projection North Pole crossing (left crossing)."""
    path = proj(Path([
        (300, 60),
        (200, 80),
        (100, 60),
        (20, 80),
    ]))

    assert_array(path.vertices, [
        (300, 60),
        (200, 80),
        (100, 60),
        (20, 80),
        (0, 75),
        (0, 90),
        (360, 90),
        (360, 75),
        (300, 60),
    ], decimal=6)

    assert_array(
        path.codes,
        [Path.MOVETO] + 7 * [Path.LINETO] + [Path.CLOSEPOLY]
    )


def test_equi_path_spole_right(proj):
    """Test equirectangular projection South Pole crossing (right crossing)."""
    path = proj(Path([
        (20, -80),
        (100, -60),
        (200, -80),
        (300, -60),
    ]))

    assert_array(path.vertices, [
        (20, -80),
        (100, -60),
        (200, -80),
        (300, -60),
        (360, -75),
        (360, -90),
        (0, -90),
        (0, -75),
        (20, -80),
    ], decimal=6)

    assert_array(
        path.codes,
        [Path.MOVETO] + 7 * [Path.LINETO] + [Path.CLOSEPOLY]
    )


def test_equi_path_spole_left(proj):
    """Test equirectangular projection South Pole crossing (left crossing)."""
    path = proj(Path([
        (300, -60),
        (200, -80),
        (100, -60),
        (20, -80),
    ]))

    assert_array(path.vertices, [
        (300, -60),
        (200, -80),
        (100, -60),
        (20, -80),
        (0, -75),
        (0, -90),
        (360, -90),
        (360, -75),
        (300, -60),
    ], decimal=6)


def test_equi_path_lon_e_0_npole_right(proj_0):
    """Test equirectangular projection North Pole right-crossing centered at 0°."""
    path = proj_0(Path([
        (20, 80),
        (120, 60),
        (200, 80),
        (300, 60),
    ]))

    assert_array(path.vertices, [
        (20, 80),
        (120, 60),
        (180, 75),
        (180, 90),
        (-180, 90),
        (-180, 75),
        (-160, 80),
        (-60, 60),
        (20, 80),
    ], decimal=6)

    assert_array(
        path.codes,
        [Path.MOVETO] + 7 * [Path.LINETO] + [Path.CLOSEPOLY]
    )


def test_equi_path_lon_e_0_spole_left(proj_0):
    """Test equirectangular projection South Pole left-crossing centered at 0°."""
    path = proj_0(Path([
        (300, -60),
        (200, -80),
        (120, -60),
        (20, -80),
    ]))

    assert_array(path.vertices, [
        (-60, -60),
        (-160, -80),
        (-180, -75),
        (-180, -90),
        (180, -90),
        (180, -75),
        (120, -60),
        (20, -80),
        (-60, -60),
    ], decimal=6)

    assert_array(
        path.codes,
        [Path.MOVETO] + 7 * [Path.LINETO] + [Path.CLOSEPOLY]
    )


def test_equi_path_x0_closed(proj_x0):
    """Test equirectangular projection on path is closed with 0 false easting."""
    path = proj_x0(Path([
        (170, 0),
        (180, 10),
        (190, 0),
        (180, -10),
    ], [Path.MOVETO] + 3 * [Path.LINETO]))

    assert len(path.vertices) == len(path.codes) == 5
    assert_array(path.vertices[0], path.vertices[-1])

    assert_array(path.vertices, [
        (-10, 0),
        (0, 10),
        (10, 0),
        (0, -10),
        (-10, 0),  # Added
    ], decimal=6)

    assert_array(
        path.codes,
        [Path.MOVETO] + 3 * [Path.LINETO] + [Path.CLOSEPOLY]
    )


def test_equi_path_x0_npole_right(proj_x0):
    """Test equirectangular projection North Pole right-crossing with 0 false easting."""
    path = proj_x0(Path([
        (20, 80),
        (100, 60),
        (200, 80),
        (300, 60),
    ]))

    assert_array(path.vertices, [
        (-160, 80),
        (-80, 60),
        (20, 80),
        (120, 60),
        (180, 75),
        (180, 90),
        (-180, 90),
        (-180, 75),
        (-160, 80),
    ], decimal=6)

    assert_array(
        path.codes,
        [Path.MOVETO] + 7 * [Path.LINETO] + [Path.CLOSEPOLY]
    )


def test_equi_path_x0_spole_left(proj_x0):
    """Test equirectangular projection South Pole left-crossing with 0 false easting."""
    path = proj_x0(Path([
        (300, -60),
        (200, -80),
        (100, -60),
        (20, -80),
    ]))

    assert_array(path.vertices, [
        (120, -60),
        (20, -80),
        (-80, -60),
        (-160, -80),
        (-180, -75),
        (-180, -90),
        (180, -90),
        (180, -75),
        (120, -60),
    ], decimal=6)


def test_equi_path_antimeridian_1_2(proj):
    """Test equirectangular projection anti-meridian crossing (1|2)."""
    path = proj(Path([
        (20, 30),   # Pt. 1
        (-10, 0),   # Pt. 2
        (20, -30),  # Pt. 3
    ]))

    assert_array(path.vertices, [
        (20, 30),    # Pt. 1 (start left polygon)
        (0, 10),     # Edge
        (0, -10),    # Edge
        (20, -30),   # Pt. 3
        (20, 30),    # Pt. 1 (close left polygon)
        (360, 10),   # Edge (start right polygon)
        (350, 0),    # Pt. 2
        (360, -10),  # Edge
        (360, 10),   # Edge (close right polygon)
    ], decimal=6)

    assert_array(
        path.codes,
        [Path.MOVETO] + 3 * [Path.LINETO] + [Path.CLOSEPOLY] +
        [Path.MOVETO] + 2 * [Path.LINETO] + [Path.CLOSEPOLY]
    )


def test_equi_path_antimeridian_2_1(proj):
    """Test equirectangular projection anti-meridian crossing (2|1)."""
    path = proj(Path([
        (-20, 30),   # Pt. 1
        (10, 0),     # Pt. 2
        (-20, -30),  # Pt. 3
    ]))

    assert_array(path.vertices, [
        (0, 10),     # Edge (start left polygon)
        (10, 0),     # Pt. 2
        (0, -10),    # Pt. 2
        (0, 10),     # Edge (close left polygon)
        (340, 30),   # Pt. 1 (start right polygon)
        (360, 10),   # Edge
        (360, -10),  # Edge
        (340, -30),  # Pt. 3
        (340, 30),   # Pt. 1 (close right polygon)
    ], decimal=6)

    assert_array(
        path.codes,
        [Path.MOVETO] + 2 * [Path.LINETO] + [Path.CLOSEPOLY] +
        [Path.MOVETO] + 3 * [Path.LINETO] + [Path.CLOSEPOLY]
    )


def test_equi_path_antimeridian_1_3(proj):
    """Test equirectangular projection anti-meridian crossing (1|3)."""
    path = proj(Path([
        (50, 0),    # Pt. 1
        (20, 30),   # Pt. 2
        (-10, 0),   # Pt. 3
        (20, -30),  # Pt. 4
    ]))

    assert_array(path.vertices, [
        (50, 0),     # Pt. 1 (start left polygon)
        (20, 30),    # Pt. 2
        (0, 10),     # Edge
        (0, -10),    # Edge
        (20, -30),   # Pt. 4
        (50, 0),     # Pt. 1 (close left polygon)
        (360, 10),   # Edge (start right polygon)
        (350, 0),    # Pt. 2
        (360, -10),  # Edge
        (360, 10),   # Edge (close right polygon)
    ], decimal=6)

    assert_array(
        path.codes,
        [Path.MOVETO] + 4 * [Path.LINETO] + [Path.CLOSEPOLY] +
        [Path.MOVETO] + 2 * [Path.LINETO] + [Path.CLOSEPOLY]
    )


def test_equi_path_antimeridian_2_2(proj):
    """Test equirectangular projection anti-meridian crossing (2|2)."""
    path = proj(Path([
        (20, 40),    # Pt. 1
        (-10, 10),   # Pt. 2
        (-10, -10),  # Pt. 3
        (20, -40),   # Pt. 4
    ]))

    assert_array(path.vertices, [
        (20, 40),    # Pt. 1 (start left polygon)
        (0, 20),     # Edge
        (0, -20),    # Edge
        (20, -40),   # Pt. 4
        (20, 40),    # Pt. 1 (close left polygon)
        (360, 20),   # Edge (start right polygon)
        (350, 10),   # Pt. 2
        (350, -10),  # Pt. 3
        (360, -20),  # Edge
        (360, 20),   # Edge (close right polygon)
    ], decimal=6)

    assert_array(
        path.codes,
        [Path.MOVETO] + 3 * [Path.LINETO] + [Path.CLOSEPOLY] +
        [Path.MOVETO] + 3 * [Path.LINETO] + [Path.CLOSEPOLY]
    )


def test_equi_path_antimeridian_3_1(proj):
    """Test equirectangular projection anti-meridian crossing (3|1)."""
    path = proj(Path([
        (10, 0),     # Pt. 1
        (-20, 30),   # Pt. 2
        (-50, 0),    # Pt. 3
        (-20, -30),  # Pt. 4
    ]))

    assert_array(path.vertices, [
        (10, 0),     # Pt. 1 (start left polygon)
        (0, 10),     # Edge
        (0, -10),    # Edge
        (10, 0),     # Pt. 1 (close left polygon)
        (360, 10),   # Edge (start right polygon)
        (340, 30),   # Pt. 2
        (310, 0),   # Pt. 3
        (340, -30),  # Pt. 4
        (360, -10),  # Edge
        (360, 10),   # Edge (close right polygon)
    ], decimal=6)

    assert_array(
        path.codes,
        [Path.MOVETO] + 2 * [Path.LINETO] + [Path.CLOSEPOLY] +
        [Path.MOVETO] + 4 * [Path.LINETO] + [Path.CLOSEPOLY]
    )


def test_equi_path_lon_e_0_antimeridian_1_2(proj_0):
    """Test equirectangular projection anti-meridian crossing centered at 0°."""
    path = proj_0(Path([
        (170, 0),    # Pt. 1
        (200, 30),   # Pt. 2
        (200, -30),  # Pt. 3
    ]))

    assert_array(path.vertices, [
        (-180, 10),   # Edge (start left polygon)
        (-160, 30),   # Pt. 2
        (-160, -30),  # Pt. 3
        (-180, -10),  # Edge
        (-180, 10),   # Edge (close left polygon)
        (170, 0),     # Pt. 1 (start right polygon)
        (180, 10),    # Edge
        (180, -10),   # Edge
        (170, 0),     # Pt. 1 (close right polygon)
    ], decimal=6)

    assert_array(
        path.codes,
        [Path.MOVETO] + 3 * [Path.LINETO] + [Path.CLOSEPOLY] +
        [Path.MOVETO] + 2 * [Path.LINETO] + [Path.CLOSEPOLY]
    )


def test_equi_path_x0_antimeridian_1_2(proj_x0):
    """Test equirectangular projection anti-meridian with 0 false easting."""
    path = proj_x0(Path([
        (20, 30),   # Pt. 1
        (-10, 0),   # Pt. 2
        (20, -30),  # Pt. 3
    ]))

    assert_array(path.vertices, [
        (-160, 30),   # Pt. 1 (start left polygon)
        (-180, 10),   # Edge
        (-180, -10),  # Edge
        (-160, -30),  # Pt. 3
        (-160, 30),   # Pt. 1 (close left polygon)
        (180, 10),    # Edge (start right polygon)
        (170, 0),     # Pt. 2
        (180, -10),   # Edge
        (180, 10),    # Edge (close right polygon)
    ], decimal=6)

    assert_array(
        path.codes,
        [Path.MOVETO] + 3 * [Path.LINETO] + [Path.CLOSEPOLY] +
        [Path.MOVETO] + 2 * [Path.LINETO] + [Path.CLOSEPOLY]
    )


def test_equi_path_antimeridian_with_2_polygons(proj):
    """Test equirectangular projection anti-meridian crossing with 2 polygons."""
    path = proj(Path([
        (20, 30),   # Pt. 1
        (-10, 0),   # Pt. 2
        (20, -30),  # Pt. 3
        (20, 30),   # Pt. 1 (closure)
        (30, 30),   # Pt. 4
        (60, 0),    # Pt. 5
        (30, -30),  # Pt. 6
        (30, 30),   # Pt. 4 (closure)

    ], [
        Path.MOVETO,     # Pt. 1
        Path.LINETO,     # Pt. 2
        Path.LINETO,     # Pt. 3
        Path.CLOSEPOLY,  # Pt. 1 (closure)
        Path.MOVETO,     # Pt. 4
        Path.LINETO,     # Pt. 5
        Path.LINETO,     # Pt. 6
        Path.CLOSEPOLY,  # Pt. 4 (closure)
    ]))

    assert_array(path.vertices, [
        (20, 30),    # Pt. 1 (start left polygon)
        (0, 10),     # Edge
        (0, -10),    # Edge
        (20, -30),   # Pt. 3
        (20, 30),    # Pt. 1 (close left polygon)
        (30, 30),    # Pt. 4
        (60, 0),     # Pt. 5
        (30, -30),   # Pt. 6
        (30, 30),    # Pt. 4 (closure)
        (360, 10),   # Edge (start right polygon)
        (350, 0),    # Pt. 2
        (360, -10),  # Edge
        (360, 10),   # Edge (close right polygon)
    ], decimal=6)

    assert_array(
        path.codes,
        [Path.MOVETO] + 3 * [Path.LINETO] + [Path.CLOSEPOLY] +
        [Path.MOVETO] + 2 * [Path.LINETO] + [Path.CLOSEPOLY] +
        [Path.MOVETO] + 2 * [Path.LINETO] + [Path.CLOSEPOLY]
    )


def test_equi_path_hole(proj):
    """Test equirectangular projection polygon with a hole."""
    path = proj(Path([
        [0, 0], [4, 0], [4, 4], [0, 4], [0, 0],
        # Note: Holes must be counterclockwise
        [1, 1], [1, 2], [2, 2], [2, 1], [1, 1],
    ],
        [Path.MOVETO] + 3 * [Path.LINETO] + [Path.CLOSEPOLY] +
        [Path.MOVETO] + 3 * [Path.LINETO] + [Path.CLOSEPOLY]
    ))

    assert_array(path.vertices, [
        (0, 0), (4, 0), (4, 4), (0, 4), (0, 0),
        (1, 1), (1, 2), (2, 2), (2, 1), (1, 1),
    ], decimal=1)

    assert_array(
        path.codes,
        2 * ([Path.MOVETO] + 3 * [Path.LINETO] + [Path.CLOSEPOLY])
    )


def test_equi_path_antimeridian_hole(proj):
    """Test equirectangular projection anti-meridian crossing polygon with a hole."""
    path = proj(Path([
        [2, 2], [2, -2], [-2, -2], [-2, 2], [2, 2],
        # Note: Holes must be counter clockwise
        [1, 1], [-1, 1], [-1, -1], [1, -1], [1, 1],
    ],
        [Path.MOVETO] + 3 * [Path.LINETO] + [Path.CLOSEPOLY] +
        [Path.MOVETO] + 3 * [Path.LINETO] + [Path.CLOSEPOLY]
    ))

    assert_array(path.vertices, [
        (2, 2), (2, -2), (0, -2), (0, 2), (2, 2),             # Left polygon
        (1, 1), (0, 1), (0, -1), (1, -1), (1, 1),             # Left hole
        (360, -2), (358, -2), (358, 2), (360, 2), (360, -2),  # Right polygon
        (360, 1), (359, 1), (359, -1), (360, -1), (360, 1),   # Right hole
    ], decimal=1)

    assert_array(
        path.codes,
        4 * ([Path.MOVETO] + 3 * [Path.LINETO] + [Path.CLOSEPOLY])
    )


def test_equi_path_cross_err(proj):
    """Test equirectangular projection for too many crossings."""
    with raises(ValueError):
        _ = proj(Path([
            (0, 0),
            (120, 0),
            (240, 0),
            (0, 0),
            (120, 0),
            (240, 0),
            (0, 0),
            (120, 0),
            (240, 0),
        ]))
