"""Test Stephan 2021 ROIs module."""

from pathlib import Path

import numpy as np
from numpy.testing import assert_array_almost_equal as assert_array

from matplotlib.patches import Rectangle

from pytest import raises

from moon_coverage.rois import ROI, CallistoROIs, GanymedeROIs, ROIsCollection
from moon_coverage.rois.rois_stephan_2021_pss import Stephan2021ROIsCollection


DOI = 'https://doi.org/10.1016/j.pss.2021.105324'


def test_stephan_invalid_errors():
    """Test abstract Stephan ROIs invalid inputs."""
    with raises(AttributeError):
        _ = Stephan2021ROIsCollection()

    with raises(FileNotFoundError):
        csv = Path(__file__).parent / 'dummy.csv'
        _ = Stephan2021ROIsCollection(csv=csv)


def test_stephan_rois_attrs():
    """Test ROIs attributes."""
    assert CallistoROIs.csv.name == 'Callisto_ROIs-Stephan_2021_PSS-Tab_SM2.csv'
    assert CallistoROIs.prefix == 'JUICE_ROI_CAL_'
    assert CallistoROIs.target == 'Callisto'
    assert CallistoROIs.doi == DOI

    assert GanymedeROIs.csv.name == 'Ganymede_ROIs-Stephan_2021_PSS-Tab_SM1.csv'
    assert GanymedeROIs.prefix == 'JUICE_ROI_GAN_'
    assert GanymedeROIs.target == 'Ganymede'
    assert GanymedeROIs.doi == DOI


def test_stephan_keys():
    """Test ROIs keys generator."""
    assert GanymedeROIs.key(1) == '1'
    assert GanymedeROIs.key(1, 10) == '1.10'
    assert GanymedeROIs.key(1, 10, 3) == 'JUICE_ROI_GAN_1_10_03'
    assert GanymedeROIs.key(1, roi=10) == 'JUICE_ROI_GAN_1_0_10'

    # Pure string input
    assert GanymedeROIs.key('1') == '1'
    assert GanymedeROIs.key('1.10') == '1.10'
    assert GanymedeROIs.key('1_10_03') == 'JUICE_ROI_GAN_1_10_03'
    assert GanymedeROIs.key('JUICE_ROI_GAN_1_10_03') == 'JUICE_ROI_GAN_1_10_03'

    assert GanymedeROIs.key('1', '10') == '1.10'
    assert GanymedeROIs.key('1', '10', '3') == 'JUICE_ROI_GAN_1_10_03'
    assert GanymedeROIs.key('1', roi='10') == 'JUICE_ROI_GAN_1_0_10'

    # Exceptions
    assert GanymedeROIs.key('# 1', '0', '#FECCCC') == '1'
    assert GanymedeROIs.key(1, '0') == '1'
    assert GanymedeROIs.key(1, 0) == '1'
    assert GanymedeROIs.key('1', '0') == '1'
    assert GanymedeROIs.key(1, 0, 1) == 'JUICE_ROI_GAN_1_0_01'
    assert GanymedeROIs.key((1, 0, 1)) == 'JUICE_ROI_GAN_1_0_01'

    # Object with a `lonlat` property should not be affected
    class Foo:
        """Dummy class with `lonlat` a property."""
        @property
        def lonlat(self):
            "Dummy lonlat property."

    foo_lonlat = Foo()
    assert GanymedeROIs.key(foo_lonlat) is foo_lonlat


def test_rois_collections():
    """Test ROIs collections."""
    assert len(GanymedeROIs) == 139
    assert len(CallistoROIs) == 85

    # Contains
    assert '1_0_01' in GanymedeROIs
    assert 'JUICE_ROI_GAN_1_0_01' in GanymedeROIs

    assert '1_0_01' in CallistoROIs
    assert 'JUICE_ROI_CAL_1_0_01' in CallistoROIs
    assert 'JUICE_ROI_GAN_1_0_01' not in CallistoROIs

    assert '2_0_01' not in CallistoROIs
    assert 'JUICE_ROI_GAN_2_0_01' not in CallistoROIs

    # Getter
    assert GanymedeROIs['JUICE_ROI_GAN_1_0_01'] == 'JUICE_ROI_GAN_1_0_01'
    assert GanymedeROIs['1_0_01'] == 'JUICE_ROI_GAN_1_0_01'
    assert GanymedeROIs[1, 0, 1] == 'JUICE_ROI_GAN_1_0_01'

    # ROI properties
    roi = GanymedeROIs['JUICE_ROI_GAN_1_0_01']

    assert isinstance(roi, ROI)
    assert roi == 'JUICE_ROI_GAN_1_0_01'

    assert roi.name == 'Harakhtes'
    assert roi.color == '#0036fc'
    assert roi.target == 'Ganymede'
    assert roi.doi == DOI

    assert_array(roi.lons_e, [252, 268, 268, 252])
    assert_array(roi.lats, [34, 34, 46, 46])

    assert_array(roi.lon_e, 260)
    assert_array(roi.lat, 40)

    assert roi.category == '1'
    assert roi.category.name == 'Polar Deposits'

    # Iter
    for roi in GanymedeROIs:
        assert isinstance(roi, ROI)
        assert roi == 'JUICE_ROI_GAN_1_0_01'
        break

    # Instrument requirements
    roi = GanymedeROIs[2, 0, 1]  # 1, 1, 0, 1, 1, 0, 1, 1, 0, 0

    assert roi.name == 'Hammamat Patera'
    assert roi['UVS']
    assert roi['JANUS HR']
    assert not roi['JANUS COLOR']
    assert roi['JANUS STEREO']
    assert roi['MAJIS']
    assert not roi['SWI']
    assert roi['RIME']
    assert roi['GALA']
    assert not roi['PEP/RPWI']
    assert not roi['J-MAG']


def test_rois_categories():
    """Test ROIs categories."""
    assert len(GanymedeROIs.categories) == 7

    # Contains
    assert 1 in GanymedeROIs
    assert '1' in GanymedeROIs

    # Get category object
    category = GanymedeROIs.categories[1]

    assert category.name == 'Polar Deposits'
    assert category.description == (
        'Surface features where influences of the transition to polar caps onto '
        'the compositional/geological properties are expected'
    )
    assert category.color == '#0036fc'
    assert category.target == 'Ganymede'
    assert category.doi == DOI

    # Get category items
    rois = GanymedeROIs['4']

    assert isinstance(rois, ROIsCollection)
    assert len(rois) == 42

    for roi in rois:
        assert isinstance(roi, ROI)
        assert roi == 'JUICE_ROI_GAN_4_1_01'
        break


def test_rois_subcategories():
    """Test ROIs sub-categories."""
    assert len(GanymedeROIs.subcategories) == 10

    # Contains
    assert '1.0' not in GanymedeROIs
    assert '4.3' in GanymedeROIs

    # Get subcategory object
    subcategory = GanymedeROIs.subcategories['4.3']

    assert subcategory.name == 'Dome / Anomaleous Dome Craters'
    assert subcategory.description == 'Impact craters with a central dome'
    assert subcategory.color == '#9a35ff'
    assert subcategory.target == 'Ganymede'
    assert subcategory.doi == DOI
    assert subcategory.category.color == '#9a35ff'
    assert subcategory.category.target == 'Ganymede'
    assert subcategory.category.doi == DOI

    # Get category items
    rois = GanymedeROIs['4.3']

    assert isinstance(rois, ROIsCollection)
    assert len(rois) == 8

    for roi in rois:
        assert isinstance(roi, ROI)
        assert roi == 'JUICE_ROI_GAN_4_3_01'
        break

    # Invalid subcategory key
    with raises(KeyError):
        _ = GanymedeROIs['1.0']


def test_rois_patch_collection():
    """Test ROIs patches collection."""
    coll = CallistoROIs()

    assert_array(coll.get_facecolor(), np.zeros((len(CallistoROIs), 4)))
    assert coll.target == 'Callisto'


def test_rois_legend_handles():
    """Test ROIs legend handles."""
    handles = CallistoROIs.handles

    assert len(handles) == 6

    handle = handles[0]

    assert isinstance(handle, Rectangle)
    assert handle.get_label() == 'Polar Deposits'
    assert_array(handle.get_facecolor(), [0, 0.21, 0.99, 1], decimal=2)
