"""Test SPICE reference module."""

from pathlib import Path

import spiceypy as sp

from pytest import approx, raises

from moon_coverage.spice import SpiceBody


KERNELS = str((Path(__file__).parent / '..' / 'data').resolve())
tpc = KERNELS + '/dummy.tpc'


def test_body_code_validity():
    """Test SPICE body code validity."""
    assert SpiceBody.is_valid(10)
    assert SpiceBody.is_valid(199)
    assert SpiceBody.is_valid(999)

    assert not SpiceBody.is_valid(0)
    assert not SpiceBody.is_valid(11)
    assert not SpiceBody.is_valid(100)
    assert not SpiceBody.is_valid(1_000)


def test_body_attrs():
    """Test SPICE body attributes."""
    sp.kclear()
    sp.furnsh(tpc)

    # Planet body
    body = SpiceBody('JUPITER')

    assert body == 'JUPITER'
    assert body == 599

    assert body.is_planet

    assert body.frame == 'IAU_JUPITER'

    assert body.mu == approx(126_686_534.92, abs=1e-2)

    assert isinstance(body.parent, SpiceBody)
    assert body.parent == 'SUN'

    assert body.barycenter == 'JUPITER BARYCENTER'

    assert body['GM'] == approx(1.267e8, abs=1e5)  # BODY599_GM

    with raises(KeyError):
        _ = body['RADII']  # BODY599_RADII is not present in `dummy.tpc`

    # Satellite body
    body = SpiceBody('Ganymede')

    assert str(body) == body.name == 'GANYMEDE'
    assert int(body) == body.id == 503
    assert body.code == '503'
    assert not body.is_planet

    assert body.frame == 'IAU_GANYMEDE'

    assert tuple(body.radii) == (2631.2, 2631.2, 2631.2)
    assert body.radius == body.r == approx(2631.2, abs=1e-2)
    assert body.re == body.rp == 2631.2
    assert body.f == 0

    assert isinstance(body.parent, SpiceBody)
    assert body.parent == 'JUPITER'
    assert body.parent.mu == approx(126_686_534.92, abs=1e-2)

    assert body.barycenter == 'JUPITER BARYCENTER'

    assert tuple(body['RADII']) == (2631.2, 2631.2, 2631.2)  # BODY503_RADII

    with raises(KeyError):
        _ = body['GM']  # BODY503_GM is not present in `dummy.tpc`


def test_body_err():
    """Test SPICE body errors."""
    with raises(KeyError):
        _ = SpiceBody('JUICE')
