"""Test SPICE reference module."""

from pathlib import Path

import numpy as np
from numpy.testing import assert_array_almost_equal as assert_array

import spiceypy as sp

from pytest import approx, raises

from moon_coverage.spice import SpiceInstrument, SpiceSpacecraft


KERNELS = str((Path(__file__).parent / '..' / 'data').resolve())
tf = KERNELS + '/dummy.tf'
ti = KERNELS + '/dummy.ti'
tls = KERNELS + '/naif0012.tls'
tsc = KERNELS + '/dummy.tsc'


def test_inst_code_validity(monkeypatch):
    """Test SPICE instrument code validity."""
    sp.kclear()
    sp.furnsh([tf, ti])

    assert SpiceInstrument.is_valid(-28_200)

    # Invalid FOV
    assert not SpiceInstrument.is_valid(-1_001)

    # Mock SPICE `getfov` to make the FOV valid
    monkeypatch.setattr(sp, 'getfov', lambda *args: (None, None, None, None, None))

    assert SpiceInstrument.is_valid(-1_001)

    # Outside the range
    assert not SpiceInstrument.is_valid(-1_000)


def test_inst_attrs():
    """Test SPICE instrument attributes."""
    sp.kclear()
    sp.furnsh([tf, ti])

    inst = SpiceInstrument('JUICE_JANUS')

    assert inst == 'JUICE_JANUS'
    assert inst == -28_200

    # Getter
    assert inst['FOV_FRAME'] == 'JUICE_JANUS'  # INS-28200_FOV_FRAME

    with raises(KeyError):
        _ = inst['FOO']  # INS-28200_FOO is not present in `dummy.ti`

    # Instrument properties
    assert inst.ns == 1504
    assert inst.nl == 2000

    assert inst.fov_along_track == approx(0.0225, abs=1e-4)
    assert inst.fov_cross_track == approx(0.0300, abs=1e-4)

    # Exception: JANUS LINES are aligned with the flight direction
    assert inst.ifov == approx((15e-6, 15e-6), abs=1e-6)
    assert inst.ifov_along_track == approx(15e-6, abs=1e-6)
    assert inst.ifov_cross_track == approx(15e-6, abs=1e-6)

    # Field of view attributes
    assert inst.frame == 'JUICE_JANUS'
    assert inst.shape == 'RECTANGLE'

    assert tuple(inst.boresight) == (0, 0, 1)

    assert len(inst.bounds) == 4
    assert_array(inst.bounds, [
        [-0.015, 0.011, 0.999],
        [-0.015, -0.011, 0.999],
        [0.015, -0.011, 0.999],
        [0.015, 0.011, 0.999],
    ], decimal=3)

    # Parent spacecraft
    sc = inst.spacecraft

    assert isinstance(sc, SpiceSpacecraft)
    assert sc == 'JUICE'
    assert sc == -28

    assert sc['INS-28200_FOV_FRAME'] == 'JUICE_JANUS'  # INS-28200_FOV_FRAME

    with raises(KeyError):
        _ = sc['FOV_FRAME']  # FOV_FRAME alone is not valid

    # Instrument without PIXEL definition
    inst = SpiceInstrument('JUICE_PEP_NIM_S1')

    assert inst.ns == inst.nl == 1
    assert inst.fov_along_track == approx(0.1745, abs=1e-4)
    assert inst.fov_cross_track == approx(1.0472, abs=1e-4)
    assert inst.ifov == approx((0.1745, 1.0472), abs=1e-4)
    assert inst.ifov_along_track == approx(0.1745, abs=1e-4)
    assert inst.ifov_cross_track == approx(1.0472, abs=1e-4)

    # Instrument with a POLYGON shape
    inst = SpiceInstrument('JUICE_UVS')

    assert inst.ns == inst.nl == 1
    assert np.isnan(inst.fov_along_track)
    assert np.isnan(inst.fov_cross_track)
    assert np.isnan(inst.ifov).all()


def test_inst_sclk():
    """Test SPICE parent spacecraft clock."""
    sp.kclear()
    sp.furnsh([tf, ti, tls, tsc])

    inst = SpiceInstrument('JUICE_JANUS')

    assert inst.sclk('2022-MAY-31-23:59:59.999 TDB') == 0.0
    assert inst.sclk('2033-01-01') == 21_896_158_850_903.266


def test_inst_err():
    """Test SPICE instrument errors."""
    with raises(KeyError):
        _ = SpiceInstrument('JUICE')


def test_inst_juice_pep_exception(monkeypatch, caplog):
    """Test JUICE PEP instrument with out-of-range NAIF instrument code."""
    sp.kclear()
    sp.furnsh([tf, ti])

    assert SpiceInstrument.is_valid(-285_301)

    inst = SpiceInstrument('JUICE_PEP_NIM_S1')

    assert inst == 'JUICE_PEP_NIM_S1'
    assert inst == -285_301

    assert isinstance(inst.spacecraft, SpiceSpacecraft)
    assert inst.spacecraft == 'JUICE'

    assert caplog.messages == [
        'Invalid JUICE/PEP instrument NAIF IDs (-285301) for `JUICE_PEP_NIM_S1`. '
        'The parent spacecraft ID is manually set to `JUICE` (-28). '
        'See issue #12 (https://juigitlab.esac.esa.int/datalab/moon-coverage/-/issues/12)'
        ' for more details.'
    ]

    # A ValueError should be thrown for all the other invalid cases
    monkeypatch.setattr(SpiceInstrument, '__str__', lambda _: 'DUMMY_INSTRUMENT')

    dummy_inst = SpiceInstrument(-285_301)
    assert str(dummy_inst) == 'DUMMY_INSTRUMENT'

    with raises(ValueError):
        _ = dummy_inst.spacecraft
