"""Test SPICE limb computation."""

import numpy as np
from numpy.testing import assert_array_almost_equal as assert_array

import spiceypy as sp

from pytest import raises

from moon_coverage.spice.limb import limb_ip_pt


def test_limb_ip_pt(monkeypatch):
    """Test limb impact parameter function."""
    # SPICE mocks
    def spice_mocker(func, value):
        monkeypatch.setattr(sp, func, lambda *args, **kwargs: value)

    # Identity rotation
    mock_values = {
        'spkezp': ([0, 2_000, 0], 0),
        'pxform': np.identity(3),
    }

    for func, value in mock_values.items():
        spice_mocker(func, value)

    # Pointing at 45° toward the target
    xyz = limb_ip_pt(1_000_000_000, -28, 503, 'JUICE', (1, 1, 0), 'IAU_GANYMEDE')
    assert_array(xyz, [1_000, -1_000, 0])

    # Pointing at 30° toward the target
    xyz = limb_ip_pt(1_000_000_000, -28, 503, 'JUICE', (1, np.sqrt(3), 0), 'IAU_GANYMEDE')
    assert_array(xyz, [500 * np.sqrt(3), -500, 0])

    # Pointing at 60° toward the target
    xyz = limb_ip_pt(1_000_000_000, -28, 503, 'JUICE', (np.sqrt(3), 1, 0), 'IAU_GANYMEDE')
    assert_array(xyz, [500 * np.sqrt(3), -1_500, 0])

    # Pointing toward the target
    xyz = limb_ip_pt(1_000_000_000, -28, 503, 'JUICE', (0, 1, 0), 'IAU_GANYMEDE')
    assert_array(xyz, [0, 0, 0])

    # Pointing away from the target
    xyz = limb_ip_pt(1_000_000_000, -28, 503, 'JUICE', (1, -1, 0), 'IAU_GANYMEDE')
    assert_array(xyz, [np.nan, np.nan, np.nan])

    # Pointing tangently to the target
    xyz = limb_ip_pt(1_000_000_000, -28, 503, 'JUICE', (1, 0, 0), 'IAU_GANYMEDE')
    assert_array(xyz, [np.nan, np.nan, np.nan])

    # Pointing with a null-vector
    xyz = limb_ip_pt(1_000_000_000, -28, 503, 'JUICE', (0, 0, 0), 'IAU_GANYMEDE')
    assert_array(xyz, [np.nan, np.nan, np.nan])

    # Add 90° rotation around the z-axis compare to J2000 reference frame
    mock_values = {
        'spkezp': ([0, 4_000, 0], 0),
        'pxform': np.array([
            [0, -1, 0],
            [1, 0, 0],
            [0, 0, 1],
        ]),
    }

    for func, value in mock_values.items():
        spice_mocker(func, value)

    # Pointing at 45° toward the target
    xyz = limb_ip_pt(1_000_000_000, -28, 503, 'JUICE', (1, -1, 0), 'IAU_GANYMEDE')

    assert_array(xyz, [2_000, 2_000, 0])

    # Pointing at -45° toward the target
    xyz = limb_ip_pt(1_000_000_000, -28, 503, 'JUICE', (1, 1, 0), 'IAU_GANYMEDE')

    assert_array(xyz, [2_000, -2_000, 0])

    # Pointing away from the target (±135°)
    xyz = limb_ip_pt(1_000_000_000, -28, 503, 'JUICE', (-1, -1, 0), 'IAU_GANYMEDE')
    assert_array(xyz, [np.nan, np.nan, np.nan])

    xyz = limb_ip_pt(1_000_000_000, -28, 503, 'JUICE', (-1, 1, 0), 'IAU_GANYMEDE')
    assert_array(xyz, [np.nan, np.nan, np.nan])

    # Error when light time correction is requested
    with raises(NotImplementedError):
        _ = limb_ip_pt(1_000_000_000, -28, 503, 'JUICE', (0, 0, 1), 'IAU_GANYMEDE',
                       abcorr='LT')
