"""Test MetaKernel object."""

import re
from pathlib import Path

from pytest import fixture, raises

import moon_coverage.spice.metakernel as _metakernel
from moon_coverage import MetaKernel
from moon_coverage.spice.metakernel import MissingKernel, MissingKernelsRemote


KERNELS = str((Path(__file__).parent / '..' / 'data').resolve())


@fixture
def mk():
    """Default metakernel."""
    return MetaKernel(KERNELS + '/metakernel.tm', kernels=KERNELS)


def mock_wget(url, fname):  # pylint: disable=unused-argument
    """Mock download function."""
    f = Path(fname)
    f.parent.mkdir(parents=True, exist_ok=True)     # Create parents if needed
    f.write_text('foo c-kernel', encoding='utf-8')  # Add dummy values


def test_metakernel_attr(mk):
    """Test MetaKernel attributes."""
    # __eq__
    assert mk == MetaKernel(KERNELS + '/////metakernel.tm', kernels=KERNELS)
    assert mk != MetaKernel(KERNELS + '/metakernel.tm', kernels=KERNELS + '/')
    assert mk != 'metakernel.tm'

    assert str(mk) == 'metakernel.tm'

    assert mk.remote == 'ftp://example.com/kernels'

    # Context manager (__enter__/__exit__/create a temporay file, exists only in the loop)
    with mk as tmp_mk:
        # Check if a temporary file was created
        _mk = Path(tmp_mk)

        assert _mk.exists()
        assert _mk.name.startswith('metakernel-')
        assert _mk.name.endswith('.tm')

        # Check if the file have the edited values
        content = _mk.read_text(encoding='utf-8')

        assert f"PATH_VALUES = ( '{KERNELS}' )" in content
        assert '$KERNELS/naif0012.tls' in content
        assert 'Dummy text' not in content

    assert not _mk.exists()

    # __iter__ (create a temporay file, exists only in the loop)
    for tmp_mk in mk:
        _mk = Path(tmp_mk)
        assert _mk.exists()

    assert not _mk.exists()

    # Missing file
    with raises(FileNotFoundError):
        _ = MetaKernel('foo.tm')


def test_metakernel_content(mk):
    """Test MetaKernel content."""
    assert 'KPL/MK' in mk.content
    assert 'Dummy kernel' in mk.content
    assert r'\begintext' in mk.content
    assert r'\begindata' in mk.content
    assert 'End of the File' in mk.content


def test_metakernel_data(mk):
    """Test MetaKernel data."""
    assert mk.data['PATH_VALUES'] == [str(KERNELS)]
    assert mk.data['PATH_SYMBOLS'] == ['KERNELS']
    assert mk.data['KERNELS_TO_LOAD'] == [
        '$KERNELS/naif0012.tls', '$KERNELS/dummy.tf']


def test_metakernel_remote(mk):
    """Test MetaKernel remote."""
    assert mk.remote == 'ftp://example.com/kernels'
    assert mk.urls == [
        'ftp://example.com/kernels/naif0012.tls',
        'ftp://example.com/kernels/dummy.tf',
    ]

    # Change remote
    mk.remote = 'https://example.com/kernels'
    assert mk.remote == 'https://example.com/kernels'
    assert mk.urls == [
        'https://example.com/kernels/naif0012.tls',
        'https://example.com/kernels/dummy.tf',
    ]

    # Try to load the remote from the content
    mk.remote = 0
    assert mk.remote == 'ftp://example.com/kernels'
    assert mk.urls == [
        'ftp://example.com/kernels/naif0012.tls',
        'ftp://example.com/kernels/dummy.tf',
    ]


def test_metakernel_kernels(mk):
    """Test MetaKernel kernels."""
    assert len(mk.kernels) == 2

    # Contains
    assert KERNELS + '/naif0012.tls' in mk.kernels
    assert KERNELS + '/dummy.tf' in mk.kernels


def test_metakernel_content_data(monkeypatch):
    """Test MetaKernel content and data parser."""
    def mock_content(content, data):
        """Mock kernel parser content and data."""
        monkeypatch.setattr(_metakernel, 'kernel_parser', lambda _: (content, data))

    # Empty metakernel (without KERNELS_TO_LOAD key)
    mock_content('missing_kernels_to_load', {'MISSING_KERNELS_TO_LOAD': None})

    err_msg = re.escape("KERNELS_TO_LOAD key is missing")

    with raises(KeyError, match=err_msg):
        _ = MetaKernel(KERNELS + '/metakernel.tm', kernels=KERNELS)

    # Update value without PATH_SYMBOLS and PATH_VALUES keys
    mock_content('empty', {'KERNELS_TO_LOAD': ['foo.bar']})

    err_msg = re.escape("PATH_VALUES key is missing")

    with raises(KeyError, match=err_msg):
        _ = MetaKernel(KERNELS + '/metakernel.tm', foo='bar')

    # Update value with symbol not in PATH_SYMBOLS
    mock_content('empty', {
        'PATH_VALUES': ['bar'],
        'PATH_SYMBOLS': ['FOO'],
        'KERNELS_TO_LOAD': ['$FOO/bar'],
    })

    err_msg = re.escape(
        "Symbol: `BAZ` is not part of PATH_SYMBOLS (`FOO`)"
    )

    with raises(ValueError, match=err_msg):
        _ = MetaKernel(KERNELS + '/metakernel.tm', baz='qux')

    # Update only one symbol among many
    mock_content('empty', {
        'PATH_VALUES': ['foo', 'bar'],
        'PATH_SYMBOLS': ['FOO', 'BAR'],
        'KERNELS_TO_LOAD': ['$BAR/dummy.ti', '$FOO/dummy.tpc'],
    })

    mk = MetaKernel(KERNELS + '/metakernel.tm', bar=KERNELS, foo=KERNELS)

    assert len(mk.kernels) == 2
    assert mk.kernels[0] == KERNELS + '/dummy.ti'
    assert mk.kernels[1] == KERNELS + '/dummy.tpc'
    assert mk.urls == [None, None]

    # No remote url in the content
    mock_content('empty', {'KERNELS_TO_LOAD': []})
    mk = MetaKernel(KERNELS + '/metakernel.tm')

    assert mk.remote is None
    assert mk.urls == []

    # Multiple remote url in the content
    mock_content('https://foo\nftp://bar', {'KERNELS_TO_LOAD': []})

    mk = MetaKernel(KERNELS + '/metakernel.tm')

    assert mk.remote == 'https://foo'

    mk = MetaKernel(KERNELS + '/metakernel.tm', remote=1)

    assert mk.remote == 'ftp://bar'

    with raises(IndexError, match='list index out of range'):
        _ = MetaKernel(KERNELS + '/metakernel.tm', remote=2)

    # Override remote in content
    mk = MetaKernel(KERNELS + '/metakernel.tm', remote='https://example.org')

    assert mk.remote == 'https://example.org'

    # Missing kernel with known remote
    mock_content('https://example.org', {
        'PATH_VALUES': ['.'],
        'PATH_SYMBOLS': ['FOO'],
        'KERNELS_TO_LOAD': ['$FOO/dummy.tf'],
    })

    err_msg = re.escape(
        "`./dummy.tf` was not found locally. "
        "You can add `download=True` to download it automatically "
        "from `https://example.org` "
        "or/and you can change the kernel path value(s) by adding "
        "`foo='path/to/my/kernels'`."
    )

    with raises(MissingKernel, match=err_msg):
        _ = MetaKernel(KERNELS + '/metakernel.tm')

    # Missing kernel with unknown remote (and multiple symbols)
    mock_content('empty', {
        'PATH_VALUES': ['.', '..'],
        'PATH_SYMBOLS': ['FOO', 'BAR'],
        'KERNELS_TO_LOAD': ['$BAR/dummy.tpc'],
    })

    err_msg = re.escape(
        "`../dummy.tpc` was not found locally. "
        "You can add `download=True` to download it automatically "
        "or/and you can change the kernel path value(s) by adding "
        "`foo|bar='path/to/my/kernels'`."
    )

    with raises(MissingKernel, match=err_msg):
        _ = MetaKernel(KERNELS + '/metakernel.tm')

    # Download kernel without remote
    err_msg = re.escape(
        "Impossible to download the missing kernel `../dummy.tpc`. "
        "You need to provide a `remote='[https|http|ftp]://kernels.domain.tld/'`."
    )

    with raises(MissingKernelsRemote, match=err_msg):
        _ = MetaKernel(KERNELS + '/metakernel.tm', download=True)

    # Download missing kernel
    mock_content('https://example.org', {
        'PATH_VALUES': ['.'],
        'PATH_SYMBOLS': ['KERNELS'],
        'KERNELS_TO_LOAD': ['$KERNELS/ck/dummy.bc'],
    })

    monkeypatch.setattr(_metakernel, 'wget', mock_wget)  # mock download function

    ck = Path(KERNELS) / 'ck' / 'dummy.bc'

    if ck.exists():
        ck.unlink()

    if ck.parent.exists():
        ck.parent.rmdir()

    mk = MetaKernel(KERNELS + '/metakernel.tm', kernels=KERNELS, download=True)

    assert mk.kernels[0] == KERNELS + '/ck/dummy.bc'
    assert ck.exists()
    assert ck.read_text(encoding='utf-8') == 'foo c-kernel'

    ck.unlink()
    ck.parent.rmdir()

    # Kernel path too long (> 255 characters)
    with raises(BufferError):
        _ = MetaKernel(KERNELS + '/metakernel.tm', kernels=255 * 'a')
