"""Test SPICE times module."""

from pathlib import Path

import numpy as np
from numpy.testing import assert_array_equal

import spiceypy as sp

from pytest import raises

from moon_coverage.debug import debug_spice_pool
from moon_coverage.spice import MetaKernel, SpicePool, check_kernels


KERNELS = str((Path(__file__).parent / '..' / 'data').resolve())
kernel = KERNELS + '/naif0012.tls'

mk = MetaKernel(KERNELS + '/metakernel.tm', kernels=KERNELS)
mk_kernels_hash = hash((
    hash(mk),
    hash(KERNELS + '/naif0012.tls'),
    hash(KERNELS + '/dummy.tf'),
))


def test_spice_pool():
    """Test SPICE pool."""
    sp.kclear()

    assert int(SpicePool) == len(SpicePool) == 0

    # __add__
    _ = SpicePool + kernel

    assert int(SpicePool) == 1
    assert hash(SpicePool) == hash((hash(kernel),))

    # Try to add a kernel already present in the pool
    with raises(ValueError):
        SpicePool.add(kernel)

    # Purge before adding a list of kernels
    SpicePool.add([None, kernel], purge=True)

    assert int(SpicePool) == 1  # Only the valid kernels are loaded.

    # __eq__
    assert SpicePool == (kernel,)              # Tuple
    assert SpicePool == [kernel]               # List
    assert SpicePool == kernel                 # Str
    assert SpicePool == hash((hash(kernel),))  # Hash of hashes
    assert SpicePool == (None, kernel)         # Discard `None` values
    assert SpicePool != 12345

    # __contains__
    assert kernel in SpicePool
    assert 'foo' not in SpicePool

    # __iter__
    for k in SpicePool:  # pylint disable: not-an-iterable
        assert k == kernel

    # __sub__
    _ = SpicePool - kernel

    with raises(ValueError):
        SpicePool.remove(kernel)  # Already removed

    assert not SpicePool
    assert len(SpicePool) == 0

    # Add an array of kernels
    _ = SpicePool + [kernel]
    assert len(SpicePool) == 1

    # __getitem__
    assert SpicePool['DELTET/DELTA_T_A'] == 32.184

    # Purge
    SpicePool.purge()
    assert not SpicePool


def test_spice_logger(caplog):
    """Test SPICE Pool logger."""
    sp.kclear()

    # Enable debugger
    debug_spice_pool(True)

    assert caplog.messages == ['Change logger level to DEBUG']
    caplog.clear()

    SpicePool.add(kernel)
    assert caplog.messages == [f'Add `{kernel}` in the SPICE pool']
    caplog.clear()

    SpicePool.remove(kernel)
    assert caplog.messages == [f'Remove {kernel}']
    caplog.clear()

    SpicePool.purge()
    assert caplog.messages == ['Purge the pool']
    caplog.clear()

    # Enable debugger with INFO level
    debug_spice_pool('info')
    assert caplog.messages == ['Change logger level to INFO']
    caplog.clear()

    SpicePool.add(kernel, purge=True)
    SpicePool.remove(kernel)

    assert caplog.messages == ['Purge the pool']
    caplog.clear()

    # Disable debugger
    debug_spice_pool(False)

    SpicePool.add(kernel, purge=True)
    SpicePool.remove(kernel)

    assert caplog.messages == []


def test_spice_windows_coverage(monkeypatch):  # noqa: disable=C901
    """Test spice windows and coverage."""
    SpicePool.add(kernel, purge=True)

    def ktotal(_):  # pylint: disable=unused-argument
        """SPICE ktotal mock."""
        return 5

    def kdata(i, _):  # pylint: disable=unused-argument
        """SPICE kdata mock."""
        return [
            ('lsk/naif0011.tls', 'TEXT', None, None),
            ('ck/dummy.bc', 'CK', None, None),
            ('pck/dummy.bpc', 'PCK', None, None),
            ('pck/dummy.tpc', 'TEXT', None, None),
            ('spk/dummy.bsp', 'SPK', None, None),
        ][i]

    def bods2c(*args):
        """SPICE bods2c mock on JUICE_SPACECRAFT."""
        return -28_000

    def bodc2n(*args):
        """SPICE bodc2n mock on JUICE_SPACECRAFT."""
        return 'JUICE_SPACECRAFT'

    def ckobj(*args):
        """SPICE ckobj mock.

        From:
            - `juice_sc_sat_crema_150l0a_conjctn_20241029_20301212_f20160326_v01.bc`

        """
        return [-28_000]

    def ckcov(*args):
        """SPICE ckcov mock.

        `JUICE_SPACECRAFT` (-28_000) coverage on:
            - `juice_sc_sat_crema_150l0a_conjctn_20241029_20301212_f20160326_v01.bc`

        """
        return [
            783500680, 783870237,
            880011923, 880901281,
            942003659, 942745795,
            975964826, 976605399,
        ]

    def wncard(cover):
        """SPICE wncard mock."""
        return len(cover) // 2

    def cell_int(*args):
        """SPICE cell_int mock."""
        return [-28_000]

    def pckfrm(*args):
        """SPICE pckfrm mock."""

    def cell_double(*args):
        """SPICE cell_int mock."""
        return [94305664.183, 757080065.183]

    def pckcov(*args):
        """SPICE pckcov mock.

        `EARTH` (3_000) coverage on:
            - `earth_031228_231229_predict.bpc`

        """

    def spkobj(*args):
        """SPICE spkobj mock."""
        return [-28_000]

    def spkcov(*args):
        """SPICE spkcov mock.

        `JUICE_SPACECRAFT` (-28_000) coverage on:
            - `juice_cog_v00.bsp`

        """
        return [707313600, 1577880000]

    monkeypatch.setattr(sp, 'ktotal', ktotal)
    monkeypatch.setattr(sp, 'kdata', kdata)
    monkeypatch.setattr(sp, 'bods2c', bods2c)
    monkeypatch.setattr(sp, 'bodc2n', bodc2n)
    monkeypatch.setattr(sp, 'ckobj', ckobj)
    monkeypatch.setattr(sp, 'ckcov', ckcov)
    monkeypatch.setattr(sp, 'wncard', wncard)
    monkeypatch.setattr(sp, 'cell_int', cell_int)
    monkeypatch.setattr(sp, 'cell_double', cell_double)
    monkeypatch.setattr(sp, 'pckfrm', pckfrm)
    monkeypatch.setattr(sp, 'pckcov', pckcov)
    monkeypatch.setattr(sp, 'spkobj', spkobj)
    monkeypatch.setattr(sp, 'spkcov', spkcov)

    assert len(SpicePool) == 5

    # ET output format
    ets = SpicePool.windows('JUICE_SPACECRAFT', fmt='ET')

    assert len(ets) == 6
    assert_array_equal(ets, [
        [783500680, 783870237],
        [880011923, 880901281],
        [942003659, 942745795],
        [975964826, 976605399],
        [94305664.183, 757080065.183],
        [707313600, 1577880000],
    ])

    # UTC output format
    utc = SpicePool.windows('JUICE_SPACECRAFT')

    assert len(utc) == 6
    assert_array_equal(utc.astype(str), [
        ['2024-10-29T19:03:30.818', '2024-11-03T01:42:47.817'],
        ['2027-11-20T19:44:13.817', '2027-12-01T02:46:51.817'],
        ['2029-11-07T07:39:49.817', '2029-11-15T21:48:45.817'],
        ['2030-12-05T09:19:16.817', '2030-12-12T19:15:29.817'],
        ['2002-12-27T23:59:59.999', '2023-12-28T23:59:55.999'],
        ['2022-05-31T23:58:50.815', '2049-12-31T23:58:50.816'],
    ])

    # TDB output format
    tdb = SpicePool.windows('JUICE_SPACECRAFT', fmt='TDB')

    assert len(tdb) == 6
    assert_array_equal(tdb, [
        ['2024-10-29 19:04:40.000 TDB', '2024-11-03 01:43:57.000 TDB'],
        ['2027-11-20 19:45:23.000 TDB', '2027-12-01 02:48:01.000 TDB'],
        ['2029-11-07 07:40:59.000 TDB', '2029-11-15 21:49:55.000 TDB'],
        ['2030-12-05 09:20:26.000 TDB', '2030-12-12 19:16:39.000 TDB'],
        ['2002-12-28 00:01:04.182 TDB', '2023-12-29 00:01:05.182 TDB'],
        ['2022-06-01 00:00:00.000 TDB', '2050-01-01 00:00:00.000 TDB'],
    ])

    # Invalid coverage boundaries (no overlap)
    with raises(ValueError):
        _ = SpicePool.coverage('JUICE_SPACECRAFT')

    # Remove CK values
    monkeypatch.setattr(sp, 'ckcov', lambda *args: [])

    # Check new windows
    ets = SpicePool.windows('JUICE_SPACECRAFT', fmt='ET')

    assert len(ets) == 2
    assert_array_equal(ets, [
        [94305664.183, 757080065.183],
        [707313600, 1577880000],
    ])

    # Take only the coverage in ET
    ets = SpicePool.coverage('JUICE_SPACECRAFT', fmt='ET')

    assert len(ets) == 2
    assert_array_equal(ets, [707313600, 757080065.183])

    # Take only the coverage in UTC
    utc = SpicePool.coverage('JUICE_SPACECRAFT', fmt='UTC')

    assert len(utc) == 2
    assert_array_equal(utc, (
        np.datetime64('2022-05-31T23:58:50.815'),
        np.datetime64('2023-12-28T23:59:55.999'),
    ))

    # Take only the coverage in TDB
    tdb = SpicePool.coverage('JUICE_SPACECRAFT', fmt='TDB')

    assert len(tdb) == 2
    assert_array_equal(tdb, [
        '2022-06-01 00:00:00.000 TDB', '2023-12-29 00:01:05.182 TDB'])

    # Invalid output format
    with raises(TypeError):
        _ = SpicePool.windows('JUICE_SPACECRAFT', fmt='INVALID_FMT')

    with raises(TypeError):
        _ = SpicePool.coverage('JUICE_SPACECRAFT', fmt='INVALID_FMT')

    # No windows
    monkeypatch.setattr(sp, 'ckobj', lambda *args: [])
    monkeypatch.setattr(sp, 'cell_int', lambda *args: [])
    monkeypatch.setattr(sp, 'spkobj', lambda *args: [])

    with raises(KeyError):
        _ = SpicePool.windows('JUICE_SPACECRAFT')


def test_spice_load_mk():
    """Test metakernel loader."""
    SpicePool.add(mk, purge=True)

    assert len(SpicePool) == 3

    # Check metakernel cache hash
    assert hash(mk) in SpicePool.MK_HASH.values()

    # Check kernel contains in the pool
    assert mk in SpicePool
    assert KERNELS + '/naif0012.tls' in SpicePool
    assert KERNELS + '/dummy.tf' in SpicePool

    # Check that all the kernels in the mk where loaded
    kernels = SpicePool.kernels

    assert kernels[0].endswith('.tm')
    assert 'metakernel' in kernels[0]
    assert 'naif0012.tls' in kernels[1]
    assert 'dummy.tf' in kernels[2]

    # Check pool content
    assert hash(SpicePool) == mk_kernels_hash

    # Empty the pool
    SpicePool.purge()
    assert hash(SpicePool) != mk_kernels_hash

    assert len(SpicePool) == 0
    assert not SpicePool.MK_HASH

    # Get metakernel hash (independent of the pool content)
    assert SpicePool.hash(mk) == mk_kernels_hash


def test_spice_pool_check_kernels(caplog):
    """Test Spice Pool kernel checker decorator."""
    debug_spice_pool(True)
    SpicePool.purge()
    caplog.clear()

    class A:
        """Dummy class with kernels and hash."""
        @property
        def kernels(self):
            """Object required kernels."""
            return (kernel, )

        def __hash__(self):
            return SpicePool.hash(self.kernels)

        @check_kernels
        def get_first_kernel(self):
            """Decorated dummy function that require the kernels to be loaded."""
            return self.kernels[0]

        @property
        @check_kernels
        def last_kernel(self):
            """Decorated dummy property that require the kernels to be loaded."""
            return self.kernels[-1]

    a = A()

    # Check the attributes
    assert a.kernels == (kernel, )
    assert hash(a) == hash((hash(kernel),))

    # Check the pool content (the kernels are not loaded)
    assert len(SpicePool) == 0
    assert not caplog.messages

    # When the decorated function is call, the pool is checked and kernels/mk are loaded.
    assert a.get_first_kernel() == kernel

    reload_msg = [
        'The content of the pool changed -> the kernels will be reloaded.',
        'Purge the pool',
        f'Add `{kernel}` in the SPICE pool',
    ]

    assert caplog.messages == reload_msg

    # Check the pool content
    assert len(SpicePool) == 1
    assert SpicePool.kernels == (kernel, )

    # Re-run the same function but this time the pool match the kernels
    caplog.clear()

    assert a.get_first_kernel() == kernel
    assert not caplog.messages

    # Checker on property (no need to reload the pool)
    assert a.last_kernel == kernel
    assert not caplog.messages

    # Class without kernels attribute
    class B:
        """Dummy class without explicit hash."""

        @check_kernels
        def dummy(self):  # pylint: disable=no-self-use
            """Decorated dummy function."""

    b = B()

    with raises(AttributeError):
        b.dummy()
