"""Test SPICE reference module."""

from pathlib import Path

import spiceypy as sp

from pytest import raises

from moon_coverage.spice import (
    SpiceBody, SpiceInstrument, SpiceObserver, SpiceRef, SpiceSpacecraft
)


KERNELS = str((Path(__file__).parent / '..' / 'data').resolve())
tf = KERNELS + '/dummy.tf'
ti = KERNELS + '/dummy.ti'


def test_body_code_validity():
    """Test SPICE body code validity."""
    assert SpiceRef.is_valid(-1)
    assert SpiceRef.is_valid(1)

    with raises(ValueError):
        _ = SpiceRef.is_valid('FOO')


def test_reference_attrs():
    """Test SPICE reference attributes."""
    sp.kclear()

    # Reference with string
    ref = SpiceRef('Jupiter Barycenter')

    assert str(ref) == ref.name == 'JUPITER BARYCENTER'
    assert int(ref) == ref.id == 5
    assert ref.code == '5'

    # Reference with positive integer
    ref = SpiceRef(5)

    assert str(ref) == 'JUPITER BARYCENTER'
    assert int(ref) == 5
    assert ref.code == '5'

    # Reference with negative integer
    ref = SpiceRef(-28)

    assert str(ref) == 'JUICE'
    assert int(ref) == -28

    # __eq__
    assert ref == 'JUICE'
    assert ref == -28

    assert ref != 'Juice'  # Case-sensitive
    assert ref != 28

    # __getitem__
    with raises(KeyError):
        _ = ref['DUMMY']

    # Nested references
    assert SpiceRef(SpiceRef('JUICE')) == -28
    assert SpiceRef(SpiceRef(-28)) == 'JUICE'

    # Not implemented errors
    ref = SpiceRef('Jupiter Barycenter')

    with raises(NotImplementedError):
        _ = ref.spacecraft

    with raises(NotImplementedError):
        _ = ref.sclk('2035-01-01')

    # Invalid SPICE reference
    with raises(ValueError):
        _ = SpiceRef('FOO')


def test_reference_promotion():
    """Test SPICE reference promotion."""
    sp.kclear()
    sp.furnsh([tf, ti])

    assert isinstance(SpiceRef('Jupiter Barycenter'), SpiceRef)
    assert isinstance(SpiceRef('Ganymede'), SpiceBody)
    assert isinstance(SpiceRef('JUICE'), SpiceSpacecraft)
    assert isinstance(SpiceRef('JUICE_JANUS'), SpiceInstrument)

    # Spice observer parent class
    assert isinstance(SpiceRef('JUICE'), SpiceObserver)
    assert isinstance(SpiceRef('JUICE_JANUS'), SpiceObserver)

    # Spice observer promotion
    assert isinstance(SpiceObserver('JUICE'), SpiceSpacecraft)
    assert isinstance(SpiceObserver('JUICE_JANUS'), SpiceInstrument)

    with raises(KeyError):
        _ = SpiceObserver('Ganymede')
