"""Test ticks module."""

from moon_coverage.ticks import (
    date_ticks, deg_ticks, hr_ticks, km_pix_ticks, km_s_ticks, km_ticks,
    lat_ticks, lon_e_ticks, lon_w_ticks, m_pix_ticks, m_s_ticks
)
from moon_coverage.ticks.ticks import UnitFormatter


def test_ticks_unit_fmt():
    """Test unit ticker formatter."""
    t = UnitFormatter()
    assert t(1) == '1'
    assert t(1.0) == '1'
    assert t(1.0001) == '1.0001'
    assert t(1.000000001) == '1'
    assert t(10) == '10'
    assert t(10.01) == '10.01'
    assert t(10_000) == '10,000'
    assert t(1_000_000) == '1e+06'
    assert t(1_000_001) == '1e+06'
    assert t(1_000_010) == '1.00001e+06'

    t = UnitFormatter('abc')
    assert t(1) == '1 abc'
    assert t(1.0) == '1 abc'
    assert t(10.01) == '10.01 abc'
    assert t(10_001) == '10,001 abc'

    t = UnitFormatter('"', sep='')
    assert t(1) == '1"'
    assert t(1.0) == '1"'
    assert t(10.01) == '10.01"'
    assert t(10_001) == '10,001"'

    # The separator is not applied if the unit is not defined
    t = UnitFormatter(sep='--')
    assert t(1) == '1'

    # Ticks explicit formatter
    t = UnitFormatter(fmt='.03f')
    assert t(1) == '1.000'


def test_ticks_scaling_factor():
    """Test unit ticker formatter with scaling factor"""
    t = UnitFormatter(unit='m', scale=1_000)
    assert t(1) == '1,000 m'
    assert t(1.001) == '1,001 m'
    assert t(1_001) == '1.001e+06 m'

    t = UnitFormatter(unit='km', scale=0.001)
    assert t(1) == '0.001 km'
    assert t(1.001) == '0.001001 km'
    assert t(1_001) == '1.001 km'

    t_times_2 = UnitFormatter(scale=2)
    assert t_times_2(1) == '2'

    t_times_4 = t_times_2 * 2
    assert t_times_4(1) == '4'

    t_times_12 = 3 * t_times_4
    assert t_times_12(1) == '12'

    t_times_6 = t_times_12 / 2
    assert t_times_6(1) == '6'

    t_neg_6 = -t_times_6
    assert t_neg_6(1) == '−6'
    assert t_neg_6(-1) == '6'

    t_plus_1 = UnitFormatter(offset=1)
    assert t_plus_1(1) == '2'

    t_plus_2 = t_plus_1 + 1
    assert t_plus_2(1) == '3'

    t_plus_4 = 2 + t_plus_2
    assert t_plus_4(1) == '5'

    t_minus_2 = t_plus_4 - 6
    assert t_minus_2(1) == '−1'

    t_linear = 5 - 3 * t_times_2
    assert t_linear(0) == '5'
    assert t_linear(1) == '−1'
    assert t_linear(2) == '−7'
    assert t_linear(-1) == '11'

    # Reverse ticks location
    assert t_linear @ 11 == -1
    assert t_linear @ [5, -1, -7, 11] == [0, 1, 2, -1]


def test_ticks_units():
    """Test units ticker."""
    assert deg_ticks(1.0) == '1°'
    assert hr_ticks(1.0) == '1 h'
    assert km_ticks(1.0) == '1 km'
    assert km_s_ticks(1.0) == '1 km/s'
    assert m_s_ticks(1.0) == '1,000 m/s'
    assert km_pix_ticks(1.0) == '1 km/pix'
    assert m_pix_ticks(1.0) == '1,000 m/pix'

    assert deg_ticks(1.001) == '1.001°'
    assert hr_ticks(1.001) == '1.001 h'
    assert km_ticks(1.001) == '1.001 km'
    assert km_s_ticks(1.001) == '1.001 km/s'
    assert m_s_ticks(1.001) == '1,001 m/s'
    assert km_pix_ticks(1.001) == '1.001 km/pix'
    assert m_pix_ticks(1.001) == '1,001 m/pix'

    assert deg_ticks(1_001) == '1,001°'
    assert hr_ticks(1_001) == '1,001 h'
    assert km_ticks(1_001) == '1,001 km'
    assert km_s_ticks(1_001) == '1,001 km/s'
    assert m_s_ticks(1_001) == '1.001e+06 m/s'
    assert km_pix_ticks(1_001) == '1,001 km/pix'
    assert m_pix_ticks(1_001) == '1,001,000 m/pix'


def test_ticks_lat():
    """Test latitude ticks."""
    assert lat_ticks(90) == 'N.P.'
    assert lat_ticks(0) == 'Eq.'
    assert lat_ticks(-90) == 'S.P.'

    # With INT
    assert lat_ticks(10) == '10°N'
    assert lat_ticks(-10) == '10°S'

    # With FLOAT
    assert lat_ticks(10.1) == '10.1°N'
    assert lat_ticks(-10.1) == '10.1°S'


def test_ticks_lon_e():
    """Test East longitude ticks."""
    assert lon_e_ticks(0) == '0°'
    assert lon_e_ticks(180) == '180°'
    assert lon_e_ticks(360) == '0°'
    assert lon_e_ticks(-180) == '180°'

    # With INT
    assert lon_e_ticks(10) == '10°E'
    assert lon_e_ticks(190) == '190°E'
    assert lon_e_ticks(-10) == '350°E'
    assert lon_e_ticks(-90) == '270°E'

    # With FLOAT
    assert lon_e_ticks(10.1) == '10.1°E'
    assert lon_e_ticks(-10.1) == '349.9°E'


def test_ticks_lon_w():
    """Test West longitude ticks."""
    assert lon_w_ticks(0) == '0°'
    assert lon_w_ticks(180) == '180°'
    assert lon_w_ticks(360) == '0°'
    assert lon_w_ticks(-180) == '180°'

    # With INT
    assert lon_w_ticks(10) == '10°W'
    assert lon_w_ticks(190) == '190°W'
    assert lon_w_ticks(-10) == '350°W'
    assert lon_w_ticks(-90) == '270°W'

    # With FLOAT
    assert lon_w_ticks(10.1) == '10.1°W'
    assert lon_w_ticks(-10.1) == '349.9°W'


def test_ticks_date():
    """Test compressed date ticks."""
    assert date_ticks(0) == '1970'
    assert date_ticks(365) == '1971'
