"""Groundtrack trajectory module."""

from pathlib import Path

import numpy as np

from pytest import raises

import moon_coverage.trajectory.config as tc
from moon_coverage.events import (
    EventsDict, EventsFile, EventsList, EventWindow
)
from moon_coverage.spice import MetaKernel, SpicePool
from moon_coverage.trajectory import TourConfig
from moon_coverage.trajectory.config import (
    ESA_CREMAS, KernelNotFoundError, KernelRemoteNotFoundError,
    KernelsDirectoryNotFoundError, MissingKernelsRemote
)


DATA = (Path(__file__).parent / '..' / 'data').resolve()
MK = DATA / 'metakernel.tm'
MK_HASH = hash((str(DATA / 'naif0012.tls'), str(DATA / 'dummy.tf')))
PHASES = DATA / 'mission_phases.csv'
TIMELINE = DATA / 'mission_timeline.csv'


def test_tour_config_attr():
    """Test tour configuration attributes."""
    SpicePool.purge()

    # Check optional attributes
    tour = TourConfig(
        mk=MK,
        kernels=(DATA / 'dummy.ti'),
        kernels_dir=DATA,
        spacecraft='foo',
        instrument='bar',
        target='baz',
        abcorr='CN+S',
        default_time_step='2 sec'
    )

    assert tour.spacecraft == 'FOO'
    assert tour.instrument == 'BAR'
    assert tour.target == 'BAZ'
    assert tour.default_time_step == '2 sec'
    assert tour.abcorr == 'CN+S'

    assert isinstance(tour.mk, MetaKernel)
    assert tour.mk.fname.name == 'metakernel.tm'               # pylint: disable=no-member
    assert hash(tour.mk) == MK_HASH

    assert isinstance(tour.metakernel, Path)
    assert tour.metakernel == MK

    assert tour.kernels == (
        # Only the explicit mk and the explicit kernels are listed in the kernels to load.
        MetaKernel(DATA / 'metakernel.tm', kernels=DATA),
        str(DATA / 'dummy.ti'),
    )

    # The hash pre-calculated before being loaded in the pool
    assert hash(tour) == hash((
        MK_HASH,
        hash(str(DATA / 'naif0012.tls')),  # 1st kernel in mk (part of the hash)
        hash(str(DATA / 'dummy.tf')),      # 2nd kernel in mk
        hash(str(DATA / 'dummy.ti')),      # Explicit kernel
    ))

    # Check the Spice Pool content
    assert len(SpicePool) == 0
    assert SpicePool != hash(tour)

    # The kernels are only loaded if a SPICE calculation require them
    assert tour.skd_version == 'v999_20210531_001'

    assert len(SpicePool) == 4
    assert SpicePool == hash(tour)

    # Reload the kernel if the pool have changed
    SpicePool.purge()
    assert tour.skd_version == 'v999_20210531_001'

    assert len(SpicePool) == 4
    assert SpicePool == hash(tour)

    # Invalid aberation correction
    with raises(KeyError):
        _ = TourConfig(mk=MK, kernels_dir=DATA, abcorr='FOO')


def test_tour_config_crema(monkeypatch):
    """Test tour CReMA loader."""
    monkeypatch.setitem(ESA_CREMAS, 'JUICE', {('3.1', 'latest'): MK})

    tour = TourConfig(
        mk='3.1',
        kernels_dir=DATA,
        spacecraft='JUICE',
    )

    assert tour.metakernel == MK
    assert tour.kernels == (
        MetaKernel(DATA / 'metakernel.tm', kernels=DATA),
    )

    # Check default optionals
    assert tour.default_time_step == '1 minute'
    assert tour.abcorr == 'NONE'


def test_tour_config_mk(monkeypatch):
    """Test tour configuration an explicit metakernel."""
    tour = TourConfig(mk=MetaKernel(MK, kernels=DATA))

    assert hash(tour.mk) == MK_HASH
    assert tour.metakernel == MK

    # Use a spacecraft name with a space
    monkeypatch.setenv('KERNELS_EUROPA_CLIPPER', str(DATA))

    tour = TourConfig(mk=MK, spacecraft='EUROPA CLIPPER')

    assert hash(tour.mk) == MK_HASH
    assert tour.metakernel == MK


def test_tour_config_mk_remote():
    """Test tour configuration metakernel with a custom remote."""
    tour = TourConfig(mk=MK, kernels_dir=DATA,
                      remote_kernels='https://example.com/kernels')

    assert hash(tour.mk) == MK_HASH
    assert tour.metakernel == MK
    assert tour.mk.remote == 'https://example.com/kernels'  # pylint: disable=no-member


def test_tour_config_kernels():
    """Test tour configuration multiple kernels and no metakernel."""
    tour = TourConfig(
        mk=None,
        kernels=[DATA / 'naif0012.tls', DATA / 'dummy.tf'],
    )

    assert tour.mk is None
    assert tour.metakernel is None
    assert tour.skd_version is None

    assert tour.kernels == (
        str(DATA / 'naif0012.tls'),
        str(DATA / 'dummy.tf'),
    )

    assert hash(tour) == hash((
        hash(str(DATA / 'naif0012.tls')),
        hash(str(DATA / 'dummy.tf')),
    ))


def test_tour_config_kernels_errors(monkeypatch):
    """Test tour configuration errors."""
    # The Spacecraft is not in ESA CReMA
    with raises(ValueError):
        _ = TourConfig(mk='3.0', spacecraft='JUNO')

    # Remove a kernel from the test folder (MissingKernel errors)
    tf = DATA / 'dummy.tf'

    # Cache the content and remove on of the file of the metakernel
    content = tf.read_text()
    tf.unlink()
    assert not tf.exists()

    # No `kernel_dir` value and KERNELS_DUMMY_SPACECRAFT does not exists
    with raises(KernelsDirectoryNotFoundError):
        _ = TourConfig(mk=MK, spacecraft='Dummy Spacecraft')

    # Provide an explicit kernels_dir` but `tf` is still missing
    with raises(KernelNotFoundError):
        _ = TourConfig(mk=MK, kernels_dir=DATA, spacecraft='JUICE')

    # Set KERNELS_JUICE as an env variable (but `tf` is still missing)
    monkeypatch.setenv('KERNELS_JUICE', str(DATA))

    with raises(KernelNotFoundError):
        _ = TourConfig(mk=MK, spacecraft='JUICE')

    # Re-create the deleted file
    tf.write_text(content)

    # Invalid `kernels_dir` folder
    with raises(KernelNotFoundError):
        _ = TourConfig(mk=MK, kernels_dir='invalid/path')

    # Create dummy MetaKernel raising `MissingKernelsRemote`
    class NoRemoteMetakernel:
        """Raise MissingKernelsRemote on MetaKernel object."""
        def __init__(self, _, **kwargs):
            raise MissingKernelsRemote('dummy_kernel.tf')

    monkeypatch.setattr(tc, 'MetaKernel', NoRemoteMetakernel)

    with raises(KernelRemoteNotFoundError):
        _ = TourConfig(mk=MK, kernels_dir=DATA)


def test_tour_config_get_traj(monkeypatch):
    """Test tour from trajectory configuration."""
    tour = TourConfig(mk=None, kernels=DATA / 'naif0012.tls')

    # Mock SPICE pool coverage (mk=3.0)
    monkeypatch.setattr(SpicePool, 'coverage', lambda *args: (
        np.datetime64('2022-05-31T23:58:51.815'),
        np.datetime64('2033-06-04T02:49:52.362'),
    ))

    assert tour.mk is None
    assert tour.kernels == (str(DATA / 'naif0012.tls'), )

    # Trajectory from a single date
    traj = tour['2033-01-01']

    assert traj.observer == 'JUICE'
    assert traj.target == 'GANYMEDE'
    assert traj.abcorr == 'NONE'

    assert traj.start == traj.stop == np.datetime64('2033-01-01')
    assert len(traj) == 1

    # Trajectory from numpy datetime64
    traj = tour[np.datetime64('2033-01-01')]

    assert traj.start == traj.stop == np.datetime64('2033-01-01')
    assert len(traj) == 1

    # Trajectory from multiple date (unsorted)
    traj = tour['2033-01-01', '2033-01-03', '2033-01-02']

    assert traj.start == np.datetime64('2033-01-01')
    assert traj.stop == np.datetime64('2033-01-03')
    assert len(traj) == 3

    # Trajectory with a range of date (with a temporal step)
    traj = tour['2033-01-01':'2033-01-02':'2h']

    assert traj.start == np.datetime64('2033-01-01')
    assert traj.stop == np.datetime64('2033-01-02')
    assert len(traj) == 13

    # Trajectory with a range of date (with a number of points)
    traj = tour['2033-01-01':'2033-01-02':10]

    assert traj.start == np.datetime64('2033-01-01')
    assert traj.stop == np.datetime64('2033-01-02')
    assert len(traj) == 10

    # Trajectory without a defined temporal step (default_step: `1 minute`)
    traj = tour['2033-01-01':'2033-01-02']

    assert traj.start == np.datetime64('2033-01-01')
    assert traj.stop == np.datetime64('2033-01-02')
    assert len(traj) == 1_441

    # Trajectory with an exotic list of times and slices
    traj = tour[
        '2033-01-04',
        '2033-JAN-02':'2033-01-03':'1h',  # 25 points
        (1041422469.184,),                # 2033-01-01

    ]

    assert traj.start == np.datetime64('2033-01-01')
    assert traj.stop == np.datetime64('2033-01-04')
    assert len(traj) == 27

    # Complete tour
    traj = tour['all']

    assert traj.start == np.datetime64('2022-05-31T23:58:51.816')
    assert traj.stop == np.datetime64('2033-06-04T02:49:52.361')
    assert len(traj) == 193_015

    # Complete tour with `None` (same as 'all')
    traj = tour[None]

    assert traj.start == np.datetime64('2022-05-31T23:58:51.816')
    assert traj.stop == np.datetime64('2033-06-04T02:49:52.361')
    assert len(traj) == 193_015

    # Complete tour with only step
    traj = tour[::'1 year']

    assert traj.start == np.datetime64('2022-05-31T23:58:51.816')
    assert traj.stop == np.datetime64('2033-06-04T02:49:52.361')
    assert len(traj) == 13

    # Sub-section of the tour from start
    traj = tour['start':'2025-01-01':'1h']

    assert traj.start == np.datetime64('2022-05-31T23:58:51.816')
    assert traj.stop == np.datetime64('2025-01-01')
    assert len(traj) == 22_682

    # Sub-section of the tour from start
    traj = tour[:'2025-01-01':'1h']

    assert traj.start == np.datetime64('2022-05-31T23:58:51.816')
    assert traj.stop == np.datetime64('2025-01-01')
    assert len(traj) == 22_682

    # Sub-section of the tour up to the end
    traj = tour['2025-01-01':'end':'1 day']

    assert traj.start == np.datetime64('2025-01-01')
    assert traj.stop == np.datetime64('2033-06-04T02:49:52.361')
    assert len(traj) == 3_078

    # Sub-section of the tour up to the end
    traj = tour['2025-01-01'::'1y']

    assert traj.start == np.datetime64('2025-01-01')
    assert traj.stop == np.datetime64('2033-06-04T02:49:52.361')
    assert len(traj) == 10


def test_tour_config_get_traj_instr():
    """Test tour from trajectory configuration for an instrument."""
    SpicePool.purge()

    tour = TourConfig(
        mk=None,
        kernels=[
            str(DATA / 'naif0012.tls'),
            str(DATA / 'dummy.tf'),
            str(DATA / 'dummy.ti'),
        ],
        spacecraft='JUICE',
        instrument='JANUS',
    )

    # Trajectory from a single date
    traj = tour['2033-01-01']

    assert traj.observer == 'JUICE_JANUS'
    assert traj.target == 'GANYMEDE'

    # Missing instrument kernel
    SpicePool.purge()

    tour = TourConfig(
        mk=None,
        kernels=str(DATA / 'naif0012.tls'),
        kernels_dir=DATA,
        spacecraft='JUICE',
        instrument='MAJIS',
    )

    with raises(ValueError):
        _ = tour['2033-01-01']


def test_tour_config_events():
    """Test tour configuration with input events."""
    tour = TourConfig(
        mk=None,
        kernels=DATA / 'naif0012.tls',
        default_time_step='1h'
    )

    phases = EventsFile(PHASES)
    events = EventsFile(TIMELINE, primary_key='Event name')
    flybys = events['FLYBY_GANYMEDE']

    # Select event (single time)
    flyby_1G1 = flybys['1G1']
    traj = tour[flyby_1G1]

    assert len(traj) == 1
    assert traj.utc == [np.datetime64('2031-07-21T07:14:23')]

    # Multiple events inputs
    flyby_2G2 = flybys['2G2']
    traj = tour[flyby_1G1, flyby_2G2]

    assert len(traj) == 2
    assert traj.utc[0] == np.datetime64('2031-07-21T07:14:23')
    assert traj.utc[1] == np.datetime64('2032-02-13T23:04:51')

    # Events ranges with an explicit time step
    traj = tour[flyby_1G1:flyby_2G2:'1 day']

    assert len(traj) == 209
    assert traj.utc[0] == np.datetime64('2031-07-21T07:14:23')
    assert traj.utc[-1] == np.datetime64('2032-02-13T23:04:51')

    # Select event window without an explicit time step
    gco500 = phases['GCO500']
    traj = tour[gco500]

    assert len(traj) == 3_276  # default time step (1h here)
    assert traj.utc[0] == np.datetime64('2035-05-22T06')
    assert traj.utc[-1] == np.datetime64('2035-10-05T16:01')

    # Select event window with an explicit time step
    traj = tour[gco500:100]

    assert len(traj) == 100
    assert traj.utc[0] == np.datetime64('2035-05-22T06')
    assert traj.utc[-1] == np.datetime64('2035-10-05T16:01')

    # List of Event and EventWindow
    traj = tour[flyby_1G1, gco500:100]

    assert len(traj) == 101
    assert traj.utc[0] == np.datetime64('2031-07-21T07:14:23')
    assert traj.utc[1] == np.datetime64('2035-05-22T06')
    assert traj.utc[-1] == np.datetime64('2035-10-05T16:01')

    # With an EventsList
    traj = tour[flybys]  # 1G1 and 2G2 only here

    assert len(traj) == 2
    assert traj.utc[0] == np.datetime64('2031-07-21T07:14:23')
    assert traj.utc[1] == np.datetime64('2032-02-13T23:04:51')

    # With event temporal operations
    traj = tour[flyby_1G1 - '12 h':flyby_1G1 + '12 h':'10 min']

    assert len(traj) == 24 * 6 + 1
    assert traj.utc[0] == np.datetime64('2031-07-20T19:14:23')
    assert traj.utc[-1] == np.datetime64('2031-07-21T19:14:23')

    # With an EventsDict/EventsFile
    tour.default_time_step = '10 days'
    traj = tour[phases]  # All phases (1 layer of EventWindow objects)

    assert len(traj) == 259
    assert traj.utc[0] == np.datetime64('2031-01-19T18:50:41')
    assert traj.utc[-1] == np.datetime64('2035-10-05T16:01:00')

    traj = tour[events]  # All timeline (multi-layers)

    assert len(traj) == 27
    assert traj.utc[0] == np.datetime64('2031-01-20T06:31:08')
    assert traj.utc[-1] == np.datetime64('2032-02-15T06:07:11')

    # With an EventsDict with a slice
    traj = tour[events:'1 h']

    assert len(traj) == 119
    assert traj.utc[0] == np.datetime64('2031-01-20T06:31:08')
    assert traj.utc[-1] == np.datetime64('2032-02-15T06:07:11')


def test_tour_config_event_exclusion():
    """Test tour configuration with event exclusion."""
    # With a single EventWindow
    event_window = EventWindow('dummy', t_start='2033-01-01T12', t_end='2033-01-02T12')

    tour = TourConfig(
        mk=None,
        kernels=DATA / 'naif0012.tls',
        exclude=event_window,
    )

    traj = tour['2033-01-01':'2033-01-03':'1 h']

    assert len(traj) == 24
    assert traj.start == np.datetime64('2033-01-01')
    assert traj.stop == np.datetime64('2033-01-03')

    assert np.datetime64('2033-01-01T11') in traj.utc
    assert np.datetime64('2033-01-02T13') in traj.utc
    assert np.datetime64('2033-01-01T12') not in traj.utc
    assert np.datetime64('2033-01-02T12') not in traj.utc

    # With an EventsList
    events_list = EventsList([event_window])

    tour = TourConfig(
        mk=None,
        kernels=DATA / 'naif0012.tls',
        exclude=events_list,
    )

    traj = tour['2033-01-01':'2033-01-03':'1 h']

    assert len(traj) == 24
    assert traj.start == np.datetime64('2033-01-01')
    assert traj.stop == np.datetime64('2033-01-03')

    assert np.datetime64('2033-01-01T11') in traj.utc
    assert np.datetime64('2033-01-02T13') in traj.utc
    assert np.datetime64('2033-01-01T12') not in traj.utc
    assert np.datetime64('2033-01-02T12') not in traj.utc

    # With an EventsDict / EventsFile
    events_dict = EventsDict([event_window])

    tour = TourConfig(
        mk=None,
        kernels=DATA / 'naif0012.tls',
        exclude=events_dict,
    )

    traj = tour['2033-01-01':'2033-01-03':'1 h']

    assert len(traj) == 24
    assert traj.start == np.datetime64('2033-01-01')
    assert traj.stop == np.datetime64('2033-01-03')

    assert np.datetime64('2033-01-01T11') in traj.utc
    assert np.datetime64('2033-01-02T13') in traj.utc
    assert np.datetime64('2033-01-01T12') not in traj.utc
    assert np.datetime64('2033-01-02T12') not in traj.utc
