"""Test trajectory instrument field of view module."""

from itertools import count
from pathlib import Path

import numpy as np
from numpy.testing import assert_array_almost_equal as assert_array

from matplotlib.collections import PathCollection
from matplotlib.path import Path as MPLPath

from pytest import fixture, raises

import moon_coverage.spice.toolbox as toolbox
import moon_coverage.trajectory.trajectory as tr
from moon_coverage import Trajectory
from moon_coverage.trajectory import InstrumentTrajectory
from moon_coverage.trajectory.fovs import FovsCollection, MaskedFovsCollection


DATA = (Path(__file__).parent / '..' / 'data').resolve()
KERNELS = [
    str(DATA / 'naif0012.tls'),
    str(DATA / 'dummy.tf'),
    str(DATA / 'dummy.ti'),
    str(DATA / 'dummy.tpc'),
]


@fixture
def traj(monkeypatch):
    """Dummy trajectory."""
    counter = count()

    def intersect_pt(times, *args, **kwargs):
        n = len(times)
        return np.transpose([
            [i, c // n, c // n]
            for i, c in zip(range(1, n + 1), counter)
        ])

    monkeypatch.setattr(toolbox, 'intersect_pt', intersect_pt)

    return Trajectory(
        KERNELS,
        'JUICE_JANUS',
        'Ganymede',
        '2033-01-01',
    )


@fixture
def fovs(traj):
    """Default FOVs collection."""
    return FovsCollection(traj, npts=5)


def test_traj_fovs_attr(fovs):
    """Test FOVs collection attributes."""
    assert fovs.observer == 'JUICE_JANUS'
    assert fovs.target == 'GANYMEDE'
    assert fovs.npts == 5
    assert fovs.limb

    assert fovs.kernels == KERNELS

    assert np.shape(fovs.pts) == (3, 1, 5)
    assert_array(fovs.pts, np.array([
        [[1, 1, 1, 1, 1]],
        [[0, 1, 2, 3, 0]],
        [[0, 1, 2, 3, 0]],
    ]))

    assert np.shape(fovs.rlonlat) == (3, 1, 5)
    assert_array(fovs.rlonlat, np.array([
        [[1, np.sqrt(3), 3, np.sqrt(19), 1]],
        [[0, 45, 63.43, 71.56, 0]],
        [[0, 35.26, 41.81, 43.49, 0]],
    ]), decimal=2)


def test_traj_fovs_paths_inside(fovs):
    """Test FOVs paths (all inside)."""
    assert len(fovs.paths) == 1
    assert isinstance(fovs.paths, list)
    assert isinstance(fovs.paths[0], MPLPath)
    assert_array(fovs.paths[0].vertices.T, [
        [0, 45, 63.43, 71.56, 0],
        [0, 35.26, 41.81, 43.49, 0],
    ], decimal=2)


def test_traj_fovs_paths_1_out(monkeypatch, fovs):
    """Test FOVs paths (one outside)."""
    @property
    def target_radii_3(*args):
        return (3, 3, 3)

    monkeypatch.setattr(tr.SpiceBody, 'radii', target_radii_3)

    assert fovs.target.radii == (3, 3, 3)

    # Check that the pts where not changed
    assert np.shape(fovs.pts) == (3, 1, 5)
    assert_array(fovs.pts, np.array([
        [[1, 1, 1, 1, 1]],
        [[0, 1, 2, 3, 0]],
        [[0, 1, 2, 3, 0]],
    ]))

    assert len(fovs.paths) == 1
    assert isinstance(fovs.paths[0], MPLPath)
    assert_array(fovs.paths[0].vertices.T, [
        [0, 45, 63.43, 71.56, 0],
        [0, 35.26, 41.81, 43.49, 0],
    ], decimal=2)


def test_traj_fovs_paths_all_out(monkeypatch, fovs):
    """Test FOVs paths (all outside)."""
    @property
    def target_radii_05(*args):
        return (.5, .5, .5)

    monkeypatch.setattr(tr.SpiceBody, 'radii', target_radii_05)

    assert fovs.target.radii == (.5, .5, .5)

    # Check that the pts where not changed
    assert np.shape(fovs.pts) == (3, 1, 5)
    assert_array(fovs.pts, np.array([
        [[1, 1, 1, 1, 1]],
        [[0, 1, 2, 3, 0]],
        [[0, 1, 2, 3, 0]],
    ]))

    assert len(fovs.paths) == 1
    assert fovs.paths[0] is None

    # FOV path collection
    assert isinstance(fovs(), PathCollection)

    assert len(fovs().get_paths()) == 1
    assert fovs().get_paths()[0] is None


def test_traj_fovs_collection(monkeypatch, fovs):
    """Test FOVs collection."""
    monkeypatch.setattr(InstrumentTrajectory, 'inc', [35.26])

    # Default
    coll = fovs()

    assert coll.get_label() == 'JUICE JANUS'
    assert_array(coll.get_facecolor(), np.empty((0, 4)))
    assert_array(coll.get_edgecolor(), [(0, 0, 0, 1)])  # black contour

    # Alternative label and fix colors
    coll = fovs(label='JANUS FOV', facecolors='tab:orange', edgecolors='C0')

    assert coll.get_label() == 'JANUS FOV'
    assert_array(coll.get_facecolor(), [(1, 0.498, 0.055, 1)], decimal=3)
    assert_array(coll.get_edgecolor(), [(0.122, 0.467, 0.706, 1)], decimal=3)

    # Collection colorized with a trajectory property
    coll = fovs('inc', cmap='gray')

    assert_array(coll.get_facecolor(), np.empty((0, 4)))
    assert_array(coll.get_edgecolor(), [(0, 0, 0, 1)])

    # Collection colorized with a trajectory property and values range
    coll = fovs(facecolors='inc', edgecolors='inc', vmin=0, vmax=90)

    assert_array(coll.get_facecolor(), [(0.899, 0.851, 0.22, 1)], decimal=3)
    assert_array(coll.get_edgecolor(), [(0.899, 0.851, 0.22, 1)], decimal=3)

    # Trajectory property error (not 1D)
    with raises(ValueError):
        _ = fovs('lonlat')


def test_traj_fovs_coll_getters(fovs):
    """Test FOVs collection behavior functions."""
    assert fovs.get_alpha() is None
    assert fovs.get_facecolors() == 'tab:orange'
    assert fovs.get_edgecolors() is None
    assert fovs.get_linewidth() == 1.5
    assert fovs.get_linestyle() == 'solid'
    assert fovs.get_zorder() == 1
    assert fovs.get_label() == ''


def test_traj_fovs_isort(monkeypatch, fovs):
    """Test FOVs indexes sorting."""
    monkeypatch.setattr(InstrumentTrajectory, 'emi', np.array([1, 3, 2, 5, 4]))
    monkeypatch.setattr(InstrumentTrajectory, 'inc', np.array([1, 3, 2, 5, 4]))

    ind = fovs.isort('emi')
    assert_array(ind, [0, 2, 1, 4, 3])
    assert_array(fovs.traj.emi[ind], [1, 2, 3, 4, 5])

    ind = fovs.isort('emi', reverse=True)
    assert_array(ind, [3, 4, 1, 2, 0])
    assert_array(fovs.traj.emi[ind], [5, 4, 3, 2, 1])

    # Order is reversed by default for `inc`, `dist` and `alt`
    ind = fovs.isort('inc')
    assert_array(ind, [3, 4, 1, 2, 0])
    assert_array(fovs.traj.inc[ind], [5, 4, 3, 2, 1])

    ind = fovs.isort('emi', reverse=False)
    assert_array(ind, [0, 2, 1, 4, 3])
    assert_array(fovs.traj.inc[ind], [1, 2, 3, 4, 5])

    with raises(KeyError):
        _ = fovs.isort('abc')


def test_traj_fovs_sort_collection(monkeypatch, traj):
    """Test FOVs sorting collection."""
    traj.ets = ['2033-01-01', '2033-01-02', '2033-01-03', '2033-01-04']

    fovs = FovsCollection(traj, npts=5)

    assert np.shape(fovs.pts) == (3, 4, 5)
    assert_array(fovs.pts[:, 0, :], np.array([
        [1, 1, 1, 1, 1],
        [0, 1, 2, 3, 0],
        [0, 1, 2, 3, 0],
    ]))
    assert_array(fovs.pts[:, 3, :], np.array([
        [4, 4, 4, 4, 4],
        [0, 1, 2, 3, 0],
        [0, 1, 2, 3, 0],
    ]))

    path_0 = [[0, 45, 63.43, 71.56, 0], [0, 35.26, 41.81, 43.49, 0]]
    path_3 = [[0, 14.04, 26.56, 36.87, 0], [0, 13.63, 24.09, 30.96, 0]]

    assert_array(fovs.paths[0].vertices.T, path_0, decimal=2)
    assert_array(fovs.paths[3].vertices.T, path_3, decimal=2)

    # Inject dummy `inc` values
    monkeypatch.setattr(InstrumentTrajectory, 'inc', [1, 3, 0, 2])

    # Check sorting values
    assert_array(fovs.isort('inc'), [1, 3, 0, 2])

    # No sort
    coll = fovs('inc', cmap='Reds_r')
    paths = coll.get_paths()

    assert_array(paths[0].vertices.T, path_0, decimal=2)
    assert_array(paths[3].vertices.T, path_3, decimal=2)

    assert_array(coll.get_facecolor(), np.empty((0, 4)))
    assert_array(coll.get_edgecolor()[..., 0], [.89, 1, .4, 0.98], decimal=2)  # Reds

    # Sort by `inc` (paths and colors)
    coll = fovs(facecolors='inc', edgecolors='inc', sort='inc', cmap='Reds_r')
    paths = coll.get_paths()

    assert_array(paths[2].vertices.T, path_0, decimal=2)
    assert_array(paths[1].vertices.T, path_3, decimal=2)

    assert_array(coll.get_facecolor()[..., 0], [1, 0.98, .89, .4], decimal=2)  # Reds
    assert_array(coll.get_edgecolor()[..., 0], [1, 0.98, .89, .4], decimal=2)  # Reds


def test_traj_fovs_masked(fovs):
    """Test FOVs masked collection."""
    masked_fovs = MaskedFovsCollection(fovs, [True])

    assert masked_fovs.mask == [True]
    assert masked_fovs.paths[0] is None
    assert masked_fovs().get_paths()[0] is None

    # Not masked fovs
    masked_fovs = MaskedFovsCollection(fovs, [False])

    assert masked_fovs.mask == [False]
    assert isinstance(masked_fovs.paths[0], MPLPath)
    assert isinstance(masked_fovs().get_paths()[0], MPLPath)
