"""Test trajectory properties calculation module."""

from pathlib import Path

import numpy as np
from numpy.testing import assert_array_almost_equal as assert_array

from pytest import fixture

import moon_coverage.trajectory.trajectory as tr
from moon_coverage.trajectory import (
    InstrumentTrajectory, SpacecraftTrajectory, Trajectory
)


DATA = (Path(__file__).parent / '..' / 'data').resolve()
KERNELS = [
    str(DATA / 'naif0012.tls'),
    str(DATA / 'dummy.tf'),
    str(DATA / 'dummy.ti'),
    str(DATA / 'dummy.tpc'),
]


@fixture
def traj(monkeypatch):
    """Trajectory with SPICE mocks."""
    # SPICE mocks
    def spice_mocker(func, value):
        monkeypatch.setattr(tr, func, lambda *args, **kwargs: np.transpose([value]))

    a = np.round(2631.2 / np.sqrt(2), 1)

    mock_values = {
        'sub_obs_pt': [2631.2, 0, 0],
        'boresight_pt': [a, a, 0],
        'sc_state': [5_000, 0, 0, 0, 0, 50],
        'groundtrack_velocity': 25,
        'attitude': [[0, 1, 0], [-1, 0, 0], [0, 0, 1]],
        'sun_pos': [300_000_000, 300_000_000, -300_000_000],
        'solar_longitude': 33,
        'true_anomaly': 15,
        'local_time': 12.5,
    }

    for func, value in mock_values.items():
        spice_mocker(func, value)

    # Additional mocks
    def illum_angles(*args, **kwargs):  # pylint: disable=unused-argument
        """Surface local illumination mock."""
        is_sub_obs_pt = args[3][1] == 0
        return np.transpose([
            [54.7, 0, 54.7] if is_sub_obs_pt else [35.26, 30, 60],
        ])

    monkeypatch.setattr(tr, 'illum_angles', illum_angles)

    return Trajectory(
        KERNELS,
        'JUICE',
        'Ganymede',
        '2033-01-01',
    )


def test_traj_properties(traj):
    """Test generic trajectory properties."""
    assert isinstance(traj, Trajectory)

    # ET and UTC
    assert_array(traj.ets, [1041422469.184], decimal=3)
    assert traj.utc.dtype == np.dtype('datetime64[ms]')
    assert traj.utc[0] == np.datetime64('2033-01-01')

    # Sub-spacecraft groundtrack (sub_obs_pt)
    assert_array(traj.sc_pts, [[2631.2], [0], [0]])
    assert_array(traj.sc_rlonlat, [[2631.2], [0], [0]])

    # Boresight surface intersection (boresight_pt)
    assert_array(traj.boresight_pts, [[1860.5], [1860.5], [0]])
    assert_array(traj.boresight_rlonlat, [[2631.2], [45], [0]], decimal=1)

    # Surface/Limb condition
    assert_array(traj.surface, [True])
    assert_array(traj.limb, [False])

    # Observer position and velocity
    assert_array(traj.sc_pos, [[5_000], [0], [0]])
    assert_array(traj.sc_velocity, [[0], [0], [50]])
    assert_array(traj.sc_speed, [50])

    assert_array(traj.dist, [5_000])
    assert_array(traj.alt, [5_000 - 2631.2])

    # Target groundtrack velocity
    assert_array(traj.groundtrack_velocity, [25])

    # C-matrix pointing
    assert_array(traj.ra, [0])
    assert_array(traj.dec, [90])

    # Change of boresight (reset boresight pointing)
    traj.boresight = (-1, 0, 0)

    assert_array(traj.ra, [270])
    assert_array(traj.dec, [0])

    # Sun position
    assert_array(traj.sun_lon_e, [45])
    assert_array(traj.sun_lat, [-35.26], decimal=2)
    assert_array(traj.ss, [[45], [-35.26]], decimal=2)

    # Seasonal solar longitude angle
    assert_array(traj.solar_longitude, [33])

    # Target true anomaly angle
    assert_array(traj.true_anomaly, [15])


def test_traj_sc_properties(traj):
    """Test spacecraft trajectory specific properties."""
    sc_traj = traj

    assert isinstance(sc_traj, SpacecraftTrajectory)

    # Sub-spacecraft groundtrack
    assert_array(sc_traj.lon_e, [0])
    assert_array(sc_traj.lat, [0])
    assert_array(sc_traj.lonlat, [[0], [0]])

    # Sub-spacecraft local time and slant distance
    assert_array(sc_traj.local_time, [12.5])
    assert_array(sc_traj.slant, [5_000 - 2631.2])

    # Sub-spacecraft local illumination and viewing geometry
    assert_array(sc_traj.illum_angles, [[54.7], [0], [54.7]])
    assert_array(sc_traj.inc, [54.7])
    assert_array(sc_traj.emi, [0])
    assert_array(sc_traj.phase, [54.7])

    assert_array(sc_traj.day, [True])
    assert_array(sc_traj.night, [False])

    assert_array(sc_traj.nadir, [True])
    assert_array(sc_traj.off_nadir, [False])

    # Sub-spacecraft local solar zenith angle
    assert_array(sc_traj.ell_norm.T, [[1, 0, 0]])
    assert_array(sc_traj.solar_zenith_angle, [54.73], decimal=2)


def test_traj_inst_properties(traj):
    """Test instrument trajectory specific properties."""
    inst_traj = traj['JANUS']

    assert isinstance(inst_traj, InstrumentTrajectory)

    # Boresight intersection groundtrack
    assert_array(inst_traj.lon_e, [45])
    assert_array(inst_traj.lat, [0])
    assert_array(inst_traj.lonlat, [[45], [0]])

    # Boresight intersection local time and slant distance
    assert_array(inst_traj.local_time, [12.5])
    assert_array(inst_traj.slant, [3649.4], decimal=1)

    # Boresight intersection local illumination and viewing geometry
    assert_array(inst_traj.illum_angles, [[35.26], [30], [60]])
    assert_array(inst_traj.inc, [35.26])
    assert_array(inst_traj.emi, [30])
    assert_array(inst_traj.phase, [60])

    assert_array(inst_traj.day, [True])
    assert_array(inst_traj.night, [False])

    assert_array(inst_traj.nadir, [False])
    assert_array(inst_traj.off_nadir, [True])

    # Boresight intersection local solar zenith angle
    assert_array(inst_traj.ell_norm.T, [[0.707, 0.707, 0]], decimal=3)
    assert_array(inst_traj.solar_zenith_angle, [35.26], decimal=2)

    # Cross-track projected iFOV -> Pixel scale
    assert_array(inst_traj.dist, [5_000])
    assert_array(inst_traj.observer.ifov, [15e-6, 15e-6], decimal=6)  # 15 µrad
    assert_array(inst_traj.pixel_scale, [0.044], decimal=3)
